#!/usr/bin/python

##############################################################################
# cqbf-conv.py: Converts between QDIMACS and the GhostQ CQBF format.
# Authors: Will Klieber
#          Samir Sapra (Unprenexing) 
##############################################################################

import sys
import os
import commands
import re
import pdb
import textwrap
import timeit
import StringIO
from pprint import pprint
from random import Random
from array import array
from copy import copy 

PrStage = True  # Print timing info for each stage.

################################################################################
# Utility functions
################################################################################

def die(text): 
    sys.stderr.write("\nERROR: " + text+"\n")
    if sys.argv[1] == '-':
        sys.exit(1)
    pdb.set_trace()

def qdie(text): 
    sys.stderr.write(text + "\n")
    sys.exit(1)

stop = pdb.set_trace
stderr = sys.stderr

timer = timeit.default_timer
start_time = timer()
elapsed_time = (lambda: timer() - start_time)

################################################################################

class LineReader:
    """Reads a file line-by-line."""
    def __init__(self, file):
        self.file = file
        self.cur = None
        self.line_num = 0
        self.advance()

    def advance(self):
        self.cur = self.file.readline()
        self.line_num += 1

    def skip_comment(self):
        while 1:
            if len(self.cur)==0:  # End-Of-File
                return
            self.cur = self.cur.strip()
            if self.cur.startswith('#') or len(self.cur)==0: 
                self.advance()
                continue
            return
        
    def close(self):
        self.file.close()


################################################################################

class ProbInst(object):
    
    def __init__(self):
        self.RepTable = None    # Maps each gate to expression defining it.
        self.QuantPfx = None    # Maps each gate to a list of QuantBlks.
        self.OutGlit = None     # Output gate literal of the circuit.
        # The following are derived from the above.
        self.QtypeFromVar = None # Maps an input var to 'a' or 'e'.
        self.LevByVar = None    # Maps each variable to its nesting level.
        self.VarsByLev = None   # Maps a nesting level to vars in it.
        self.VarPars = None     # Maps a var to the gates it feeds.  See FindParents.

    def DumpRepTable(self):     # For debugging.
        for (gvar, expr) in sorted(self.RepTable.items()):
            print "RepTable[%i] = %s" % (gvar, repr(expr))

    def IsGateLit(self, x):
        return (abs(x) in self.RepTable)
        #ret = (self.LevByVar[abs(x)] > 1)
        ##if (not glo.NoRenum): assert(ret == (abs(x) > self.NumVars))
        #return ret

    def NewGvar(self):
        self.NumVars += 1
        ret = self.NumVars
        return ret
        

class QuantBlk:
    def __init__(self, qtype, qvars):
        self.qtype = qtype
        self.qvars = qvars

    def __repr__(self):
        return "QuantBlk(" + repr(self.qtype) + ", " + repr(self.qvars) + ")"


def read_input_file(filename):
    try:
        file_ptr = (open(filename, 'r') if (filename != '-') else sys.stdin)
    except IOError, e:
        qdie(str(e))
    in_file = LineReader(file_ptr)

    global start_time
    start_time = timer()
    FirstLine = in_file.cur.strip()
    if (FirstLine.startswith("c") or FirstLine.startswith("p cnf")):
        return read_qdimacs_file(in_file)
    for prefix in ["#", "LastInputVar", "CktQBF"]:
        if FirstLine.lower().startswith(prefix.lower()):
            return read_circuit_file(in_file)
    die("Unrecognized file format.")


def read_circuit_file(in_file):
    def LineDie(msg):
        stderr.write("Error on line %i. " % in_file.line_num)
        stderr.write(msg + "\n")
        sys.exit(1)
    def ReadIntParam(pname):
        try:
            ret = int(re.match(pname + ' +(-?[0-9]+)\s*', in_file.cur).group(1))
            in_file.advance()
        except:
            die("Error trying to read '%s' parameter on line %i." % 
                    (pname, in_file.line_num))
        return ret
    prob = ProbInst()
    in_file.skip_comment()
    if (in_file.cur.startswith("CktQBF")): in_file.advance()
    in_file.skip_comment()
    if (in_file.cur.startswith("LastInputVar")): ReadIntParam("LastInputVar")
    if (in_file.cur.startswith("FirstGateVar")): ReadIntParam("FirstGateVar")
    prob.NumVars = ReadIntParam("LastGateVar")
    prob.NumOrigVars = prob.NumVars
    prob.OutGlit = ReadIntParam("OutputGateLit")
    if (in_file.cur.startswith("PreprocTimeMilli")): ReadIntParam("PreprocTimeMilli")
    prob.NumElim = 0
    prob.QuantPfx = {}
    prob.RepTable = {}
    NumVarsDecl = prob.NumVars

    HasAnyQ = False

    while True:
        in_file.skip_comment()
        CurLine = in_file.cur
        if (CurLine == "<q>") and not HasAnyQ:
            GateNum = abs(prob.OutGlit)
        else:
            try: 
                GateNum = int(re.match('<q gate=(-?[0-9]+)>', CurLine).group(1))
            except: 
                if (HasAnyQ): break
                else: LineDie("Expected a line of the form '<q>' or '<q gate=i>'.")
        HasAnyQ = True
        QuantBlks = []
        while True:
            in_file.advance()
            in_file.skip_comment()
            if len(in_file.cur)==0: LineDie("File ended unexpectedly.")
            CurLine = in_file.cur
            if (CurLine[0].lower() not in ['a','e']): break
            CurQuant = QuantBlk(CurLine[0], [int(x) for x in CurLine[1:].split()])
            QuantBlks.append(CurQuant)
        if (CurLine != "</q>"): LineDie("Expected '</q>'.")
        in_file.advance()
        prob.QuantPfx[GateNum] = QuantBlks

    def MakeXor(NewGate, args):
        if len(args)==1:
            return args[0]
        if (NewGate == 0):
            NewGate = prob.NumVars+1
            prob.NumVars += 1
        temp1 = prob.NumVars+1
        temp2 = prob.NumVars+2
        prob.NumVars += 2
        x = args[0]
        y = MakeXor(0, args[1:])
        prob.RepTable[temp1] = tuple(['or', x, y])
        prob.RepTable[temp2] = tuple(['or', -x, -y])
        prob.RepTable[NewGate] = tuple(['and', temp1, temp2])
        return NewGate

    while True:
        if len(in_file.cur)==0: break
        CurLine = in_file.cur.strip().lower()
        if CurLine.startswith('#') or len(CurLine)==0: 
            in_file.advance()
            continue
        try:
            [GateVar, op, args] = re.match('([0-9]+) += +([a-zA-Z]+)\((.+)\)', in_file.cur).groups()
            GateVar = int(GateVar)
            if GateVar > NumVarsDecl:
                die("Line %i: Variable number (%i) exceeds maximum declared (%i)." %
                    (in_file.line_num, GateVar, NumVarsDecl))
            args = [int(x) for x in args.replace(",", " ").split()]
            if (op in ["and", "or"]):
                expr = tuple([op] + args)
                prob.RepTable[GateVar] = expr 
            elif (op == 'xor'):
                MakeXor(GateVar, args)
            else:
                die("Unrecognized operator: '%s'" % op)
        except Exception, e:
            die("Error trying to parse line %i." % (in_file.line_num))
        in_file.advance()

    in_file.close()
    if (PrStage): sys.stderr.write("Done reading file, %.4f sec\n" % elapsed_time())
    return prob


def read_qdimacs_file(cnf_file):
    # Skip comment lines
    while cnf_file.cur.startswith("c") or cnf_file.cur == "\n":
        cnf_file.advance()
    
    # Read the problem line
    prob_line = cnf_file.cur
    try:
        [NumVars, NumClauses] = [int(num) for num in 
                re.match('p cnf +([0-9]+) +([0-9]+)\s*', prob_line).groups()]
    except:
        die("Error reading the problem line.")
    cnf_file.advance()
    
    # Read the quantifiers
    QuantBlks = []
    while cnf_file.cur[0].lower() in ['a', 'e']:
        qtype = cnf_file.cur[0].upper()
        try:
            qvars = [int(lit) for lit in cnf_file.cur[1:].split()]
        except:
            die("Error reading quantifier on line #%d." % (cnf_file.line_num))
        if qvars[len(qvars)-1] != 0:
            die("Error: Line %i does not end in a 0." % cnf_file.line_num)
        QuantBlks.append( QuantBlk(qtype, qvars[0:len(qvars)-1]) )
        cnf_file.advance()

    # Read the clauses
    HasVar = array('B', [False] * (NumVars+1))
    Clauses = [None] * NumClauses
    for i in range(0, NumClauses):
        try:
            cur = tuple([int(lit) for lit in cnf_file.cur.split()])
        except:
            die("Error reading clause #%d:\n%s" % (i, cnf_file.cur))
        if len(cur) == 0: die("Fewer clauses exist than declared.")
        if cur[len(cur)-1] != 0:
            die("Error: Line %i does not end in a 0." % cnf_file.line_num)
        cur = cur[0:len(cur)-1]  # Remove terminating "0"
        for var in (abs(x) for x in cur):
            if var > NumVars:
                die("Line %i: Variable number (%i) exceeds maximum declared (%i)." %
                    (cnf_file.line_num, var, NumVars))
            HasVar[var] = True
        Clauses[i] = cur
        cnf_file.advance()
    if cnf_file.cur.strip() != "": die("More clauses exist than declared.")
    NumOrigVars = 0
    for var in range(1, NumVars+1):
        if HasVar[var]:
            NumOrigVars += 1

    cnf_file.close()
    return cnf_to_nnf(NumOrigVars, NumVars, QuantBlks, Clauses)


def cnf_to_nnf(NumOrigVars, NumVars, QuantBlks, Clauses):
    prob = ProbInst()
    prob.NumOrigVars = NumOrigVars
    prob.NumVars = NumVars

    NumOrigClauses = len(Clauses)
    if (PrStage): sys.stderr.write("Done reading file, %.4f sec\n" % elapsed_time())

    # Compute which vars are eligible to be representative vars
    if QuantBlks == []:
        elig = array('B', [True] * (NumVars+1))
    else:
        elig = array('B', [False] * (NumVars+1))
        inr = QuantBlks[-1]  # innermost quantifier
        if (inr.qtype.lower() == 'e'):
            for var in inr.qvars:
                if var > NumVars:
                    die("Bad var number %i: exceeds number of declared vars." % var)
                elig[var] = True
        else: assert(inr.qtype == 'A')

    QbFromLit = {}
    QtypeFromLit = {}
    for (ixQb, qb) in enumerate(QuantBlks):
        for v in qb.qvars:
            QbFromLit[v] = ixQb
            QtypeFromLit[v] = qb.qtype
    
    DelClauses = set()
    RepTable = {}

    flag = (not glo.NoRevEngr)
    round = 0
    NumElim = 0
    while flag:

        round += 1
        flag = False

        # Compute paired literals
        paired = {}
        for var in range(1, NumVars+1):
            paired[var] = dict()
            paired[-var] = dict()
            
        for (ixCl, CurCl) in enumerate(Clauses):
            if len(CurCl) != 2:
                continue
            paired[CurCl[0]][CurCl[1]] = ixCl
            paired[CurCl[1]][CurCl[0]] = ixCl
        
        xsub = [set() for var in range(0, NumVars+1)]

        # (v <==> (x1 | x2 | x3)) expands to
        # (~v | x1 | x2 | x3)   &   (~x1 | v)  &  (~x2 | v)  &  (~x3 | v)
        #
        # (v <==> (x1 & x2 & x3)) expands to
        # (~x1 | ~x2 | ~x3 | v)  &  (~v | x1)  &  (~v | x2)  &  (~v | x3)
    
     
        DelClauses = set()
        ClauseSet = set(Clauses)
        for (ixCl, CurCl) in enumerate(Clauses):
            if len(CurCl) < 2:
                continue
            RepLits = []
            #print "CLAUSE ", CurCl
            MaxQb = max(QbFromLit[abs(x)] for x in CurCl)
            for TestLit in CurCl:
                if QbFromLit[abs(TestLit)] != MaxQb: 
                    assert( not elig[abs(TestLit)])
                    continue
                if (QtypeFromLit[abs(TestLit)].lower() != 'e'):
                    continue
                #if not elig[abs(TestLit)]: continue
                for CurLit in CurCl:
                    if CurLit == TestLit: continue
                    if -TestLit not in paired[-CurLit]:
                        break
                    if abs(ixCl - paired[-CurLit][-TestLit]) > len(CurCl):
                        break
                else:
                    RepLits.append(TestLit)
            if len(RepLits) != 1:
                continue
            RepLit = RepLits[0]
            SubLits = [lit for lit in CurCl if lit != RepLit]
            #
            has_cycle = False
            for CurLit in SubLits:
                if abs(RepLit) in xsub[abs(CurLit)]:
                    has_cycle = True
            if has_cycle: continue
            #
            NumElim += 1
            for CurLit in SubLits:
                xsub[abs(RepLit)].add(abs(CurLit))
                xsub[abs(RepLit)] |= xsub[abs(CurLit)]
            #
            if RepLit in RepTable: die("Unimplemented")
            [pos_op, neg_op] = ['and', 'or']
            if RepLit > 0: RepTable[RepLit] = tuple([pos_op] + [-lit for lit in SubLits])
            if RepLit < 0: RepTable[-RepLit] = tuple([neg_op] + SubLits)
            #flag = True
            #
            DelClauses.add(CurCl)
            for CurLit in SubLits:
                CurDel = (-RepLit, -CurLit)
                if (CurDel not in ClauseSet):
                    CurDel = (-CurLit, -RepLit)
                assert(CurDel in ClauseSet)
                DelClauses.add(CurDel)

        Clauses = [clause for clause in Clauses if (clause not in DelClauses)]
        break


    if (PrStage): sys.stderr.write("Done finding gate variables, %.4f sec\n" % elapsed_time())
        

    ##################################################################

    DelClauses = []

    def TryMux(WorkCl, glit, sel):
        # A gate defn of the form (glit = (sel ? a : b)) is encoded by the
        # following four clauses:
        # ((sel  & glit  & a)  |
        #  (sel  & ~glit & ~a) | 
        #  (~sel & glit  & b)  | 
        #  (~sel & ~glit & ~b))
        # 
        if (abs(glit) == abs(sel)): return None
        PosA = NegA = PosB = NegB = 0
        def GetIn(cl):
            return (cl - set([glit, -glit, sel, -sel])).pop()
        for cl in WorkCl:
            if    (-sel in cl) and  (glit in cl): NegA = GetIn(cl)
            elif  (-sel in cl) and (-glit in cl): PosA = GetIn(cl)
            elif   (sel in cl) and  (glit in cl): NegB = GetIn(cl)
            elif   (sel in cl) and (-glit in cl): PosB = GetIn(cl)
            else: return None
        if (0 in [PosA, PosB, NegA, NegB]): return None
        if (PosA != -NegA): return None
        if (PosB != -NegB): return None
        if len(set(abs(x) for x in [glit,sel,PosA,PosB])) < 3:
            return None
        return (PosA, PosB)
        
    def LookForMux(WorkCl, CurCl):
        CurCl = [abs(x) for x in CurCl]
        for glit in CurCl:
            if (not elig[glit]): continue
            if (glit in RepTable): continue
            PosSel = tuple(set(CurCl) - set([glit]))
            for j in [0,1]:
                test = TryMux(WorkCl, glit, PosSel[j])
                if (test != None): return (glit, PosSel[j], test[0], test[1])
        return None

    def IsDesc(child, par, hit):
        if (par in hit): return False
        hit.add(par)
        par = abs(par)
        if (par not in RepTable): return False
        for x in RepTable[par][1:]:
            if abs(x)==abs(child): return True
            if IsDesc(child, x, hit): return True
        return False

    if (1 and not glo.NoRevEngr): 
        for (ixCl, CurCl) in enumerate(Clauses):
            if (ixCl < 3): continue
            WorkCl = [set(Clauses[ixCl-j]) for j in [0,1,2,3]]
            if [len(x) for x in WorkCl] != [3,3,3,3]: continue
            ret = LookForMux(WorkCl, CurCl)
            if ret != None:
                (glit, sel, PosA, PosB) = ret
                dhit = set()
                if IsDesc(glit, sel, dhit) or IsDesc(glit, PosA, dhit) or IsDesc(glit, PosB, dhit):
                    continue
                vp1 = prob.NewGvar()
                vp2 = prob.NewGvar()
                RepTable[vp1] = tuple(['and', sel, PosA])
                RepTable[vp2] = tuple(['and', -sel, PosB])
                RepTable[glit] = tuple(['or', vp1, vp2])
                DelClauses += [(Clauses[ixCl-j]) for j in [0,1,2,3]]


    DelClauses = set(DelClauses)

    Clauses = [clause for clause in Clauses if (clause not in DelClauses)]

    NewClauses = list(reversed([clause for clause in Clauses if (clause not in DelClauses)]))




    ##################################################################

    # Convert the remaining top-level conjuncts.
    OutGlit = None
    if len(NewClauses) == 1 and len(NewClauses[0]) == 1:
        OutGlit = NewClauses[0][0]
    else:
        OutGlit = []
        for CurCl in NewClauses:
            # (v <==> (x1 | x2 | x3)) expands to
            # (~v | x1 | x2 | x3)   &   (~x1 | v)  &  (~x2 | v)  &  (~x3 | v)
            if len(CurCl) == 1:
                OutGlit.append(CurCl[0])
                continue
            NewVar = prob.NewGvar()
            RepTable[NewVar] = tuple(['or'] + list(CurCl))
            OutGlit.append(NewVar)
        if len(OutGlit) == 1:
            OutGlit = OutGlit[0]
        else:
            # (v <==> (x1 & x2 & x3)) expands to
            # (~x1 | ~x2 | ~x3 | v)  &  (~v | x1)  &  (~v | x2)  &  (~v | x3)
            NewVar = prob.NewGvar()
            RepTable[NewVar] = tuple(['and'] + OutGlit)
            OutGlit = NewVar

    if OutGlit < 0:
        RepTable[-OutGlit] = NegateExpr(RepTable[-OutGlit])
        OutGlit = -OutGlit

        
    prob.QuantPfx = {}
    prob.QuantPfx[OutGlit] = QuantBlks
    prob.OutGlit = OutGlit
    prob.RepTable = RepTable
    prob.NumElim = NumElim
    return prob


def FindGateOccur(p):
    p.NumGateOccur = array('B', [0] * (p.LastGateVar+1))
    for expr in p.RepTable.values():
        for lit in expr[1:]:
            if not p.IsGateLit(lit): continue
            p.NumGateOccur[abs(lit)] = min(2, p.NumGateOccur[abs(lit)]+1)

def FindParents(p):
    p.VarPars = {}
    for (RepVar, expr) in p.RepTable.items():
        for lit in expr[1:]:
            p.VarPars.setdefault(abs(lit), [])
            p.VarPars[abs(lit)].append(RepVar)

def OccursOnce(p, glit):
    gvar = abs(glit)
    if (gvar == abs(p.OutGlit)): return True
    pars = p.VarPars[gvar]
    if len(pars) != 1: return False
    return OccursOnce(p, pars[0])

def NegateExpr(expr):
    if (expr == "true"): return "false"
    if (expr == "false"): return "true"
    expr = list(expr)
    expr[0] = {'and':'or', 'or':'and'}[expr[0]]
    expr[1:] = [-arg for arg in expr[1:]]
    return tuple(expr)
    
def CombineSimplify(p, ixExpr, HitSet):
    """Convert AND(x,AND(y,z)) to AND(x,y,z), etc."""
    if HitSet==None: HitSet = set()
    if (ixExpr in HitSet): return
    HitSet.add(ixExpr)
    OldTop = p.RepTable[ixExpr]
    if (OldTop == "true" or OldTop == "false"): return
    p.PresLits[ixExpr] = set()
    HasDual = [0]
    NewArgs = set()
    def add(CurList, x):
        if (type(x)==int and (x in p.TrivLits)):
            x = {True:"true", False:"false"}[p.TrivLits[x]]
        if (x == "true"):
            if OldTop[0] == 'and': return
            assert(OldTop[0] == 'or')
            HasDual[0] = True
            return
        if (x == "false"):
            if OldTop[0] == 'or': return
            assert(OldTop[0] == 'and')
            HasDual[0] = True
            return
        if (-x in CurList): HasDual[0] = x
        CurList.add(x)
    for ixSubExpr in OldTop[1:]:
        if not p.IsGateLit(abs(ixSubExpr)):
            add(NewArgs, ixSubExpr)
            continue
        CombineSimplify(p, abs(ixSubExpr), HitSet)
        SubPres = p.PresLits[abs(ixSubExpr)]
        if ixSubExpr < 0:
            SubExpr = NegateExpr(p.RepTable[-ixSubExpr])
            SubPres = set(-x for x in SubPres)
        else:
            SubExpr = p.RepTable[ixSubExpr]
        if (SubExpr == "true" or SubExpr == "false"):
            add(NewArgs, SubExpr)
            continue

        if (len(SubExpr[1:]) == 1):
            add(NewArgs, SubExpr[1:][0])
            p.PossibleJunk.add(abs(ixSubExpr))
            continue

        p.PresLits[ixExpr] |= SubPres
        
        if (SubExpr[0] == OldTop[0]) and (glo.OptCombine >= p.NumGateOccur[abs(ixSubExpr)]) and (ixSubExpr not in p.QuantPfx) and (-ixSubExpr not in NewArgs):
            for NewLit in SubExpr[1:]:
                add(NewArgs, NewLit)
            p.PossibleJunk.add(abs(ixSubExpr))
            p.NumCombined += 1
        else:
            add(NewArgs, ixSubExpr)
    if (HasDual[0]):
        #assert(False)
        if (OldTop[0] == 'or'):     p.RepTable[ixExpr] = "true"
        elif (OldTop[0] == 'and'):  p.RepTable[ixExpr] = "false"
        else: assert(False)
        return
    if len(NewArgs)==0:
        if (OldTop[0] == 'or'):     p.RepTable[ixExpr] = "false"
        elif (OldTop[0] == 'and'):  p.RepTable[ixExpr] = "true"
        else: assert(False)
        return
    p.PresLits[ixExpr] |= set([x for x in NewArgs if not p.IsGateLit(x)])
    NewArgs = [OldTop[0]] + sorted(NewArgs, key=abs)
    p.RepTable[ixExpr] = tuple(NewArgs)

def TrimJunk(p):
    """Removes expressions made redundant by CombineSimplify."""
    RefdExprs = set()
    def AddRefd(ixExpr):
        if (ixExpr in RefdExprs):
            return
        RefdExprs.add(ixExpr)
        if (ixExpr not in p.RepTable):
            return
        Expr = p.RepTable[ixExpr]
        for arg in Expr[1:]:
            AddRefd(abs(arg))
    AddRefd(abs(p.OutGlit))
    UselessExprs = []
    Subsumed = set()
    for ixExpr in p.RepTable.keys():
        if (ixExpr not in RefdExprs):
            if (ixExpr not in p.PossibleJunk):
                UselessExprs.append(p.MapNewToOld[ixExpr])
            else:
                Subsumed.add(p.MapNewToOld[ixExpr])
            del p.RepTable[ixExpr]
            if (ixExpr in p.QuantPfx):
                del p.QuantPfx[ixExpr]

    #if len(UselessExprs) > 0:
    #    sys.stderr.write("Note: Input file has useless subformulas, headed " + \
    #            "by the following variables: " + str(UselessExprs) + "\n")
    return (Subsumed, UselessExprs)

def FindVarLevs(p):
    RecLim = 5000
    sys.setrecursionlimit(RecLim)
    LevByVar = [0] * (p.NumVars + 1)
    VarsByLev = {}
    GateVars = set(p.RepTable.keys())
    hit = set()
    def FindVarLev(var, rec):
        if (rec > RecLim):
            stop()
            qdie("ERROR: Exceeded recursion limit!\n" + 
                 "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n")
        var = abs(var)
        if (LevByVar[var] != 0): return LevByVar[var]
        if var in hit:
            qdie("ERROR: Cycle in RepTable!\n") 
        hit.add(var)
        if (var not in GateVars): 
            ret = 1
        else:
            ChildLevs = [FindVarLev(x,rec+1) for x in p.RepTable[var][1:]]
            ret = max(ChildLevs)+1
        LevByVar[var] = ret
        VarsByLev.setdefault(ret, set())
        VarsByLev[ret].add(var)
        return ret
    FindVarLev(p.OutGlit, 100)
    for QuantBlks in p.QuantPfx.values():
        for OrigQb in QuantBlks:
            qtype = OrigQb.qtype
            assert(qtype.lower() in ('a', 'e'))
            QbSet = set(OrigQb.qvars)
            for lit in OrigQb.qvars:
                if LevByVar[lit] != 1:
                    QbSet.remove(lit)
            OrigQb.qvars = list(sorted(QbSet))
    if (PrStage): sys.stderr.write("Done FindVarLevs, %.4f sec\n" % elapsed_time())
    p.VarsByLev = VarsByLev
    p.LevByVar = LevByVar
    return 

def FindQtypes(p):
    p.QtypeFromVar = {}
    for qp in p.QuantPfx.values():
        for qb in qp:
            for var in qb.qvars:
                if (var in p.QtypeFromVar):
                    die("Variable %i is quantified in more than once place!" % var)
                p.QtypeFromVar[var] = qb.qtype.lower()

def ChkUnquantVars(p):
    QuantVars = set()
    for qp in p.QuantPfx.values():
        for qb in qp:
            QuantVars |= set(qb.qvars)
    unq = set()
    for (lev, lvars) in p.VarsByLev.items():
        unq |= (lvars - QuantVars)
    if len(unq) > 0:
        print "Warning: unquantified variables: ", unq

def RenumExprs(p):
    """Renumber the expressions.  Postcond: Every new expr number will be even."""
    VarsByLev = p.VarsByLev
    InpVars = VarsByLev[1]
    QuantVars = set()
    for qp in p.QuantPfx.values():
        for qb in qp:
            QuantVars |= set(qb.qvars)
    #
    FreeVars = sorted(InpVars - QuantVars)
    if (len(FreeVars) > 0): 
        print "There are unquantified variables in the QBF matrix:"
        print FreeVars
        stop()
        qdie("")
    #
    map = {}
    increm = glo.VarIncrem
    i = glo.InputVarStart
    p.QtypeFromVar = {}
    for (qgate, blks) in SortedQuantPfx(p):
        for qb in blks:
            assert(qb.qtype.upper() in ['A','E'])
            for OldVar in sorted(set(qb.qvars) & InpVars):
                map[OldVar] = i
                i += increm
    p.NumVars = i - increm

    NewRepTbl = {}
    def conv(val):
        if type(val)==str: return val
        assert(type(val)==int)
        if val > 0: return map[val]
        if val < 0: return -map[abs(val)]

    i = int('1' + '0'*len(str(p.NumVars)))
    if (p.NumVars >= 100):
        i = int( str(int(str(p.NumVars)[0])+1) + '0'*(len(str(p.NumVars))-1) )
    base = i
    p.FirstExpr = base + increm 
    i = p.FirstExpr
    for lev in sorted(VarsByLev.keys()):
        if lev <= 1: continue
        for OldExprIx in (sorted(VarsByLev[lev])):
            map[OldExprIx] = i
            NewRepTbl[i] = tuple([conv(x) for x in p.RepTable[OldExprIx]])
            i += increm
    p.LastGateVar = max(map.values())

    p.OutGlit = conv(p.OutGlit)
    OldQuantPfx = p.QuantPfx
    p.QuantPfx = {}
    for (qgate, blks) in OldQuantPfx.items():
        for qb in blks:
            qb.qvars = [abs(conv(x)) for x in qb.qvars]
        p.QuantPfx[conv(qgate)] = blks
    p.MapNewToOld = dict([(val, key) for (key, val) in map.items()])
    LevByOldVar = p.LevByVar
    p.LevByVar = {}
    p.VarsByLev = {}
    for (key, expr) in p.RepTable.items():
        if (LevByOldVar[key] == 0): continue
        NewExpr = NewRepTbl[map[key]]
        assert(NewExpr[0] == expr[0])
        for i in range(1, len(expr)):
            old = map[abs(expr[i])]
            if (expr[i] < 0): old = -old
            assert(NewExpr[i] == old)
    for (OldVar, lev) in enumerate(LevByOldVar):
        if OldVar not in map:
            continue
        NewVar = map[OldVar]
        p.LevByVar[NewVar] = lev
        p.VarsByLev.setdefault(lev, set())
        p.VarsByLev[lev].add(NewVar)
    p.OldRepTable = p.RepTable
    p.RepTable = NewRepTbl

def GatesSorted(p, dir, MinLev=2):
    for (lev, vars) in sorted(p.VarsByLev.items(), key=(lambda (lev,v): lev*dir)):
        if (lev < MinLev):
            continue
        for v in vars:
            yield v

def Gates_BotUp(p, MinLev=2):
    for x in GatesSorted(p, 1, MinLev):
        yield x

def Unprenex(p):
    if not glo.TryUnprenex: return
    p.NumPushed = 0
    dp = 0  # debug print
    assert(type(p) == ProbInst)
    FindParents(p)
    if dp: print p.VarsByLev
    if dp: print "\n"

    def getGateDesc(g):
        if g not in GateDesc:
            if p.LevByVar[g] == 1:
                GateDesc[g] = set([g]);
            else:
                GateDesc[g] = set()
                for gi in p.RepTable[g][1:]:
                    GateDesc[g] |= getGateDesc(abs(gi))
        return GateDesc[g]
    GateDesc = {}
    if (p.OutGlit < 0):
        return
    getGateDesc(p.OutGlit) # p.OutGlit must be >0
    #print GateDesc;

    for (lev, gates) in sorted(p.VarsByLev.items()):
        if (lev < 3): continue  # Skip input vars and lowest-level gates.
        gates = copy(gates)
        gates &= set(p.QuantPfx.keys())  # Set-intersect to exclude non-quantified gates
        for g in gates:
            for QB_idx in xrange(len(p.QuantPfx[g])-1, -1, -1):
                QB = p.QuantPfx[g][QB_idx]
                if dp: print g,",",QB_idx,": ", QB
                if dp: print "(Type %s %s)" % (QB.qtype, QB.qtype.lower())
                x_qtype = QB.qtype.lower()
                for var_idx in xrange(len(QB.qvars)-1, -1, -1):
                    x = QB.qvars[var_idx]
                    if dp: print "Var %d:" % x
                    g_new = g
                    pos = QB_idx
                    while 1:
                        if dp: print "        Gate %d, %d (%s) of %d" \
                              % (g_new, pos, \
                              len(p.QuantPfx[g_new]) and \
                                 p.QuantPfx[g_new][pos].qtype, \
                              len(p.QuantPfx[g_new]) - 1)
                        if pos < len(p.QuantPfx[g_new]) - 1:
                            if dp: print "        New QType: %s" % \
                                  p.QuantPfx[g_new][pos+1].qtype
                            if p.QuantPfx[g_new][pos+1].qtype.lower() \
                                   == x_qtype:
                                pos += 1
                            else:
                                break # stop: we hit a complementary
                                      # quantifier
                        else: # here pos == len(p.QuantPfx[g_new]) - 1
                            subscopes = [child \
                                         for child in p.RepTable[g_new][1:] \
                                         if x in GateDesc[abs(child)]]
                            if dp: print "        Subscopes: ", subscopes
                            if len(subscopes) == 1 and \
                                   subscopes[0] > 0 and \
                                   p.LevByVar[subscopes[0]] > 1:
                                g_new = subscopes[0]
                                p.QuantPfx.setdefault(g_new, [])
                                pos = -1
                            else:
                                break # stop: we've reached all the
                                      # way down to the immediate dominator
                    # Time to delete and re-insert
                    if (g_new, pos) != (g, QB_idx):
                        p.NumPushed += 1
                        if dp: print "Moving %d from (%d,%d) to (%d, %d)" % \
                              (x, g, QB_idx, g_new, pos)
                        del QB.qvars[var_idx]
                        if pos == -1:
                            p.QuantPfx[g_new].insert(0, QuantBlk(x_qtype, [x]))
                        else:
                            p.QuantPfx[g_new][pos].qvars.append(x)
                if len(QB.qvars) == 0:
                    del p.QuantPfx[g][QB_idx]
                if dp: print
            pass
            # TODO: Try to push quantifiers inward

    # GATE SPLITTING #################################################
    SML_MAGIC_NUM=4
    MED_MAGIC_NUM1=8
    MED_MAGIC_NUM2=8
    BIG_MAGIC_NUM=16
    for (lev, gates) in sorted(p.VarsByLev.items()):
        if (lev < 2): continue  # Skip input vars
        #for g in gates:
        if dp: print "Level %d: " % (lev),
        if dp: print sorted([len(GateDesc[g]) for g in gates])
        for g in gates:
            # 'continue's
            if g not in p.QuantPfx or len(p.QuantPfx[g]) == 0: continue
            if len(p.RepTable[g]) < BIG_MAGIC_NUM: continue
            
            if dp: print "Level %d, Gate %d [%d]: " % (lev, g, len(GateDesc[g])),
            if dp: print sorted([len(GateDesc[abs(child)]) for child in p.RepTable[g][1:]])
            
            Vars2gates = {}
            VarsByLen = {}
            VarsByLen[0] = set()
            VarsByLen[1] = set()
            VarsByLen[2] = set()
            VarsByLen[3] = set()
            VarsByLen[4] = set()
            VarsByLen[5] = set()
            for ch in p.RepTable[g][1:]:
                child = abs(ch)

                # Potential optimization: Figure out how to soundly
                # ignore large child gates

                for var in GateDesc[child]:
                    if var not in p.QuantPfx[g][-1].qvars: continue
                    #   alternate way of iterating
                    
                    Vars2gates.setdefault(var,[])
                    Vars2gates[var].append(child) #get too big?
                    newvarlen = len(Vars2gates[var])
                    if newvarlen == 1:
                        VarsByLen[newvarlen].add(var)
                    elif newvarlen <= SML_MAGIC_NUM+1: #5:
                        VarsByLen[newvarlen-1].remove(var)
                        VarsByLen[newvarlen].add(var)
            if dp: print "Potential Inverse Map:", p.QuantPfx[g][-1].qvars
            if dp: print "Invserse Map:", VarsByLen
            if dp: print "Invserse Map (contd):", Vars2gates

            for varlen in xrange(SML_MAGIC_NUM, 1, -1): #[4, 3, 2]:
                while len(VarsByLen[varlen]) > 0:
                    var = iter(VarsByLen[varlen]).next() #.pop() later
                    if dp: print var
                    assert var in p.QuantPfx[g][-1].qvars
                    if dp: print "Useful to split"

                    g_new = p.NewGvar()
                    if dp: print p.RepTable[g]
                    p.RepTable[g_new] \
                        = tuple([p.RepTable[g][0]] \
                                + [child \
                                   for child in p.RepTable[g][1:] \
                                   if abs(child) in Vars2gates[var]])
                    if dp: print p.RepTable[g_new]
                    p.RepTable[g] = tuple(\
                        [p.RepTable[g][0]] \
                        + [child for child in p.RepTable[g][1:] \
                           if abs(child) not in Vars2gates[var]] \
                        + [g_new])
                    if dp: print p.RepTable[g]

                    # Handle below (more elegant methinks)
                    if dp: print "Moving quantifier for ", var
                    origQB = p.QuantPfx[g][-1]
                    #origQB.qvars.remove(var)
                    #p.QuantPfx[p.NumVars] = [QuantBlk(origQB.qtype, [var])]
                    #for child_idx in xrange(len(p.RepTable[g])-1, 0,-1):
                    #    child = p.RepTable[g][child_idx]
                    #    if abs(child) in Vars2gates[var]:
                    #        p.RepTable[p.NumVars].append(child)
                    #        del p.RepTable[g][child_idx] #slow
                    #p.RepTable[g].append(p.NumVars)

                    separatedgates = Vars2gates[var]
                    affectedvars = set()
                    for sg in separatedgates:
                        affectedvars |= GateDesc[sg] #might get too big. Opt?
                    if dp: print affectedvars

                    sg_vars_subsumed = []
                    sg_vars_notsubsumed = []
                    for sg_var in Vars2gates:
                        if sg_var not in affectedvars: continue
                        if set(Vars2gates[sg_var]) <= set(separatedgates):
                            sg_vars_subsumed.append(sg_var)
                        else:
                            sg_vars_notsubsumed.append(sg_var)
                            
                    if dp: print sg_vars_subsumed
                    if dp: print sg_vars_notsubsumed
                            
                    origQB.qvars = list(
                        set(origQB.qvars) - set(sg_vars_subsumed))
                    p.QuantPfx[g_new] = [
                        QuantBlk(origQB.qtype, sg_vars_subsumed)]
                    if dp: print p.QuantPfx[g_new]
                    for sg_var in sg_vars_subsumed + sg_vars_notsubsumed:
                        count = len(Vars2gates[sg_var])
                        if count > 4: VarsByLen[5].remove(sg_var)
                        else: VarsByLen[count].remove(sg_var)
                        del Vars2gates[sg_var]
    FindVarLevs(p)
    FindParents(p)


    # Special case for dmeX_X instances
    GateDesc = {}
    getGateDesc(p.OutGlit)
    MoveQ = {}

    for (b, qp) in p.QuantPfx.items():
        if len(qp) != 1: continue  # only one quantifier block in the prefix.
        if len(qp[0].qvars) != 1: continue
        u = qp[0].qvars[0]
        if (u not in GateDesc[b]): continue
        if (len(GateDesc[b]) != 2): continue
        e = list(GateDesc[b] - set([u]))[0]
        assert(GateDesc[b] == set([e,u]))
        if (set(p.VarPars[e]) != set(p.RepTable[b][1:])): continue
        MoveQ[e] = b

    for (b, qp) in p.QuantPfx.items():
        for qb in qp:
            qb.qvars = [x for x in qb.qvars if (x not in MoveQ)]
    for (e, b) in MoveQ.items():
        p.QuantPfx[b].insert(0, QuantBlk(p.QtypeFromVar[e], [e]))

        

                    


def SortedQuantPfx(p):
    return sorted(p.QuantPfx.items(), key=(lambda qp: (-p.LevByVar[qp[0]], -abs(qp[0]))))
                

class TrivLitDict:
    def __init__(self):
        self.d = dict()
    
    def __setitem__(self, k, v):
        assert(v in [True, False])
        assert(type(k)==int)
        if (k < 0):
            k = -k
            v = (not v) 
        self.d[k] = v
    
    def __getitem__(self, k):
        if (k < 0): return (not self.d[-k])
        return self.d[k]
    
    def __contains__(self, k):
        return self.d.__contains__(abs(k))

def FindTrivLits(p):
    p.TrivLits = TrivLitDict()
    TopExpr = p.RepTable[p.OutGlit]
    if (TopExpr[0] == 'and'):
        for lit in TopExpr[1:]:
            if not p.IsGateLit(lit):
                p.TrivLits[lit] = True
            

TrivTrueFile = StringIO.StringIO(
"""
p cnf 4 2
a 2 0
e 4 0
-2 -4 0
2 4 0
""")

TrivFalseFile = StringIO.StringIO(
"""
p cnf 4 2
e 2 0
a 4 0
-2 -4 0
2 4 0
""")


class GloOpt:
    pass


def main():
    argv = sys.argv
    if len(argv) < 2 or (argv[1] in ["-h", "--help", "-help"]):
        print "Usage: 'python %s inputfile [-o outfile] [options]'" % (sys.argv[0])
        sys.exit(1)

    global PrStage
    global glo
    glo = GloOpt()

    glo.VarIncrem = 2

    filename = sys.argv[1]
    PrettyName = filename
    OnlyAnd = False
    WriteOldMap = False
    glo.WriteQdimacs = False
    glo.NoRevEngr = False
    glo.InputVarStart = 2
    glo.OptCombine = 1
    glo.NoRenum = False
    glo.NoSimplify = False
    glo.ConvOr = True 
    glo.UnprenexSplit = False
    glo.TryUnprenex = False
    OutFilename = ""
    outf = sys.stdout
    i=2
    while (i < len(argv)):
        sArg = argv[i]
        if sArg.startswith('--'):
            sArg = sArg[1:]
        if (False): pass
        elif (sArg=='-no-conv-or'):     glo.ConvOr = False
        elif (sArg=='-no-allow-or'):    OnlyAnd = True
        elif (sArg=='-old-map'):        WriteOldMap = True
        elif (sArg=='-write-qdimacs'):  glo.WriteQdimacs = True
        elif (sArg=='-no-rev-engr'):    glo.NoRevEngr = True
        elif (sArg=='-no-renum'):       glo.NoRenum = True
        elif (sArg=='-no-simplify'):    glo.NoSimplify = True
        elif (sArg=='-incr1'):          glo.VarIncrem = 1
        elif (sArg=='-quiet'):          PrStage = False
        elif (i == len(argv)-1):  
            qdie("Error: option '%s' doesn't exist or it needs a parameter" % sArg)
        elif (sArg=='-combine'):
            i+=1
            glo.OptCombine = int(argv[i])
        elif (sArg=='-unprenex-split'):  # Allow splitting up gates
            i+=1
            glo.UnprenexSplit = bool(int(argv[i]))
        elif (sArg=='-unprenex'):  # Allow splitting up gates
            i+=1
            glo.TryUnprenex = bool(int(argv[i]))
        elif (sArg=='-inp-var-start'):
            i+=1
            glo.InputVarStart = int(argv[i])
        elif (sArg=='-name'):
            i+=1
            PrettyName = argv[i]
        elif (sArg=='-o'):
            i+=1
            OutFilename = argv[i]
            if OutFilename.endswith(".qdimacs"):
                glo.WriteQdimacs = True
        else: qdie("Error: unrecognized option '%s'" % sArg)
        i += 1

        
    def TransmogrifyProb(p):
        p.LastGateVar = max(p.RepTable.keys())
        def KillDups():
            if len(p.QuantPfx) > 1:
                return
            again = 1
            while again:
                again = 0
                InvRepTable = dict([(v,k) for (k,v) in sorted(p.RepTable.items())])
                xlat = {}
                for (RepVar, expr) in p.RepTable.items():
                    xlat[RepVar] = InvRepTable[expr]
                #stop()
                for (old, new) in xlat.items():
                    if old != new:
                        again = 1
                        #print "%i = %i : " % (old, new),  p.RepTable[new]
                        del p.RepTable[old]
                for (RepVar, expr) in p.RepTable.items():
                    p.RepTable[RepVar] = tuple([(xlat.get(x,x) if x > 0 else -xlat.get(-x,-x))
                                                for x in expr])
                del xlat

        #FindVarLevs(p)

        p.Subsumed = set()
        p.MapNewToOld = dict([(x,x) for x in range(0, p.LastGateVar+1)])
        #if not glo.NoRenum: 
        #    RenumExprs(p)
        FindQtypes(p)
        FindTrivLits(p)
        again = True
        # if (not glo.NoSimplify):
        #     FindVarLevs(p)
        #     FindParents(p)
        #     for var in range(1, p.NumVars):
                
        while (again and (not glo.NoSimplify)):
            again = False
            KillDups()
            p.PresLits = {}  # Lits that are present in the subformula
            StartNumGates = len(p.RepTable)
            p.PossibleJunk = set()
            p.NumCombined = 0
            FindGateOccur(p)
            CombineSimplify(p, p.OutGlit, set())
            if (p.NumCombined > 0): again = True
            if p.RepTable[p.OutGlit] == "true":
                glo.NoRevEngr = True
                stderr.write("**Problem is trivially %s.**\n" % str(p.RepTable[p.OutGlit]))
                return TransmogrifyProb(read_qdimacs_file(LineReader(TrivTrueFile)))
            if p.RepTable[p.OutGlit] == "false":
                glo.NoRevEngr = True
                stderr.write("**Problem is trivially %s.**\n" % str(p.RepTable[p.OutGlit]))
                return TransmogrifyProb(read_qdimacs_file(LineReader(TrivFalseFile)))
            while (len(p.RepTable[p.OutGlit][1:]) == 1):
                #sys.stderr.write("\n" + "Reached untested code.  Verify it before proceeding.\n")
                #pdb.set_trace()
                NewOut = p.RepTable[p.OutGlit][1:][0]
                if (not p.IsGateLit(NewOut)):
                    qdie("Problem is trivial.\nUnimplemented!\n")
                if (NewOut < 0):
                    assert(NewOut not in p.QuantPfx)
                    NewOut = -NewOut
                    p.RepTable[NewOut] = NegateExpr(p.RepTable[NewOut])
                p.QuantPfx[NewOut] = p.QuantPfx[p.OutGlit] + p.QuantPfx.get(NewOut, [])
                del p.QuantPfx[p.OutGlit]
                del p.RepTable[p.OutGlit]
                p.OutGlit = NewOut
            TrivKeys = [k for (k,v) in p.RepTable.items() if v in ("true", "false")]
            if (p.OutGlit in TrivKeys):
                die("Problem is trivially " + str(p.RepTable[p.OutGlit]))
            TrivKeys = set(TrivKeys)
            for x in TrivKeys:
                del p.RepTable[x]
                if (x in p.QuantPfx):
                    del p.QuantPfx[x]
            for (key, expr) in p.RepTable.items():
                assert(len(set([abs(x) for x in expr[1:]]) & TrivKeys)==0)
            (Subsumed, Useless) = TrimJunk(p)
            p.Subsumed |= Subsumed
            assert(p.OutGlit > 0)
            QuitEarly = (elapsed_time() > 3)
            MonoSubst = []
            if not QuitEarly:
                FindTrivLits(p)
                if len(p.TrivLits.d) > 0: again = True
                TopPresLits = p.PresLits[p.OutGlit]
                if 1:
                    MonoSubst = []
                    for lit in range(0, p.NumVars):
                        if (lit in TopPresLits) == (-lit in TopPresLits):
                            continue
                        assert(not p.IsGateLit(lit))
                        if   (lit not in TopPresLits): polarity = False
                        elif (-lit not in TopPresLits): polarity = True
                        else: assert(0)
                        if   p.QtypeFromVar[lit] == 'e': p.TrivLits[lit] = polarity
                        elif p.QtypeFromVar[lit] == 'a': p.TrivLits[lit] = (not polarity)
                        else: assert(0)
                        MonoSubst.append(lit * [-1,1][p.TrivLits[lit]])
                        again = True
                    #print "Monotone: ", MonoSubst
            if (PrStage): sys.stderr.write("Done CombineSimplify (%2i combined, %2i gates elim, %2i mono lits elim), %.4f sec\n" % (p.NumCombined, StartNumGates - len(p.RepTable), len(MonoSubst), elapsed_time()))
            del p.NumCombined
            del p.PossibleJunk
            del p.PresLits
            if QuitEarly:
                stderr.write("Quit simplifying early.\n")
                again = False
        del p.TrivLits



        FindVarLevs(p)
        p.NumPushed = -1
        Unprenex(p)
        if (PrStage): 
            if p.NumPushed == -1:
                sys.stderr.write("Skipped unprenexing ('-unprenex' option not specified).\n")
            else:
                sys.stderr.write("Done unprenexing (%i), %.4f sec\n" % (p.NumPushed, elapsed_time()))
        if not glo.NoRenum: 
            RenumExprs(p)
            if (PrStage): sys.stderr.write("Done renumbering, %.4f sec\n" % elapsed_time())

        Negd = set()
        if OnlyAnd:
            if p.RepTable[p.OutGlit][0] == "or":
                p.OutGlit = -(p.OutGlit)
            for (RepVar, Expr) in p.RepTable.items(): 
                if Expr[0] == 'or': 
                    Negd.add(RepVar)
                    p.MapNewToOld[RepVar] = -p.MapNewToOld[RepVar]

        if glo.ConvOr and len(p.QuantPfx.keys())==1 and (not OnlyAnd):
            FindParents(p)
            for (RepVar, Expr) in sorted(p.RepTable.items(), key=(lambda (k,v): -p.LevByVar[k])): 
                for arg in Expr[1:]:
                    if len(p.VarPars[abs(arg)])==1 and ((arg < 0) ^ (RepVar in Negd)):
                        assert(arg not in p.QuantPfx)
                        assert(-arg not in p.QuantPfx)
                        Negd.add(abs(arg))
                        p.MapNewToOld[RepVar] = -p.MapNewToOld[RepVar]

        if Negd != set():
            for qgate in p.QuantPfx.keys():
                if (qgate in Negd):
                    sys.stderr.write("\nERROR: Tried to negate a quantified gate.")
                    if glo.WriteQdimacs:
                        sys.stderr.write("\n(Converting non-prenex to QDIMACS is not implemented.)\n")
                    sys.exit(1)
                    
            for (RepVar, Expr) in p.RepTable.items():
                Expr = list(Expr)
                if RepVar in Negd:
                    if (Expr[0] == "or"): Expr[0] = "and"
                    elif (Expr[0] == "and"): Expr[0] = "or"
                    Expr[1:] = (-x for x in Expr[1:])
                Expr[1:] = ((x if (abs(x) not in Negd) else -x) for x in Expr[1:])
                p.RepTable[RepVar] = tuple(Expr)
        if (PrStage): sys.stderr.write("Done processing, %.4f sec\n" % elapsed_time())
        

        return p

    def PrintProb(p, outf):
        NumOrigVars = p.NumOrigVars
        if (not glo.NoRenum):
            factor = NumOrigVars / (p.NumVars/2)
            if PrStage:
                sys.stderr.write(("Input variables reduced from %5i to %4i in " 
                    + PrettyName + ". (%ix)\n") % \
                    (NumOrigVars, p.NumVars/2, factor))

        outf.write("CktQBF\n")
        outf.write("LastInputVar %i\n" % p.NumVars)
        #outf.write("FirstGateVar %i\n" % p.FirstExpr)
        outf.write("LastGateVar %i\n"  % max(p.RepTable.keys()))
        outf.write("OutputGateLit %i\n"   % p.OutGlit)
        outf.write("PreprocTimeMilli %i\n"   % int(elapsed_time() * 1000))
        
        if WriteOldMap:
            outf.write('\n' + textwrap.fill(
                "#MapNewToOld (NewVar, OldVar) : (" + 
                ", ".join("(%i,%i)" % pair for pair in sorted(p.MapNewToOld.items()))
                + ")", 80, subsequent_indent='#   ') + "\n\n")
            outf.write('\n' + textwrap.fill(
                "#Subsumed : (" + str(sorted(p.Subsumed)) + ")", 80, subsequent_indent='#   ') + "\n\n")

        ixQb = 1
        outf.write("\n")

        for (qgate, blks) in p.QuantPfx.items():
            NewBlks = []
            for qb in blks:
                if len(qb.qvars)==0: continue  # no variables quantified
                if len(NewBlks) > 0 and NewBlks[-1].qtype == qb.qtype:
                    NewBlks[-1].qvars += qb.qvars
                else:
                    NewBlks.append(qb)
            p.QuantPfx[qgate] = NewBlks

        for (qgate, blks) in SortedQuantPfx(p):
            if blks == []:
                continue
            outf.write("<q gate=%i>\n" % qgate)
            for quant in blks:
                if len(quant.qvars)==0: continue  # no variables quantified
                outf.write("#Qb%i\n" % ixQb)
                ixQb += 1
                outf.write("      %s %s\n" % 
                    (quant.qtype.lower(), " ".join([str(lit) for lit in quant.qvars])))
            outf.write("</q>\n")
        outf.write("\n")


        CurLev = 0
        FindParents(p)
        for (RepVar, Expr) in reversed(sorted(p.RepTable.items())):  #, key=(lambda x: p.MapNewToOld[x])
            NewLev = p.LevByVar[RepVar]
            if (CurLev != NewLev):
                CurLev = NewLev
                outf.write("#NESTING LEVEL %i\n" % CurLev)
            OccurMsg = ""
            if 0:
                OccOnce = OccursOnce(p, RepVar)
                OccurMsg = (" # occurs once" if OccOnce else " # occurs multiple times")
            outf.write("%i = %s(%s)%s\n" % 
                (RepVar, Expr[0], ", ".join([str(lit) for lit in sorted(Expr[1:], key=abs)]), OccurMsg))
        if (PrStage): sys.stderr.write("Done writing output file, %.4f sec\n" % elapsed_time())

    
    def PrintQdimacs(p, outf):
        #ChkUnquantVars(p)
        OutIsConj = (p.OutGlit > 0)
        assert(OnlyAnd)

        TopClauseGates = set()
        if OutIsConj:
            TopClauseGates = set([abs(x) for x in p.RepTable[p.OutGlit][1:] if (-x > 0 and p.IsGateLit(x))])
            for (GateVar, expr) in p.RepTable.items():
                if (GateVar == p.OutGlit): continue
                TopClauseGates -= set([abs(x) for x in expr[1:]])
                    
        CalcNumCl = 0
        for (GateVar, expr) in p.RepTable.items():
            if (GateVar == p.OutGlit): continue
            if OutIsConj and (GateVar in TopClauseGates):
                CalcNumCl += 1
            else:
                CalcNumCl += 1 + len(expr[1:])
        if OutIsConj: 
            TopLits = set([x for x in p.RepTable[p.OutGlit][1:] if abs(x) not in TopClauseGates])
        else:
            TopLits = set([p.OutGlit])
        CalcNumCl += len(TopLits)

        if WriteOldMap:
            outf.write('\n' + textwrap.fill(
                "c MapNewToOld (NewVar, OldVar) : (" + 
                ", ".join("(%i,%i)" % pair for pair in sorted(p.MapNewToOld.items()))
                + ")", 80, subsequent_indent='c   ') + "\n\n")
        outf.write("p cnf %i %i\n" % (p.LastGateVar, CalcNumCl))
        
        FlatBlks = []
        for (qgate, blks) in SortedQuantPfx(p):
            for quant in blks:
                if len(quant.qvars)==0: continue  # no variables quantified
                FlatBlks.append(quant)

        for quant in FlatBlks:
            outf.write(quant.qtype.lower() + " " + 
                (" ".join([str(lit).lower() for lit in quant.qvars])))
            if (quant is FlatBlks[-1]):
                RepKeys = set(p.RepTable.keys()) - TopClauseGates
                if (OutIsConj): RepKeys.remove(abs(p.OutGlit))
                if quant.qtype.lower() == 'e':
                    outf.write(" " + " ".join(str(x) for x in sorted(RepKeys)))
                else:
                    outf.write(" 0\n")
                    outf.write("e " + " ".join(str(x) for x in sorted(RepKeys)))
            outf.write(" 0\n")

        ActNumCl = 0
        
        def sxlat(x):
            return str(x)
            #if x>0: return str(p.MapNewToOld[x])
            #else:   return str(-p.MapNewToOld[abs(x)])   

        for lit in TopLits:
            outf.write(sxlat(lit) + " 0\n")
            ActNumCl += 1

        for (GateVar, conj) in reversed(sorted(p.RepTable.items())):
            # (v <==> (x1 & x2 & x3)) expands to
            # (~x1 | ~x2 | ~x3 | v)  &  (~v | x1)  &  (~v | x2)  &  (~v | x3)
            assert(conj[0] == 'and')
            args = conj[1:]
            if OutIsConj and (GateVar == p.OutGlit): continue
            if OutIsConj and (GateVar in TopClauseGates):
                outf.write(" ".join([sxlat(-x) for x in args]) + " 0\n")
                ActNumCl += 1
            else:
                outf.write(sxlat(GateVar) + " " + " ".join([sxlat(-x) for x in args]) + " 0\n")
                ActNumCl += 1
                for lit in args:
                    outf.write(sxlat(-GateVar) + " " + sxlat(lit) + " 0\n")
                    ActNumCl += 1

        assert(ActNumCl == CalcNumCl)


    def doit():
        prob = read_input_file(filename)
        prob = TransmogrifyProb(prob)
        if (OutFilename != ""): 
            outf = open(OutFilename, 'w')
        else:
            outf = sys.stdout
        if glo.WriteQdimacs:
            PrintQdimacs(prob, outf)
        else:
            PrintProb(prob, outf)

    if 1: doit()
    else:
        try: doit()
        except KeyboardInterrupt:
            print "BREAK!"
            sys.exit(3)

main()
