/******************************************************************************
* GhostQ: A QBF solver with ghost variables.
* Version 0.12 (Apr 2010)
* Author: Will Klieber (Carnegie Mellon University)
* Note: Sorry for the current untidy nature of this file; I still need to
* clean up all the dead code and reorganize the code for clarity.
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <algorithm>
#include <map>
#include <set>
#include <vector>
#include <list>
#include <queue>
using namespace std;


/*****************************************************************************
* Utility Macros.    }}}{{{1
*****************************************************************************/

#define auto(var, init) typeof(init) var = (init)

#define foreach(var, container) \
  for (auto(var, (container).begin());  var != (container).end();  ++var)

#define printferr(fmt,args...) \
    ({fflush(stdout); fprintf(stderr, fmt, ##args);})

#define talloc(type, n) ((type*) calloc(n, sizeof(type)))


/**************************************************************
* TODO:                                            }}}{{{1    *
* - Remove subsumed game-states.
* - Periodically clean up game-state database.
**************************************************************/


/*****************************************************************************
* Vim: highlight func name in func definition, folding.     }}}{{{1    
******************************************************************************
* syn match cppFuncDef "\~\?\zs\h\w*\ze([^)]*\()\s*\(const\)\?\)\?{ *$"
* hi def link cppFuncDef Todo
* set fdm=marker
*****************************************************************************/


/**************************************************************
* Global constants.    }}}{{{1
**************************************************************/

const short NO_WINR = 0x7f;
const int DEFN_TAG = 1;
const int ORIG_TAG = 1;
const int LRN_TAG = 2;
const int NOT_DONE = 3;
#define ALL_PLY -2
#define ZERO_DECLEV 0
#define ZERO_LIT 0


/* Global command-line and compile-time options. */
#define UsePhant 1  // Use ghost (phantom) literals instead of controllers
const int AltMeth1 = 0;  // alternate method for learning (don't use this)
int ExDbg = false;
int GloDbgLev = 10;
int GloRandSeed;
int GloSanChk = 1;
int GloTimeOut = 0;
long long StartTime;
int RestartCycle = 128;
int NoRestart = false;
int OldCubeWay = false;
int UseLrnType4 = false;
int FindFullCtlrs = !UsePhant;
int FindSemiCtlrs = false;
int UseSemiCtlrs = true;
int UseMonotone = 0;
int NoUnivGhost = 0;
FILE* PrfLog = NULL;

const int LINE_LEN = 90;


/*****************************************************************************
* Utility Functions.    }}}{{{1
*****************************************************************************/

void raw_nop() {}
void (*volatile volnop)() = &raw_nop;  // For debugging; optimizer won't remove.

void rawstop() {}    // Place a breakpoint here with gdb.
void (*volatile stop)() = &rawstop;  // For debugging; optimizer won't remove.

inline int StrEqc(const char* s1, const char* s2) {
    // Precond: At least one of the strings is not null.
    if ((s1 == NULL) ^ (s2 == NULL)) {return 0;}
    return (strcasecmp(s1, s2) == 0);
}

char* StrFmt(const char* fmt, ...) {
    va_list ap;
    char* pOut = NULL;
    va_start(ap, fmt);
    vasprintf(&pOut, fmt, ap);
    va_end(ap);
    return pOut;
}

int die_f(const char* fmt, ...) {
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    abort();
    return 0;
}

int die(const char* msg="") {
    fprintf(stderr, "%s", msg);
    abort();
    return 0;
}

long long GetTimeMicro() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return ((long long) tv.tv_sec) * 1000000 + (tv.tv_usec);
}

long long GetUserTimeMicro() {
    struct rusage usage;
    getrusage(RUSAGE_SELF, &usage);
    struct timeval tv = usage.ru_utime;
    return ((long long) tv.tv_sec) * 1000000 + (tv.tv_usec);
}

int RandSeedFromTime() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    int ret = tv.tv_sec * 10000 + (tv.tv_usec / 100);
    ret = (ret & (((uint)(~0)) >> 1));  // Set the MSB to zero.
    return ret;
}

void DumpIntSet(set<int>* pSet, const char* pEnd) {
    set<int>::iterator iter = pSet->begin();
    printf("[ ");
    for (; iter != pSet->end(); ++iter) {
        printf("%i ", *iter);
    }
    printf("]%s", pEnd);
}

bool gt_f(int const& x, int const& y) {
    return (x > y);
}

bool lt_f(int const& x, int const& y) {
    return (x < y);
}


/*****************************************************************************
* BitSet.    }}}{{{1
* Unlike the STL bitset, this doesn't require that the number
* of bits be a compile-time constant.
* TODO: Optimize this so that each bit only consumes 1 bit of memory. 
*****************************************************************************/
struct BitSetT {
    char* arr;

    void alloc(int n) {arr = talloc(char, n);}
    void dealloc() {free(arr); arr=NULL;}
    char operator[](uint ix) {return arr[ix];}
    void set(uint ix, bool val) {arr[ix] = val;}
    //void clear(int n) {memset(arr, 0, n);}

    #define FPROTO_BitSetT
    #include "ghostq.proto"
    #undef FPROTO_BitSetT

};


/*****************************************************************************
* Fixed-Size Array with Bounds Info.    }}}{{{1
*****************************************************************************/
template <typename T>
struct ap {  // "Array Pointer"
    T* arr;
    int n;

    /* Constructor */
    ap() : arr(NULL), n(0) { }; 
    ap(int* _arr, int _n) : arr(_arr), n(_n) { }; 

    T& operator[](unsigned ix) {return arr[ix];}
    T& last() {return arr[n-1];}
    void alloc(int size) {
        n = size;
        arr = talloc(T, size);
    }

    T* begin() {return &arr[0];}
    T* end() {return &arr[n];}

    #define FPROTO_ap
    #include "ghostq.proto"
    #undef FPROTO_ap

};

template <typename T>
int ap<T>::indexof(T x) {
    for (int i=0; i < n; i++) {
        if (arr[i] == x) {
            return i;
        }
    }
    return -1;
}

template <typename T>
void ap<T>::sort(int (*cmp)(const T*, const T*)) {
    qsort(this->arr, this->n, sizeof(T), (int (*)(const void*, const void*))(cmp));
    // Note: Use the STL sort algorithm instead; it's faster because it's inlined.
}

template <typename T>
void ap<T>::remove_adj_dups() {
    T* pNewEnd = unique(begin(), end());
    n = pNewEnd - begin();
}

template <typename T>
int DeepEq(ap<T> a1, ap<T> a2) {
    if (a1.n != a2.n) return false;
    for (int i=0; i < a1.n; i++) {
        if (a1.arr[i] != a2.arr[i]) return false;
    }
    return true;
}

/*****************************************************************************
* Growable Array.    }}}{{{1
* We use this instead of the STL vector class in order to manually
* optimize memory usage.
*****************************************************************************/
template <typename T>
struct garr {
    T* arr;
    int n;
    int MaxSize;

    #define FPROTO_garr
    #include "ghostq.proto"
    #undef FPROTO_garr

    /* Constructor */
    garr() : arr(NULL), n(0), MaxSize(0) { }; 
    garr(int size, int prealloc) : n(size), MaxSize(prealloc) {
        arr = talloc(T, prealloc);
    }; 
    
    T& operator[](unsigned ix) {return arr[ix];}

    T& last() {return arr[n-1];}

    void init(int size=0, int prealloc=8)  {
        n = size;
        MaxSize = prealloc = max(prealloc, size);
        if (MaxSize == 0) {arr = NULL;}
        else {arr = (T*) malloc(prealloc * sizeof(T));}
    };

    void resize(int NewSize) {
        if (NewSize > MaxSize) {realloc_grow(NewSize);}
        n = NewSize;
    };

    T* begin() {return &arr[0];}
    T* end() {return &arr[n];}

    void append(T item) {resize(n+1); arr[n-1] = item;};

};


template <typename T>
int garr<T>::realloc_grow(int DemandSize) {
    int NewSize;
    if (MaxSize < 65536) {
        if (MaxSize < 8) {
            MaxSize = 8;
        }
        NewSize = MaxSize * 2;
    } else {
        NewSize = (MaxSize + (MaxSize / 4)) & ~4096;
    }
    if (NewSize < DemandSize) {
        NewSize = DemandSize;
    }
    MaxSize = NewSize;
    arr = (T*) realloc(arr, NewSize*sizeof(T));
    return 0;
};

template <typename T>
void garr<T>::realloc_shrink(int NewSize=-1) {
    if (NewSize == -1) {
        NewSize = n;
    }
    MaxSize = NewSize;
    arr = (T*) realloc(arr, NewSize*sizeof(T));
};

template <typename T>
T* garr<T>::AppendBlank() {
    resize(n+1);
    arr[n-1] = T();
    return &arr[n-1];
}

template <typename T>
void garr<T>::extend(garr<T> items) {
    int base = n;
    int m = items.n;
    resize(n + m);
    for (int i=0; i < m; i++) {
        arr[base+i] = items[i];
    }
};

template <typename T>
void garr<T>::dealloc() {
    n = MaxSize = 0;
    free(arr);
    arr = NULL;
}

template <typename T>
garr<T> garr<T>::dup() {
    garr<T> ret;
    ret.init(n, n);
    for (int i=0; i < n; i++) {
        ret[i] = arr[i];
    };
    return ret;
}

template <typename T>
int garr<T>::indexof(T x, int start=0) {
    for (int i=start; i < n; i++) {
        if (arr[i] == x) {
            return i;
        }
    }
    return -1;
}

template <typename T>
T garr<T>::RemoveAndFill(int ix) {
    T ret = arr[ix];
    arr[ix] = arr[n-1];
    n--;
    return ret;
}

template <typename T>
T garr<T>::PopLast() {
    assert(n > 0);
    T ret = arr[n-1];
    n--;
    return ret;
}

template <typename T>
void garr<T>::RemoveVal_AndFill(T val) {
    for (int i=0; i < n; i++) {
        if (arr[i] == val) {
            RemoveAndFill(i);
            return;
        }
    }
    assert(false);
}

template <typename T>
garr<T> garr<T>::dup_shuffle() {
    garr<T> tmp = this->dup();
    garr<T> ret;
    ret.init(n);
    for (int i=0; i < n; i++) {
        ret[i] = tmp.RemoveAndFill(rand() % (n-i));
    }
    tmp.dealloc();
    return ret;
}

template <typename T>
garr<T> garr<T>::dup_apply(T (*fn)(T)) {
    garr<T> ret;
    ret.init(n, n);
    for (int i=0; i < n; i++) {
        ret[i] = fn(arr[i]);
    };
    return ret;
}

template <typename T>
void garr<T>::apply(T (*fn)(T)) {
    for (int i=0; i < n; i++) {
        arr[i] = fn(arr[i]);
    };
}

template <typename T>
void garr<T>::sort(int (*cmp)(const T*, const T*)) {
    qsort(this->arr, this->n, sizeof(T), (int (*)(const void*, const void*))(cmp));
    // Note: Use the STL sort algorithm instead; it's faster because it's inlined.
}


template <typename T>
void garr<T>::remove_adj_dups() {
    T* pNewEnd = unique(begin(), end());
    n = pNewEnd - begin();
}

template <typename T, typename C>
void init_garr_from_set(garr<T>* pGarr, set<T,C>& OrigSet) {
    pGarr->resize(OrigSet.size());
    pGarr->n = 0;
    typename set<T>::iterator iter = OrigSet.begin();
    for (; iter != OrigSet.end(); ++iter) {
        pGarr->append(*iter);
    }
}

template <typename T, typename C>
void init_garr_from_list(garr<T>* pGarr, list<T,C>& OrigSet) {
    pGarr->resize(OrigSet.size());
    pGarr->n = 0;
    typename list<T>::iterator iter = OrigSet.begin();
    for (; iter != OrigSet.end(); ++iter) {
        pGarr->append(*iter);
    }
}

template <typename T>
set<T> garr<T>::to_set() {
    set<T> ret(this->arr, this->arr+n);
    return ret;
}

typedef garr<int> garri;


template <typename T>
garr<T> dup_as_garr(ap<T>& cur) {
    garr<T> ret;
    ret.init(cur.n, cur.n);
    for (int i=0; i < cur.n; i++) {
        ret[i] = cur.arr[i];
    };
    return ret;
}

template <typename T>
ap<T> dup_as_ap(garr<T>& cur) {
    ap<T> ret;
    ret.arr = talloc(T, cur.n);
    ret.n = cur.n;
    for (int i=0; i < cur.n; i++) {
        ret[i] = cur.arr[i];
    };
    return ret;
}


/*****************************************************************************
* Encoding of Literals.    }}}{{{1
*****************************************************************************/

/* Literals are encoded as numbers.  The least-significant bit (LSB)
 * is 0 if the literal is in its positive form, and
 * is 1 if the literal is negated.
 * Currently, we require that the input file uses only even numbers;
 * this makes debugging easier.
 */
int encv(int x) {
    if (x >= 0) {return x;}
    else {return (abs(x) + 1);}
}
int decv(int x) {
    if ((x & 1) == 1) {return -(x & ~1);}
    else {return x;}
}
inline int negv(int x) {return (x ^ 1);}
inline int absv(int x) {return (x & ~1);}
inline int signv(int x) {return (x & 1);}
inline int IsNegLit(int x) {return (x & 1);}
inline int IsPosLit(int x) {return !IsNegLit(x);}

template <typename T>
void DumpLitsEnc(T& arr, const char* newl="\n") {
    const char* sep = "";
    printf("[");
    int n = 0;
    foreach (it, arr) {
        n += printf("%s%i", sep, decv(*it));
        if (n > 85) {
            if (n < 95) printf("\n");
            n = 0;
        }
        sep = " ";
    }
    printf("]%s", newl);
}


void DumpLitsEnc(garri& arr, const char* newl="\n") {
    const char* sep = "";
    printf("[");
    int n = 0;
    foreach (it, arr) {
        n += printf("%s%i", sep, decv(*it));
        if (n > 85) {
            if (n < 95) printf("\n");
            n = 0;
        }
        sep = " ";
    }
    printf("]%s", newl);
}


/*void DumpLitsEnc(ap<int>& arr, const char* newl="\n") {
    printf("[");
    for (int i=0; i < arr.n; i++) {
        printf("%i%s", decv(arr[i]), (i == arr.n-1 ? "" : " "));
    }
    printf("]%s", newl);
}*/


/*****************************************************************************
* Lexer/Parser.    }}}{{{1
*****************************************************************************/
enum {
    INT_TOK_TYPE = 1,
    IDFR_TOK_TYPE,
    OTHER_TOK_TYPE
};

#define LEX_ERR(str) {fprintf(stderr, "%s\n", str); abort();}

/* String Array Pointer */
struct sap {
    char* p;    // pointer
    int n;      // size
};


class Lexer {
    public:
    const char* filename;
    FILE* infile;
    sap CurLine;
    sap CurTok;
    int CurPos;
    int LineNum;
    int CurTokType;

    #define FPROTO_Lexer
    #include "ghostq.proto"
    #undef FPROTO_Lexer

    /* Constructor */
    Lexer(const char* filename_) : filename(filename_) {
        init();
    }

    void init() {
        CurLine = (sap){NULL, 0};
        CurTok  = (sap){NULL, 0};
        CurPos = 0;
        LineNum = 0;
        if (StrEqc(filename, "-")) {
            infile = stdin;
        } else {
            infile = fopen(filename, "r");
        }
        if (infile == 0) {
            LEX_ERR(StrFmt("File '%s' does not exist.", filename));
        }
        AdvLine();
        AdvTok();
        while (StrEqc(CurTok.p, "\n")) {AdvTok();}
    };

    void AdvLine() {
        CurLine.n = getline(&CurLine.p, (size_t*)&CurLine.n, infile);
        CurPos = 0;
        LineNum++;
    }
};


void Lexer::AdvTok() {
    int InitPos;
    CurTokType = OTHER_TOK_TYPE;
    again:
    if (CurLine.n == -1) {
        CurTok = (sap){NULL, -1};
        return;
    }
    if (CurPos >= CurLine.n) {
        AdvLine();
        goto again;
    }
    //CurTok.p = &CurLine.p[CurPos];
    InitPos = CurPos;
    switch (CurLine.p[CurPos++]) 
    {
        case '\r':  // For MS-DOS text files
            goto again;

        case ' ': case '\t': case ',':
            goto again;

        case '#':
            // Lines beginning with "#" are comment lines
            CurPos = CurLine.n - 1;
            goto again;

        case '\n': case '(': case ')': case '=':
            CurTok.n = 1;
            goto copytok;

        case '0' ... '9': case '-':
            CurTokType = INT_TOK_TYPE;
            while (true) {
              assert(CurPos <= CurLine.n);
              switch (CurLine.p[CurPos++]) {
                case '0' ... '9': continue;
                default: goto done_idfr;
              }
            }

        case 'a' ... 'z':
        case 'A' ... 'Z': case '_':
            CurTokType = IDFR_TOK_TYPE;
            while (true) {
              assert(CurPos <= CurLine.n);
              switch (CurLine.p[CurPos++]) {
                case '0' ... '9':  case 'a' ... 'z':  case 'A' ... 'Z':  case '_':
                  continue;
                default:
                  goto done_idfr;
              }
            }

        case '<': 
            CurPos++;
            goto done_idfr;

        done_idfr:
        CurPos--;
        CurTok.n = CurPos - InitPos;
        goto copytok;


        default: {
            char* pChrStr;
            unsigned char chr = CurLine.p[CurPos-1];
            if ((32 <= chr) && (chr <= 126)) {
                pChrStr = StrFmt("'%c'", chr);
            } else {
                pChrStr = StrFmt("0x%X", chr);
            }
            fprintf(stderr, "Bad character (%s) on line %i, column %i.\n",
                pChrStr, LineNum, CurPos);
            exit(1);
        }
    }
    copytok:
    CurTok.p = (char*) realloc(CurTok.p, CurTok.n+1);
    memcpy(CurTok.p, &CurLine.p[InitPos], CurTok.n);
    CurTok.p[CurTok.n] = 0;
        
}

char* Lexer::StrCurTokPos() {
    return StrFmt("line %i, col %i-%i", LineNum, CurPos - CurTok.n + 1, CurPos);
}

void Lexer::ChkTok(const char* expected) {
    if (!StrEqc(CurTok.p, expected)) {
        LEX_ERR(StrFmt("Parsing error (%s): Expected '%s'.",
            StrCurTokPos(), expected));
    }
    AdvTok();
}

void Lexer::ReadLineSep() {
    if (!StrEqc(CurTok.p, "\n")) {
        LEX_ERR(StrFmt("Parsing error (%s): Expected a line break.", StrCurTokPos()));
    }
    while (StrEqc(CurTok.p, "\n")) {
        AdvTok();
    }
}

void Lexer::SkipLine() {
    AdvLine();
    AdvTok();
    while (StrEqc(CurTok.p, "\n")) {
        AdvTok();
    }
}
    

int Lexer::ReadIntRaw() {
    if (CurTokType != INT_TOK_TYPE) {
        LEX_ERR(StrFmt("Parsing error (%s): Expected an integer.", StrCurTokPos()));
    }
    int ret = atoi(CurTok.p);
    AdvTok();
    return ret;
}

int Lexer::ReadIntEnc() {
    int x = ReadIntRaw();
    if (x % 2 != 0) {
        LEX_ERR(StrFmt("Parsing error (%s): Expected an even int, but found an odd.",
            StrCurTokPos()));
    }
    return encv(x);
}

const char* Lexer::OpNameSym(char* pNameStr) {
    if (StrEqc(pNameStr, "AND"))    {return "AND";}
    if (StrEqc(pNameStr, "OR"))     {return "OR";}
    return NULL;
}

const char* Lexer::ReadOpIdfr() {
    if (CurTokType != IDFR_TOK_TYPE) {
        LEX_ERR(StrFmt("Parsing error (%s): Expected an identifier.", StrCurTokPos()));
    }
    const char* ret = OpNameSym(CurTok.p);
    if (ret == NULL) {
        LEX_ERR(StrFmt("Parsing error (%s): Unrecognized operator name '%s'.",
            StrCurTokPos(), CurTok.p));
    }
    AdvTok();
    return ret;
}

garri Lexer::ReadArgs() {
    garri ret;
    ret.init(0, 8);
    ChkTok("(");
    while (CurTokType == INT_TOK_TYPE) {
        ret.append(ReadIntEnc());
    }
    if (StrEqc(CurTok.p, ")")) {
        AdvTok();
    } else {
        LEX_ERR(StrFmt("Parsing error (%s): Expected an integer or a ')'.",
            StrCurTokPos()));
    }
    ret.realloc_shrink();
    return ret;
}

int Lexer::ReadIntParamEnc(const char* pParamName) {
    ChkTok(pParamName);
    int ret = ReadIntEnc();
    ReadLineSep();
    return ret;
}




/*****************************************************************************
* Decision Level stuff, and other stuff.    }}}{{{1
*****************************************************************************/
/* Note: QbFromLit[x] is 16 bits long. */
#define MAX_QB ((1 << 15) - 1)
#define MAX_VARS ((1 << 24) - 1)

#define NO_DECLEV ((uint)(~0) >> 1)


typedef int FmlaT;
typedef int LitT;


typedef map<int, garr<struct GsT*> >  QbDepListT;  // Maps quantifier index to list of dependent game states. 
typedef QbDepListT::iterator QbDepListIterT;


/*****************************************************************************
* Headers.    }}}{{{1
*****************************************************************************/
struct QuantBlkT {
    int     qtype;      // 'A' or 'E'.
    int     ix;         // index number
    int     bat;        // gate of which this is a prefix
    garri   lits;
    QuantBlkT() : qtype(0), lits() { }; 
};

struct QuantPfxT {
    garr<QuantBlkT> blks;
};

struct ExprT {
    const char* sOp;
    garri args;
    struct GsT* pLongLs;
};

struct GateDefT {
    int IsAnd;
    ap<LitT> args;
};


/*****************************************************************************
* ProbInstT Header.    }}}{{{1
*****************************************************************************/
struct ProbInst {
    int LastInpLit;
    int FirstGateVar;
    int LastGlit;
    int LastPlit;
    int OutGlit;
    int NumQuantVars;
    garr<QuantBlkT> QuantBlk;   // indexed by qb index number
    garr<GsT*> *LitHave;   
    garr<GsT*> clauses;
    int NumDoneClauses;
    garr<struct GsT*> PhantLitSets;
    garri *ParentLit;
    garri *ImmedForce;  // y is in ImmedForce[x] if (x ==> y).
    int PreprocTimeMilli;

    #define FPROTO_ProbInst
    #include "ghostq.proto"
    #undef FPROTO_ProbInst

    int IsValidInpVar(int x)    {return (2 <= x  &&  x <= LastInpLit);}
    int IsValidExprIx(int x) {return (2 <= x  &&  x <= LastGlit);}
    int IsValidQbIx(int x)   {return (1 <= x  &&  x < QuantBlk.n);}

};


struct LvCacheEntry {
    int IsAlive;
    long long timestamp;
};


/*****************************************************************************
* Global vars and utility funcs.    }}}{{{1
*****************************************************************************/

/* Global Variables. */
struct ProbInst* pGloProb;
struct AsgnT* pGloAsgn;
int GloStep = 1;
int* pNumClauses;
int GloIsPostMortem = false;
int GloNumLrnRejects = 0;
int GloNumLrnWins = 0;
int NumSfAborts = 0;
int NumSfUses = 0;
int NumSfRevives = 0;
bool NeedNewl = false;
int NumImpBlocked = 0;
int NumExecAttempts = 0;
int NumProps = 0;
int NumCtlrExecAttempts = 0;
int NumImpUnripe = 0;
int NumImpRevives = 0;
int NumStepBt = 1;
int NumBigBt = 1;
int NumDecs = 0;
int NumProp_OneStep=0;
int NumProp_ManyStep_Base=0;
int NumProp_ManyStep_Lrn=0;

int LastPickGate = 0;

#define AND_GATE 'c'
#define OR_GATE 'd'
int LastPhant[2];
int PhantOffset[2];
char *GateType;
short *QbFromLit;          // maps a lit to the index of its quant block.
short *MinQbFromGlit;
int *LitToBat;
//int *GlitToBat;
int *QbToBat;
GsT** GlitToGs;
GateDefT *GateDef;
QuantPfxT* QuantPfx;        // indexed by gate number

GsT* AbsGlitToGs(int glit) {
    assert(GlitToGs[glit]==0 || GlitToGs[negv(glit)]==0);
    return (GsT*)((long)(GlitToGs[glit]) | (long)(GlitToGs[negv(glit)]));
}

GsT* NegGlitToGs(int glit) {
    return GlitToGs[negv(glit)];
}

void FixNeedNewl() {
    if (NeedNewl) {printf("\n"); NeedNewl = false;}
}

int IsInputLit(int lit) {
    return (lit <= pGloProb->LastInpLit);
}

int IsGateLit(int lit) {
    return !(lit <= pGloProb->LastInpLit);
}

int IsPhant(int lit) {
    return lit > pGloProb->LastGlit;
}

int IsAndGate(int lit) {
    assert(IsGateLit(lit));
    return GateDef[lit].IsAnd;
}

int GloQtype(int ixQb) {
    return pGloProb->QuantBlk[ixQb].qtype;
}

int QtypeFromLit(int lit) {
    if (!IsInputLit(lit)) {return '-';}
    return GloQtype(QbFromLit[lit]);
}

int QtypeNumFromLit(int lit) {
    switch (QtypeFromLit(lit)) {
        case 'A':  return 0;
        case 'E':  return 1;
        default:   assert(false);
    }
}

//int LitToBat(int lit) {
//    return pGloProb->QuantBlk[QbFromLit[lit]].bat;
//}

/*int QuantBlkT::GetIx() {
    int ret = ix;
    assert(ret == this - pGloProb->QuantBlk.begin());
    return ret;
}*/


struct LessGsById {
    bool operator() (const GsT* const& pLs1, const GsT* const& pLs2) const;
};
//typedef set<struct GsT*,LessGsById> GwSetT;
typedef set<pair<int, struct GsT*> > GwSetT;
//typedef list<struct GsT*> GwSetT;

#define LRN_TYPE_1 1
#define LRN_TYPE_2 2
#define LRN_TYPE_3 3
#define LRN_TYPE_4 4

/*****************************************************************************
* Header for Game-State Entry data structure.    }}}{{{1
*****************************************************************************/
struct GsT {
    int id;
    char OrigOrLrn;
    char FreeFmla;
    short LrnType;
    int IsBattle;
    int GateLit;
    int ctlr;
    //int IsInit;
    set<LitT> ctlrs;
    garri* pChoices;


    int WatchedLit[2];
    int WatchedRes;
    int WatchRawAE[4];
    GwSetT::iterator itWatchRes;
    GwSetT::iterator itWatchRawAE[4];

    ap<int> ReqLits;    // Triggering Lits (corresponding to $L^{now}$ in the paper)
    ap<int> TrigAE[2];  // Triggering universal lits (TrigAE[0]) and
                        // triggering existential lits (TrigAE[1]), sorted.
    ap<int> TrigG;      // Triggering gate lits, sorted.
    ap<int> ImpLits;    // Reserved Lits (corresponding to $L^{fut}$ in the paper)

    int& WatchAE(int player, int x) {
        return WatchRawAE[player*2 + x];
    }

    inline int FirstGateWatch() {
        if (FreeFmla == NO_WINR) return WatchAE(0,0);
        //assert(FreeFmla==0 || FreeFmla==1);
        return WatchAE(negv(FreeFmla), 0);
    }

    #define FPROTO_GsT
    #include "ghostq.proto"
    #undef FPROTO_GsT

};

void GsT::init() {
    ctlrs = set<LitT>();
    //IsInit = true;
}

void GsT::dealloc() {
}

int set_count(GwSetT& s, GsT* item) {return s.count(pair<int, struct GsT*>(item->id, item));}  // for debugging

bool LessGsById::operator() (const GsT* const& pLs1, const GsT* const& pLs2) const {
    //assert(pLs1->id != 0 && pLs2->id != 0);
    if (pLs1->id == pLs2->id) {return pLs1 < pLs2;}
    return (pLs1->id < pLs2->id);
}

bool FnGreaterGsById(const GsT* const& pLs1, const GsT* const& pLs2) {
    return (pLs1->id > pLs2->id);
}

bool FnLessGsById(const GsT* const& pLs1, const GsT* const& pLs2) {
    return (pLs1->id < pLs2->id);
}

int CmpGsById(GsT* const* p1, GsT* const* p2) {
    return ((*p1)->id - (*p2)->id);
}



int CmpLsByPtr(GsT* const* ppLs1, GsT* const* ppLs2) {
    GsT* pLs1 = *ppLs1;
    GsT* pLs2 = *ppLs2;
    return (pLs1 - pLs2);
}



    
//GsT EmptyLs;



/*****************************************************************************
* ChlitListT: List of chosen (non-forced) literals.    }}}{{{1
*****************************************************************************/
class ChlitListT {
    public:
    garri flat;
    #define FPROTO_ChlitListT
    #include "ghostq.proto"
    #undef FPROTO_ChlitListT
};

void ChlitListT::append(int lit) {
    flat.append(lit);
}

int ChlitListT::pop() {
    return flat.PopLast();
}

int ChlitListT::last() {
    return flat.last();
}

void ChlitListT::dump() {
    int CurQb = 0;
    printf("[");
    for (int i=1; i < flat.n; i++) {
        int lit = flat[i];
        if (i != 1) {printf("%c ", (QbFromLit[lit]==CurQb ? ',' : ';'));}
        CurQb = QbFromLit[lit];
        printf("%i", decv(lit));
    }
    printf("]\n");
}

/*****************************************************************************
* Stuff  }}}{{{1
*****************************************************************************/


struct VsidsT {
    uint x[2];
};


#define UNIT_PROP_ANTECED 1
#define MONO_ANTECED 2
#define CONTRADICT_ANTECED 3

struct AntecedT {
    GsT* pDepGs;
    int FinalCtlr;
    int AltType;
    int AltLit;

    int GetType() {
        if (pDepGs != NULL) return UNIT_PROP_ANTECED;
        else return AltType;
    }
    int IsUnitProp() {return (pDepGs != NULL);}

    /* Constructor */
    AntecedT() : pDepGs(0), FinalCtlr(0), AltType(0), AltLit(0) {};
    AntecedT(GsT* _gs, int _ctlr, int _alt=0, int _mlit=0) : 
            pDepGs(_gs), FinalCtlr(_ctlr), AltType(_alt), AltLit(_mlit) {};

    bool operator==(AntecedT x) {
        return (memcmp(this, &x, sizeof(AntecedT)) == 0);
    }
};

AntecedT EmptyAnteced;


bool AntecedByGreaterGsId(AntecedT const& a1, AntecedT const& a2) {
    if (a1.pDepGs->id <  a2.pDepGs->id) return true;
    if (a1.pDepGs->id != a2.pDepGs->id) return false;
    if (a1.FinalCtlr < a2.FinalCtlr) return true;
    return false;
}


/****************************************************************************/
struct GsPriQ {
    garr<AntecedT> a;
    int first;
    //set<GsT*> s;

    void init() {
        a.init();
        //s = set<GsT*>();
    }
    void append(AntecedT ante) {
        //int IsNew = s.insert(pGs).second;
        //if (!IsNew) {return;}
        a.append(ante); 
        push_heap(a.begin(), a.end(), AntecedByGreaterGsId);
    }
    AntecedT pop()  {
        if (a.n == 0) return EmptyAnteced;
        pop_heap(a.begin(), a.end(), AntecedByGreaterGsId);
        AntecedT ret = a.PopLast();
        //s.erase(ret);
        return ret;
    }
    AntecedT peek() {return (a.n == 0 ? EmptyAnteced : a[0]);}
    bool is_empty() {return (a.n == 0);}

    void clear() {
        a.n=0; 
        //s.clear();
    }

    /*void init() {first=0; a.init();}
    void append(AntecedT ante) {a.append(ante);}
    AntecedT pop()  {return (first >= a.n ? EmptyAnteced : a[first++]);}
    AntecedT peek() {return (first >= a.n ? EmptyAnteced : a[first]);}
    bool is_empty() {return (first >= a.n);}
    void clear() {first=0; a.n=0;}*/
};


typedef pair<int,int> IntIntPair;
typedef pair<GsT*,int> GsIntPair;

struct GsSetByPly {
    set<GsT*,LessGsById> ply[2];
};


template <typename T, bool (*lt)(T const& x, T const& y)>
struct heap {
    garr<T> h;

    void insert(T x) {
        h.append(x);
        push_heap(h.begin(), h.end(), lt);
    }

    T pop() {
        assert(h.n > 0);
        T ret = h[0];
        while (h.n > 0 && h[0] == ret) {
            pop_heap(h.begin(), h.end(), lt);
            h.n--;
        }
        return ret;
    }

    T peek() {
        return h[0];
    }

    bool empty() {return (h.n == 0);}

    T top() {return h[0];}
    void push(T x) {insert(x);}
    
};



class GsBuilder;


#define DEFN_GST_PRI 1
#define DEFN_INS_GST_PRI 1
#define MONO_GST_PRI 2
#define FORC_GST_PRI 3
#define WIN_GST_PRI 4
#define MAX_GST_PRI 4

   
#define EM_DRYRUN 0
#define EM_SCHED 1
#define EM_FORCE 2
#define EM_WIN 3


typedef int ChronoT;

/*****************************************************************************
* AsgnT Header.    }}}{{{1
*****************************************************************************/
struct AsgnT {
    ChlitListT chlits;   // Chosen lits, arranged according to quantifier block
    GwSetT *LitWatch;   // LitWatch[lit] lists the game-states that are watching lit.
    uint* LitToDL;       // maps lit to the decision level in which it was chosen or forced
    //set<LitT> FreeVars;
    garri* UndoListByDL;
    LitT* UndoTarg;
    garr<GsIntPair>* FullCtlrUndo;
    garr<GsIntPair>* SemiCtlrUndo;
    set<GsT*,LessGsById>* FullCtlrToGSes;
    map<GsT*,LitT,LessGsById>* SemiCtlrToGSes;
    AntecedT* DepByLit;
    BitSetT LitBmap;
    ChronoT* LitToChrono;
    long long* LitToAbsTime;  // For debugging only
    LitT* ChronoToLit;
    FmlaT FreeFmla;
    set<LitT> LitsInConflict;
    set<LitT> BattleVars;
    garri BattleVarList;

    int TargGate;
    int CurChrono;
    int LastInitGs;
    int LastDefnGs;
    int DeathChrono;
    long long NumAppends;

    GsPriQ PendingExecs[MAX_GST_PRI+1]; 

    //int IsRestarting;
    garri InitChoices;
    //garri TargStack;
    int ixChoi;
    int NumSinceCut;
    int NumUntilRestart;
    long long LastPrintTime;
    int NumRestarts;
    set<pair<int, vector<int> > > runs;

    LvCacheEntry* LvCache;
    LitT* LitToLiveParent;  // maps lit to nondead orig clause with it
    set<LitT>* ParentToWatchers;   // maps parent literal to set of watching lits
    int NumMonoProps;  // for use as a timestamp
    set<LitT> PendingMonotoneLits;
    heap<LitT,lt_f> PendingDeadPars;
    long long *LastChkPar;

    char* TriedLit;
    int *LastHitDL;  // for debugging
    VsidsT* vsids;

    


    #define FPROTO_AsgnT
    #include "ghostq.proto"
    #undef FPROTO_AsgnT
    
    inline bool HasLit(int x) {return LitToDL[x] != NO_DECLEV;}
    inline bool HasLitOrNeg(int x) {return HasLit(x) || HasLit(negv(x));}
    inline bool HasLitOrPhant(int x, int ply) {
        if (LitToDL[x] != NO_DECLEV) return true;
        if (LitToDL[x + PhantOffset[ply]] != NO_DECLEV) return true;
        return false;
    }

    inline int GetLastChlit() {
        return chlits.last();
    }

    inline uint GetCurDL() {
        return LitToDL[GetLastChlit()];
    }

    bool IsChlit(int lit) {
        return (ChlitByDL(LitToDL[lit]) == lit);
    }

    inline GwSetT::iterator LitWatchInsert(int lit, GsT* pGs) {
        //assert(LitWatch[lit].count(pGs) == 0);
        return LitWatch[lit].insert(pair<int,GsT*>(pGs->id, pGs)).first;
        //return LitWatch[lit].insert(LitWatch[lit].end(), pGs);
    }
};


void AsgnT::init() {
    LitToDL = talloc(uint, pGloProb->LastPlit+1);
    //FreeVars = set<LitT>();
    UndoListByDL = talloc(garri, pGloProb->LastInpLit+1);
    UndoTarg = talloc(LitT, pGloProb->LastInpLit+1);
    FullCtlrUndo = talloc(garr<GsIntPair>, pGloProb->LastInpLit+1);
    SemiCtlrUndo = talloc(garr<GsIntPair>, pGloProb->LastInpLit+1);
    FullCtlrToGSes = new set<GsT*,LessGsById>[pGloProb->LastPlit+1];
    SemiCtlrToGSes = new map<GsT*,LitT,LessGsById>[pGloProb->LastPlit+1];
    DepByLit = talloc(AntecedT, pGloProb->LastPlit+1);
    LitToChrono = talloc(ChronoT, pGloProb->LastPlit+1);
    LitToAbsTime = talloc(long long, pGloProb->LastPlit+1);
    ChronoToLit = talloc(LitT, pGloProb->LastPlit+1);
    LvCache = talloc(LvCacheEntry, pGloProb->LastPlit+1);
    LastHitDL = talloc(int, pGloProb->LastInpLit+1);
    vsids = talloc(VsidsT, pGloProb->LastPlit+1);
    LitToLiveParent = talloc(LitT, pGloProb->LastPlit+1);
    LastChkPar = talloc(long long, pGloProb->LastPlit+1);
    ParentToWatchers = new set<LitT>[pGloProb->LastPlit+1];
    TriedLit = talloc(char, pGloProb->LastPlit+1);
    //DoneGsSet = set<GsT*>();
    LitsInConflict = set<LitT>();
    BattleVars = set<LitT>();
    BattleVarList = garri();
    chlits.flat.init();
    LastInitGs = (1 << 30);
    NumAppends = 0;
    for (int i=0; i <= MAX_GST_PRI; i++) {
        PendingExecs[i].init();
    }
    LitBmap.alloc(pGloProb->LastPlit+1);
    //PlitsPending.init();
    for (int i=0; i <= pGloProb->LastPlit; i++) {
        LitToDL[i] = NO_DECLEV;
        LitToChrono[i] = ((ChronoT) NO_DECLEV);
    }
    CurChrono = 0;
    NumMonoProps = 1;
    this->AppendChlit(ZERO_LIT);
    this->LitToDL[ZERO_LIT] = ZERO_DECLEV;
    this->LitBmap.set(ZERO_LIT, 1);
}

void AsgnT::DumpFullCtlrs() {
    for (int glit = 0; glit < pGloProb->LastGlit; glit++) {
        bool GotOne = false;
        int noch = 0;
        foreach (itGs, FullCtlrToGSes[glit]) {
            if (!GotOne++) {
                noch += printf("Glit %4i controls clauses: ", glit);
            }
            if (noch >= LINE_LEN - 6) printf("\n");
            noch += printf("%i ", (*itGs)->id);
        }
        if (GotOne) printf("\n");
    }
}

GsT* AsgnT::GetDepByLit(int lit) {
    return DepByLit[lit].pDepGs;
}


int Phantomize(int lit, int ply) {
    return lit + PhantOffset[ply];
}

int Unphantomize(int lit) {
    if (lit >= PhantOffset[0]) {
        if (lit >= PhantOffset[1]) {
            lit -= PhantOffset[1];
        } else {
            lit -= PhantOffset[0];
        }
    }
    return lit;
}

int PlyFromLit(int lit) {
    if (IsInputLit(lit)) {
        return QtypeNumFromLit(lit);
    } else {
        if (lit > PhantOffset[1]) return 1;
        else if (lit > PhantOffset[0]) return 0;
        else assert(0);
    }
}


int NarfType(int lit) {
    if (IsGateLit(lit)) {return 0;}
    if (QtypeFromLit(lit) == 'E') {return 3;}
    if (QtypeFromLit(lit) == 'A') {return 2;}
    assert(false);
}

int CmpLitByNarf(const int *pLit1, const int *pLit2) {
    int x1 = (NarfType(*pLit1) << 26) + *pLit1;
    int x2 = (NarfType(*pLit2) << 26) + *pLit2;
    return x1 - x2;
}
        
int CmpLitByQb(const int *pLit1, const int *pLit2) {
    int q1 = QbFromLit[*pLit1];
    int q2 = QbFromLit[*pLit2];
    if (q1 == q2) {return 0;}
    else if (q1 < q2) {return -1;}
    else {return 1;}
}
        
int CmpLitByDL(const int *pLit1, const int *pLit2) {
    int d1 = pGloAsgn->LitToDL[*pLit1];
    if (d1==NO_DECLEV &&  pGloAsgn->LitToDL[negv(*pLit1)]==NO_DECLEV) {d1--;}
    int d2 = pGloAsgn->LitToDL[*pLit2];
    if (d2==NO_DECLEV &&  pGloAsgn->LitToDL[negv(*pLit2)]==NO_DECLEV) {d2--;}
    if (d1 == d2) {return 0;}
    else if (d1 < d2) {return -1;}
    else {return 1;}
}

int CmpLitByChrono(const int *pLit1, const int *pLit2) {
    ChronoT d1 = pGloAsgn->LitToChrono[*pLit1];
    ChronoT d2 = pGloAsgn->LitToChrono[*pLit2];
    //return d1 - d2; // NO!
    if (d1 == d2) {return (*pLit1 - *pLit2);}
    else if (d1 < d2) {return -1;}
    else {return 1;}
}

struct LitLessByDL {
    bool operator() (const int& Lit1, const int& Lit2) const {
        int d1 = pGloAsgn->LitToDL[Lit1];
        int d2 = pGloAsgn->LitToDL[Lit2];
        if (d1 == d2) {
            if (Lit1 == Lit2) return false;
            assert(!((pGloAsgn->IsChlit(Lit1) && (pGloAsgn->IsChlit(Lit2)))));
            if (pGloAsgn->IsChlit(Lit1)) return true;
            if (pGloAsgn->IsChlit(Lit2)) return false;
            return (Lit1 < Lit2);
        }
        return (d1 < d2);
    }
};


/*bool LitLessByChronof(const int& Lit1, const int& Lit2) {
    ChronoT d1 = pGloAsgn->LitToChrono[Lit1];
    ChronoT d2 = pGloAsgn->LitToChrono[Lit2];
    if (d1 == d2) {assert(Lit1 == Lit2);}
    return (d1 < d2);
}*/

struct LitLessByChrono {
    bool operator() (const int& Lit1, const int& Lit2) const {
        ChronoT d1 = pGloAsgn->LitToChrono[Lit1];
        ChronoT d2 = pGloAsgn->LitToChrono[Lit2];
        //if (d1 == d2) {assert(Lit1 == Lit2);}
        return (d1 < d2);
    }
};

struct LitGtByChrono {
    bool operator() (const int& Lit1, const int& Lit2) const {
        ChronoT d1 = pGloAsgn->LitToChrono[Lit1];
        ChronoT d2 = pGloAsgn->LitToChrono[Lit2];
        //if (d1 == d2) {assert(Lit1 == Lit2);}
        return (d1 > d2);
    }
};



/*****************************************************************************
* Functions dealing mainly with game-state objects.    }}}{{{1
*****************************************************************************/

bool GsT::IsBaseGs() {
    return (id <= pGloAsgn->LastInitGs);
}

int GsT::IsTrigLit(int CurLit) {
    ap<LitT> trigs;
    if (IsInputLit(CurLit)) {trigs = TrigAE[QtypeNumFromLit(CurLit)];}
    else {trigs = TrigG;}
    return binary_search(trigs.begin(), trigs.end(), CurLit);
}

int GsT::GetFirstFreeAE(int ply) {
    foreach (itLit, TrigAE[ply]) {
        if (pGloAsgn->HasLit(*itLit)) continue;
        if (pGloAsgn->HasLit(negv(*itLit))) continue;
        return *itLit;
    }
    return 0;
}

int GsT::GetFreeResQb() {
    foreach (itLit, ImpLits) {
        if (pGloAsgn->HasLit(*itLit)) continue;
        return QbFromLit[*itLit];
    }
    return MAX_QB;
}

int GsT::GetWinner() {
    return (FreeFmla==1 ? 'E' : (FreeFmla==0 ? 'A' : die()));
}

int GsT::IsBlocked() {
    if (HasBlockedImps()) return true;
    return false;
}

int GsT::HasBlockedImps() {
    for (int i=0; i < ImpLits.n; i++) {
        if (pGloAsgn->HasLit(negv(ImpLits[i]))) {
            return ImpLits[i];
        }
    }
    return false;
}

int AsgnT::GetUpstreamFreeLit(int flit, ap<int>& LitList) {
    int WatchQb = QbFromLit[flit];
    if (IsGateLit(flit)) WatchQb += 1; //return 1; //WatchQb++;
    for (int i=0; i < LitList.n; i++) {
        int CurLit = LitList[i];
        if (QbFromLit[CurLit] < WatchQb && !pGloAsgn->HasLit(CurLit)) {
            return CurLit;
        }
    }
    return 0;
}

int AsgnT::GetFirstFreeLit(ap<int>& LitList) {
    for (int i=0; i < LitList.n; i++) {
        int CurLit = LitList[i];
        if (!pGloAsgn->HasLit(CurLit)) {
            return CurLit;
        }
    }
    return 0;
}

int AsgnT::GetFirstFreeLitph(ap<int>& LitList, int ply) {
    for (int i=0; i < LitList.n; i++) {
        int CurLit = LitList[i];
        if (!HasLit(CurLit) && !HasLit(Phantomize(CurLit, ply))) {
            return CurLit;
        }
    }
    return 0;
}

int AsgnT::GetFreeLitph(ap<int>& LitList, int ply) {
    int ret = 0;
    foreach (it, LitList) {
        int CurLit = *it;
        if (!HasLit(CurLit) && !HasLit(Phantomize(CurLit, ply))) {
            if (ret != 0) {
                return (1 << 30);
            }
            ret = CurLit;
        }
    }
    return ret;
}

int GsT::HasAllWatchedTrigs() {
    for (int i=0; i < 4; i++) {
        if (!pGloAsgn->HasLit(WatchRawAE[i])) return false;
    }
    return true;
}

int GsT::TrigRipeI(int ply) {
    if (!pGloAsgn->HasLit(WatchAE(ply, 1))) {return 0;}
    if (pGloAsgn->HasLit(negv(WatchAE(ply, 0)))) {return 0;}
    if (pGloAsgn->HasLit(WatchAE(ply, 0))) {return 2;}
    if (IsGateLit(WatchAE(ply, 0))) {return 0;}
    return 1;
}

int GsT::TrigRipe(int ply) {
    if (!pGloAsgn->HasLitOrPhant(WatchAE(ply, 1), ply)) {return 0;}
    if (pGloAsgn->HasLit(negv(WatchAE(ply, 0)))) {return 0;}
    if (pGloAsgn->HasLitOrPhant(WatchAE(ply, 0), ply)) {return 2;}
    //if (IsGateLit(WatchAE(ply, 0))) {return 0;}
    return 1;
}

int GsT::Is_ExactlyOne_LosrLit_Free() {
    assert(OrigOrLrn == LRN_TAG);
    int losr = 1 - FreeFmla;
    return (pGloAsgn->HasLit(WatchAE(losr, 0)) ^ pGloAsgn->HasLit(WatchAE(losr, 1)));
}

int GsT::AreBothTrigsFree() {
    assert(OrigOrLrn == LRN_TAG);
    int ply = 1-FreeFmla;
    if (pGloAsgn->HasLit(WatchAE(ply, 0))) {return false;}
    if (pGloAsgn->HasLit(WatchAE(ply, 1))) {return false;}
    return true;
}

int GsT::IsEitherTrigFree() {
    assert(OrigOrLrn == LRN_TAG);
    int ply = 1-FreeFmla;
    if (!pGloAsgn->HasLit(WatchAE(ply, 0))) {return true;}
    if (!pGloAsgn->HasLit(WatchAE(ply, 1))) {return true;}
    return false;
}

int GsT::HasCtlr() {
    if (!pGloAsgn->HasLit(WatchAE(0,1))) return false;
    if (!pGloAsgn->HasLit(WatchAE(1,1))) return false;
    //if (WatchAE(0,0) != WatchAE(1,0)) return false;
    if (pGloAsgn->HasLit(WatchAE(0,0)) ^ pGloAsgn->HasLit(WatchAE(1,0))) return false;
    if (!IsGateLit(FirstGateWatch())) return false;
    //if (id > pGloAsgn->LastInitGs) return false;
    return true;
}

int GsT::GetFullCtlr(int pass, int inv) {
    /* Note: This function is tail-recursive. */
    if (!HasCtlr()) return (pass ^ inv);
    int ret = FirstGateWatch() ^ inv;
    int IsAnd = IsAndGate(ret);
    return GlitToGs[ret ^ !IsAnd]->GetFullCtlr(ret ^ !IsAnd, !IsAnd);
}


void GsT::InitTrigLits(ap<int> args) {
    garri gTrigAE[2] = {garri(), garri()};
    garri gTrigG = garri();
    for (int i=0; i < args.n; i++) {
        if (IsGateLit(args[i])) {
            gTrigG.append(args[i]);
        } else {
            int qtype = QtypeFromLit(args[i]);
            if      (qtype == 'A') {gTrigAE[0].append(args[i]);}
            else if (qtype == 'E') {gTrigAE[1].append(args[i]);}
            else assert(false);
        }
    }
    gTrigAE[0].sort(CmpLitByQb);
    gTrigAE[1].sort(CmpLitByQb);
    //TrigG.sort(CmpLitByQb);
    this->TrigAE[0] = ap<int>(gTrigAE[0].arr, gTrigAE[0].n);
    this->TrigAE[1] = ap<int>(gTrigAE[1].arr, gTrigAE[1].n);
    this->TrigG   =   ap<int>(gTrigG.arr,     gTrigG.n);
}

int GsT::PrintIndex() {
    // This is for debugging.
    int ret = pGloProb->clauses.indexof(this);
    printf(  "pGloProb->clauses.arr[%i]\n", ret);
    return ret;
}

void GsT::RegisterLitsHave() {
    for (int i=0; i < ReqLits.n; i++) {
        int lit = ReqLits[i];
        pGloProb->LitHave[lit].append(this);
    }
    //volnop();
    if (pGloAsgn->BattleVars.count(absv(this->GateLit)) == 1) {
        IsBattle = true;
        assert(this->FreeFmla == 1^signv(this->GateLit));
    }
}

int GsT::WhichForced(int TestLit, set<int>* pUndoSet = NULL) {
    /* Precond: TestLit is in this game state. */
    GsT* pCurLs = this;
    if (IsInputLit(TestLit)) {
        int qtype = QtypeFromLit(TestLit);
        FmlaT CurTransFmla = pCurLs->FreeFmla;
        if ((CurTransFmla == true  && qtype == 'E') || 
            (CurTransFmla == false && qtype == 'A'))
        {
            return TestLit;
        }
    }
    return negv(TestLit);
}




/*****************************************************************************
* Stuff  }}}{{{1
*****************************************************************************/

int AsgnT::IsBattleLit(int glit) {
    return binary_search(BattleVarList.begin(), BattleVarList.end(), absv(glit));
}

int AsgnT::GetSemiCtlrPar(int ctlr, int GsId) {
    return SemiCtlrToGSes[ctlr].find(pGloProb->clauses.arr[GsId])->second;
}

void AsgnT::DumpSemiCtlrToGSes(int ctlr) {
    foreach (it, SemiCtlrToGSes[ctlr]) {
        printf(" %i", it->first->id);
    }
    printf("\n");
}

GsT* GsById(int id) {
    return pGloProb->clauses[id];
}

inline int AsgnT::GetWinner() {
    return (FreeFmla==1 ? 'E' : (FreeFmla==0 ? 'A' : die()));
}

void AsgnT::CutVsids(int power=3) {
    for (int lit=0; lit <= pGloProb->LastInpLit; lit++) {
        vsids[lit].x[0] >>= power;
        vsids[lit].x[1] >>= power;
    }
}


void AsgnT::InitParents() {
    for (int lit = pGloProb->LastGlit; lit >= 2; lit--) {
        garri pars = pGloProb->ParentLit[lit];
        if (pars.n == 0) {
            continue;
        }
        //LitToLiveParent[lit] = pars[0];
        PickParent(lit);
    }
}

void DumpIntSet(set<int,struct LitLessByChrono>* pSet, const char* pEnd) {
    auto(iter, pSet->begin());
    printf("[ ");
    for (; iter != pSet->end(); ++iter) {
        printf("%i ", *iter);
    }
    printf("]%s", pEnd);
}


void GsT::dump() {
    DumpRaw();
}

void GsT::DumpBrief() {
    printf("  TrigA "); DumpLitsEnc(TrigAE[0], ",  ");
    printf("TrigE "); DumpLitsEnc(TrigAE[1], "\n");
    printf("  TrigG "); DumpLitsEnc(TrigG, ",\n");
    printf("  Imps "); DumpLitsEnc(ImpLits, "\n");
    /*if (OrigOrLrn==ORIG_TAG) {printf("GateLit=%i,  ", GateLit);}
    else printf("FreeFmla=%i,  ", FreeFmla);
    printf("WatchAE: [%i, %i], [%i, %i].  %s litset.\n", 
        decv(WatchAE(0,0)), decv(WatchAE(0,1)),
        decv(WatchAE(1,0)), decv(WatchAE(1,1)),
        (OrigOrLrn==ORIG_TAG ? "Original" : "Learned"));*/
}

void GsT::DumpRaw() {
    if (this == NULL) {printf("(null)\n"); return;}
    printf("clauses[%i] (%s).  ", id, OrigOrLrn==ORIG_TAG ? "orig" : "lrn");
    printf("GateLit=%i.\n", GateLit);
    printf("winner=%i('%c').\n", FreeFmla, (FreeFmla==NO_WINR ? '-' : (FreeFmla ? 'E' : 'A')));
    printf("ReqLits: "); DumpLitsEnc(ReqLits);
    printf("TrigA: "); DumpLitsEnc(TrigAE[0]);
    printf("TrigE: "); DumpLitsEnc(TrigAE[1]);
    printf("TrigG: "); DumpLitsEnc(TrigG);
    printf("ImpLits: "); DumpLitsEnc(ImpLits);
    //printf("WatchedLit: [%i, %i].  %s litset, ", decv(WatchedLit[0]), decv(WatchedLit[1]),
    //    (OrigOrLrn==ORIG_TAG ? "Original" : "Learned"));
    printf("Watch: [%i, %i], [%i, %i], %i.\n", 
        decv(WatchAE(0,0)), decv(WatchAE(0,1)),
        decv(WatchAE(1,0)), decv(WatchAE(1,1)),
        decv(WatchedRes));
}

void GsT::DumpCtlrs() {
    foreach (it, ctlrs) {
        printf("%i ", *it);
    }
    printf("\n");
}

        
int AsgnT::ChlitByDL(int dl) {
    return chlits.flat[dl];
}

int AsgnT::QbFromDL(int dl) {
    return QbFromLit[ChlitByDL(dl)];
}

int AsgnT::QtypeFromDL(int dl) {
    return GloQtype(QbFromDL(dl));
}

int AsgnT::GlitToMinChrono(int lit, int ply) {
    int ret = LitToChrono[lit];
    if (IsInputLit(lit)) return ret;
    int alt = LitToChrono[Phantomize(lit, ply)];
    return min(ret, alt);
}



/*****************************************************************************
* Functions for pending queue of game-states.  }}}{{{1
*****************************************************************************/
void AsgnT::AddPendingLs(GsT* pCurLs, int score, int ctlr=0) {
    PendingExecs[score].append(AntecedT(pCurLs, 0, 0));
}

void AsgnT::AddPendingMonot(int flit) {
    PendingExecs[MONO_GST_PRI].append(AntecedT(0, 0, MONO_ANTECED, flit));
}

AntecedT AsgnT::PopFirstPendingLs(int flag) {
    if (!PendingExecs[DEFN_GST_PRI].is_empty()) {
        return PendingExecs[DEFN_GST_PRI].pop();
    }
    if (flag) return AntecedT();
    if (!PendingExecs[MONO_GST_PRI].is_empty()) {
        return PendingExecs[MONO_GST_PRI].pop();
    }
    if (!PendingExecs[FORC_GST_PRI].is_empty()) {
        return PendingExecs[FORC_GST_PRI].pop();
    }
    if (!PendingExecs[WIN_GST_PRI].is_empty()) {
        return PendingExecs[WIN_GST_PRI].pop();
    }
    return AntecedT();
}

bool AsgnT::HasPendingForced() {
    if (!PendingExecs[DEFN_GST_PRI].is_empty()) return true;
    if (!PendingExecs[MONO_GST_PRI].is_empty()) return true;
    if (!PendingExecs[FORC_GST_PRI].is_empty()) return true;
    return false;
}

int AsgnT::HasTargVar() {
    return HasLit(TargGate) || HasLit(negv(TargGate));
}

/*****************************************************************************
* Functions for watched literals.  }}}{{{1
*****************************************************************************/

void AsgnT::PickWatchedLitsG(GsT* pCurLs, int FreeGates[4], int ply) {
    int iw=0;
    int NumLits  =  pCurLs->TrigG.n;
    int* PossLits = pCurLs->TrigG.arr;
    //uint* LitToDL = pGloAsgn->LitToDL;
    for (int ixLit = 0; ixLit < NumLits; ixLit++) {
        int CurLit = PossLits[ixLit];
        if (LitToDL[CurLit] == NO_DECLEV && (!UsePhant || LitToDL[CurLit + PhantOffset[ply]] == NO_DECLEV)) {
            FreeGates[iw+(ply<<1)] = CurLit;
            iw++;
            if (iw == 2) break;
        } 
    }
}
//void (*volatile volPickG)(GsT*, int[4]) = &PickWatchedLitsG;


int AsgnT::PickWatchedLitsAE(GsT* pCurLs, int ply, int NewWatches[2], int FreeGates[4]) {

    int PrevWatched[2] = {pCurLs->WatchAE(ply, 0), pCurLs->WatchAE(ply, 1)};
    PickWatchedLitsG(pCurLs, FreeGates, ply);
        
    int iw=0;

    int NumLits  =  pCurLs->TrigAE[ply].n;
    int* PossLits = pCurLs->TrigAE[ply].arr;
    for (int ixLit = 0; ixLit < NumLits; ixLit++) {
        int CurLit = PossLits[ixLit];
        if (LitToDL[CurLit] == NO_DECLEV) {
            NewWatches[iw++] = CurLit;
            if (iw == 2) return 0;
        } 
    }

    if (iw < 2 && (!UsePhant || pCurLs->id <= LastInitGs || ply == (1 ^ pCurLs->FreeFmla))) {
        if (FreeGates[2*ply+1] != 0) {
            NewWatches[iw++] = FreeGates[2*ply+0];
            if (iw==2) return 0;
            NewWatches[iw++] = FreeGates[2*ply+1];
            return 0;
        } else if (FreeGates[2*ply+0] != 0) {
            NewWatches[iw++] = FreeGates[2*ply+0];
            if (iw==2) return 0;
        }
    }

    if (NewWatches[0] == 0) {
        NewWatches[0] = PrevWatched[0];
        NewWatches[1] = PrevWatched[1];
    } else if (NewWatches[1] == 0) {
        if (PrevWatched[0] == NewWatches[0]) {
            NewWatches[1] = PrevWatched[1];
        } else {
            NewWatches[1] = PrevWatched[0];
            if (HasLit(PrevWatched[1])) {assert(LitToDL[PrevWatched[0]] >= LitToDL[PrevWatched[1]]);}
        }
    } else assert(false);
    if (NewWatches[0] != 0) assert(NewWatches[0] != NewWatches[1]);
    return 0;
}



void AsgnT::FixWatchedLits(GsT* pCurLs) {
    //if (pCurLs->ReqLits.n == 1) return;  // New learned GSes don't pick ResLit correctly.
    assert(pCurLs->id != 0);  // Can't insert into LitWatch without an id!
    /*for (int i=0; i < 4; i++) {
        if (pCurLs->WatchRawAE[i] == 0) continue; 
        //assert(pCurLs->itWatchRawAE[i] == pGloAsgn->LitWatch[pCurLs->WatchRawAE[i]].find(pCurLs));
        assert(LitWatch[pCurLs->WatchRawAE[i]].count(pCurLs));
    }*/
    int FreeGates[4] = {0,0,0,0};
    int NewWatchesB[4] = {0,0,0,0};
    //PickWatchedLitsG(pCurLs, FreeGates);
    //volPickG(pCurLs, FreeGates);
    PickWatchedLitsAE(pCurLs, 0, NewWatchesB, FreeGates);
    PickWatchedLitsAE(pCurLs, 1, NewWatchesB+2, FreeGates);
    if (NewWatchesB[2] != 0) assert(NewWatchesB[2] != NewWatchesB[3]);

    GwSetT::iterator itWatchNew[4] = {
        pCurLs->itWatchRawAE[0],
        pCurLs->itWatchRawAE[1],
        pCurLs->itWatchRawAE[2],
        pCurLs->itWatchRawAE[3]
    };
    int* OldWatches = pCurLs->WatchRawAE;
    GwSetT::iterator* itWatchOld = pCurLs->itWatchRawAE;
    for (int i=0; i < 4; i++) {
        bool ok;
        int cur = pCurLs->WatchRawAE[i];
        if (cur==0) continue;
        for (int j=0; j < 4; j++) {if (NewWatchesB[j] == cur) goto SkipDel;}
        ok = (i < 2) || (
                (OldWatches[0] != OldWatches[i]) &&
                (OldWatches[1] != OldWatches[i]));
        if (ok) {
            //assert(pCurLs->itWatchRawAE[i] == pGloAsgn->LitWatch[cur].find(pCurLs));
            pGloAsgn->LitWatch[cur].erase(pCurLs->itWatchRawAE[i]);
        }
        itWatchNew[i] = GwSetT::iterator();
        SkipDel:;
    }
    for (int i=0; i < 4; i++) {
        bool ok;
        int cur = NewWatchesB[i];
        if (cur==0) continue;
        for (int j=0; j < 4; j++) {
            if (pCurLs->WatchRawAE[j] == cur) {
                itWatchNew[i] = itWatchOld[j];
                goto SkipAdd;
            }
        }
        ok = (i < 2) || (
                (NewWatchesB[0] != NewWatchesB[i]) &&
                (NewWatchesB[1] != NewWatchesB[i]));
        if (ok) {
            itWatchNew[i] = pGloAsgn->LitWatchInsert(cur, pCurLs);
        } else {
            itWatchNew[i] = GwSetT::iterator();
        }
        SkipAdd:;
        //assert(pGloAsgn->LitWatch[cur].count(pCurLs) == 1);
    }
    for (int i=0; i < 4; i++) {
        if (NewWatchesB[i] == 0) {assert(pCurLs->WatchRawAE[i] == 0); continue; }
        pCurLs->WatchRawAE[i] = NewWatchesB[i];
        pCurLs->itWatchRawAE[i] = itWatchNew[i];
        //assert(pCurLs->itWatchRawAE[i] == pGloAsgn->LitWatch[pCurLs->WatchRawAE[i]].find(pCurLs));
    }

    int ResLitW = 0;
    if (pCurLs->OrigOrLrn == LRN_TAG && pCurLs->ImpLits.n > 0 && pCurLs->Is_ExactlyOne_LosrLit_Free()) {
        int losr = 1 - pCurLs->FreeFmla;
        int LastTrig = pCurLs->WatchAE(losr, 0);
        ResLitW = GetFirstFreeLit(pCurLs->ImpLits);
        int OldRes = pCurLs->WatchedRes;
        if (OldRes != ResLitW && ResLitW != 0) {
            if (QbFromLit[ResLitW] < QbFromLit[LastTrig]) {
                assert(ResLitW == PickNewResWatch(pCurLs));
                if (OldRes != 0) {pGloAsgn->LitWatch[OldRes].erase(pCurLs->itWatchRes);}
                pCurLs->itWatchRes = LitWatchInsert(ResLitW, pCurLs);
                pCurLs->WatchedRes = ResLitW;
            }
        }
    }
    /*for (int i=0; i < 4; i++) {
        if (pCurLs->WatchRawAE[i] == 0) continue; 
        //assert(pCurLs->itWatchRawAE[i] == pGloAsgn->LitWatch[pCurLs->WatchRawAE[i]].find(pCurLs));
        assert(LitWatch[pCurLs->WatchRawAE[i]].count(pCurLs));
    }*/

}

int AsgnT::PickNewResWatch(GsT* pCurLs) {
    int ResLitW = 0;
    if (!pCurLs->Is_ExactlyOne_LosrLit_Free()) return 0;
    int losr = 1 - pCurLs->FreeFmla;
    int LastTrig = pCurLs->WatchAE(losr, 0);
    ResLitW = GetFirstFreeLit(pCurLs->ImpLits);
    if (ResLitW==0) return 0;
    if (QbFromLit[ResLitW] < QbFromLit[LastTrig]) {return ResLitW;}
    else return 0;
}


/*****************************************************************************
* Functions for controllers.  }}}{{{1
*****************************************************************************/

void AsgnT::GetSemiCtlrs(int glit, int parent, map<LitT,LitT>& ret) {
    int SubCtlr = AbsGlitToGs(glit)->GetFullCtlr(0, !IsAndGate(glit));
    if (absv(SubCtlr) != 0) {
        ret.insert(IntIntPair(glit, parent));
        GetSemiCtlrs(SubCtlr, glit, ret);
        return;
    }
    bool AlreadyHave = (ret.count(glit) == 1);
    if (AlreadyHave) return;
    ret.insert(IntIntPair(glit, parent));
    if (IsAndGate(glit)) return;
    GsT* pCurLs = AbsGlitToGs(glit);
    foreach (itLit, pCurLs->TrigG) {
        int CurLit = negv(*itLit);
        if (HasLit(negv(CurLit))) continue;
        GetSemiCtlrs(CurLit, glit, ret);
    }
}

void AsgnT::SemiCtlrProp(GsT* pCurLs, int FullCtlr, bool OnlyAdd, bool NoUpdUndo=false) {
    map<LitT,LitT> NewSemis = map<LitT,LitT>();
    if (FullCtlr != 0) {
        int UpperCtlr = pCurLs->FirstGateWatch();
        if (GloSanChk >= 2) assert(UpperCtlr == pCurLs->GetLastGlit());
        GetSemiCtlrs(FullCtlr, UpperCtlr, NewSemis);
    }
    SemiCtlrProp(pCurLs, NewSemis, OnlyAdd, NoUpdUndo);
}
    
void AsgnT::SemiCtlrProp(GsT* pCurLs, map<LitT,LitT>& NewSemis, bool OnlyAdd, bool NoUpdUndo=false) {
    if (!FindSemiCtlrs) return;
    {
        auto(itOld, pCurLs->ctlrs.begin());
        auto(itEndOld, pCurLs->ctlrs.end());
        auto(itNew, NewSemis.begin());
        auto(itEndNew, NewSemis.end());
        while (itNew != itEndNew || itOld != itEndOld) {
            int NewLit = ((itNew == itEndNew) ? (1<<30) : itNew->first);
            int OldLit = ((itOld == itEndOld) ? (1<<30) : *itOld);
            /* TODO: Be more discriminating here; don't add every one to the queue! */
            if (UseSemiCtlrs && NewLit != (1<<30) && NewLit < OldLit) {
                AntecedT NewAnteced = AntecedT(pCurLs, NewLit);
                PendingExecs[FORC_GST_PRI].append(NewAnteced);
            }
            if (NewLit == OldLit) {
                ++itNew;
                ++itOld;
            } else if (NewLit < OldLit) {
                if (!NoUpdUndo) {
                    if (SemiCtlrToGSes[NewLit].count(pCurLs)==0) {
                        SemiCtlrUndo[GetCurDL()].append(GsIntPair(pCurLs, -NewLit));
                    }
                }
                SemiCtlrToGSes[NewLit].insert(GsIntPair(pCurLs, itNew->second));
                ++itNew;
            } else {
                if (!OnlyAdd) {
                    if (!NoUpdUndo) {
                        SemiCtlrUndo[GetCurDL()].append(GsIntPair(pCurLs, OldLit));
                    }
                    assert(SemiCtlrToGSes[OldLit].count(pCurLs)==1);
                    //SemiCtlrToGSes[OldLit].erase(pCurLs);
                }
                ++itOld;
            }
        }
    }
    /*if (GloDbgLev >= 1960 && pCurLs->id == 998) {
        if (NewSemis != pCurLs->ctlrs) {
            printf("NewSemis: ");
            DumpIntSet(&NewSemis, "\n");
        }
    }*/
    if (!OnlyAdd) {
        pCurLs->ctlrs.clear();
    }
    foreach (it, NewSemis) {
        pCurLs->ctlrs.insert(it->first);
        assert(SemiCtlrToGSes[it->first].count(pCurLs)==1);
    }
    /*if (pCurLs->ctlrs.size() > 1) {
        stop();
    }*/
    //pCurLs->ctlrs = NewSemis;
}

void AsgnT::CtlrProp(GsT* pCurLs, bool NoUpdUndo=false) {
    if (!FindFullCtlrs) return;
    if (pCurLs->OrigOrLrn == LRN_TAG) {
        int NewCtlr = pCurLs->GetFullCtlr(0,0);
        SemiCtlrProp(pCurLs, NewCtlr, 0, NoUpdUndo);
        if (NewCtlr != 0) {
            int OldCtlr = pCurLs->ctlr;
            if (NewCtlr != OldCtlr) {
                if (GloDbgLev >= 2007) printf("FixW: clauses[%i]->ctlr = %i\n", pCurLs->id, NewCtlr);
                if (!NoUpdUndo) {
                    FullCtlrUndo[GetCurDL()].append(GsIntPair(pCurLs, OldCtlr));
                }
                pCurLs->ctlr = NewCtlr; 
                FullCtlrToGSes[NewCtlr].insert(pCurLs);
            }
            PendingExecs[FORC_GST_PRI].append(AntecedT(pCurLs, NewCtlr));
        }
    } else if (IsAndGate(pCurLs->GateLit)) {
        int glit = pCurLs->GateLit;
        if (!pCurLs->HasCtlr()) {
            return;
        }
        {
            int NewCtlr = pCurLs->GetFullCtlr(0,0);
            int CurDL = GetCurDL(); 
            // Consider both the positive and negative forms of the gate variable.
            for (int v=0; v<=1; v++) {
                int NewCtlrL = NewCtlr ^ v;
                foreach (itGs, FullCtlrToGSes[glit^v]) {
                    if (GloDbgLev >= 2007) printf("Upd:  clauses[%i]->ctlr = %i\n", (*itGs)->id, NewCtlrL);
                    GsT* pLrn = *itGs;
                    int narf = pLrn->GetFullCtlr(0,0);
                    assert(narf == NewCtlrL);
                    FullCtlrUndo[CurDL].append(GsIntPair(*itGs, (*itGs)->ctlr));
                    (*itGs)->ctlr = NewCtlrL; 
                    FullCtlrToGSes[NewCtlrL].insert(*itGs);
                    SemiCtlrProp(*itGs, NewCtlrL, 0, NoUpdUndo);
                    PendingExecs[FORC_GST_PRI].append(AntecedT(*itGs, NewCtlrL));
                }
            }
        }
        for (int v=0; v<=1; v++) {
            /*static garr<GsT*> FrobGSes;
            assert(FrobGSes.n == 0);
            foreach (itGs, SemiCtlrToGSes[glit^v]) {
                FrobGSes.append(*itGs);
            }*/
            foreach (itGs, SemiCtlrToGSes[glit^v]) {
                int SubCtlr = glit^v; 
                GsT* pLrn = itGs->first;
                if ((pLrn)->ctlrs.count(SubCtlr)==0) continue;
                // TODO: Don't recompute the whole thing, just incremental.
                if (0) {
                    int tmpFullCtlr = pLrn->GetFullCtlr(0,0);
                    SemiCtlrProp(pLrn, tmpFullCtlr, 0, NoUpdUndo);
                } else {
                    SemiCtlrProp(pLrn, SubCtlr, 1, NoUpdUndo);
                    //assert((pLrn)->ctlrs.count(SubCtlr)==0);
                }
            }
            //FrobGSes.n = 0;
        }
    }
}

/*****************************************************************************
* Stuff  }}}{{{1
*****************************************************************************/

int AsgnT::Backtrack() {
    LastPickGate = 0;
    GloStep++;
    NumStepBt++;
    while (!PendingDeadPars.empty()) {
        PendingDeadPars.pop();
    }
    PendingMonotoneLits.clear();
    int UndoChlit = GetLastChlit();
    uint CurDL = LitToDL[UndoChlit];
    if (GloDbgLev >= 2100) printf("Popping %i (DL %i) (step %i)\n", decv(UndoChlit), CurDL, GloStep);
    assert(CurDL != ZERO_DECLEV);

    int UndoDL = LitToDL[UndoChlit];
    garri& CurUndoList = UndoListByDL[UndoDL];
    //CurUndoList.append(UndoChlit);
    
    (UndoChlit == chlits.pop()) || die();

    bool qPrintRemoved = (GloDbgLev >= 2110);
    if (qPrintRemoved) printf("Removing [ ");
    for (int i = CurUndoList.n - 1; i >= 0; i--) {
        int flit = CurUndoList[i];
        if (qPrintRemoved) printf("%i ", decv(flit));
        if (LitToDL[flit] != CurDL) {assert(false);}
        if (flit != UndoChlit) {
            if (DepByLit[flit].IsUnitProp()) assert(!GetDepByLit(flit)->IsBlocked());
        }
        //FreeVars.insert(absv(flit));
        LitToDL[flit] = NO_DECLEV;
        LitToChrono[flit] = ((ChronoT) NO_DECLEV);
        CurChrono--;
        DepByLit[flit] = EmptyAnteced;
    }

    for (int i=0; i <= MAX_GST_PRI; i++) {
        PendingExecs[i].clear();
    }
    if (qPrintRemoved) printf("]\n");
    assert(HasTargVar() == false);
    pGloAsgn->FreeFmla = NOT_DONE;
    LitsInConflict.clear();
    UndoListByDL[UndoDL].n = 0;
    
    for (int i = SemiCtlrUndo[UndoDL].n-1; i >= 0; i--) {
        GsT* pFixGs = SemiCtlrUndo[UndoDL][i].first;
        int LitAdj = SemiCtlrUndo[UndoDL][i].second;
        if (GloDbgLev >= 2960 && pFixGs->id == 998) {
            printf("Upd clause[%i]: %i\n", pFixGs->id, LitAdj);
        }
        if (LitAdj < 0) {
            LitAdj = -LitAdj;
            pFixGs->ctlrs.erase(LitAdj);
            SemiCtlrToGSes[LitAdj].erase(pFixGs);
        } else {
            pFixGs->ctlrs.insert(LitAdj);
            assert(SemiCtlrToGSes[LitAdj].count(pFixGs) == 1);
            //SemiCtlrToGSes[LitAdj].insert(pFixGs);
        }
    }
    SemiCtlrUndo[UndoDL].n = 0;

    set<GsT*> BadGsSet = set<GsT*>();
    for (int i = FullCtlrUndo[UndoDL].n-1; i >= 0; i--) {
        GsT* pFixGs = FullCtlrUndo[UndoDL][i].first;
        int LowerCtlr = pFixGs->ctlr;
        int UpperCtlr = FullCtlrUndo[UndoDL][i].second;
        if (pFixGs->OrigOrLrn==LRN_TAG && LowerCtlr != UpperCtlr) {
            FullCtlrToGSes[pFixGs->ctlr].erase(pFixGs);
        }
        pFixGs->ctlr = UpperCtlr;
        if (UpperCtlr == -1) {
            assert(0);
            //pFixGs->ctlr = 0;
            //CtlrProp(pFixGs, true);
        }
        if (GloSanChk >= 1) {
            if (pFixGs->ctlr != 0) {
                assert(FullCtlrToGSes[pFixGs->ctlr].count(pFixGs) == 1);
            }
            if (pFixGs->ctlr == pFixGs->GetFullCtlr(0,0)) {
                BadGsSet.erase(pFixGs);
            } else {
                BadGsSet.insert(pFixGs);
            }
        }
        if (GloDbgLev >= 2007) printf("Undo: clauses[%i]->ctlr = %i (prev %i)\n",
            pFixGs->id, pFixGs->ctlr, LowerCtlr);
    }
    foreach (itGs, BadGsSet) {
        GsT* pFixGs = *itGs;
        printf("pGloProb->clauses.arr[%i] : %i != %i\n",
            pFixGs->id, pFixGs->ctlr, pFixGs->GetFullCtlr(0,0));
    }
    assert(BadGsSet.empty());
    FullCtlrUndo[UndoDL].n = 0;

    /*while (!InCurQuantPfx(UndoChlit)) {
        if (TargStack.n == 0) {assert(0); break;}
        TargGate = TargStack.PopLast();
    }*/
    if (UndoTarg[UndoDL] != 0) {
        TargGate = UndoTarg[UndoDL];
        UndoTarg[UndoDL] = 0;
    }
    if (GetLastChlit() == 0) assert(TargGate == pGloProb->OutGlit);

    /*if (true) {
        for (int i=0; i < DoneLitSets.n; i++) {
            GsT* pCurLs = DoneLitSets[i];
            assert(!pCurLs->HasForcedMove());
        }
    }*/
    return UndoChlit;
}


void AsgnT::AddUndoLit(int chlit, int flit) {
    int UndoDL = LitToDL[chlit];
    //if (GloSanChk > 0) assert(UndoListByDL[UndoDL].indexof(flit) == -1);
    UndoListByDL[UndoDL].append(flit);
    LitToChrono[flit] = CurChrono;
    ChronoToLit[CurChrono] = flit;
}

void AsgnT::AppendChlit(int NewChlit, bool IsRedo=false) {
    CurChrono++;
    NumAppends++;
    assert(!HasLit(negv(NewChlit)));
    assert(!HasLit(NewChlit));
    assert(DepByLit[NewChlit].pDepGs == NULL && DepByLit[NewChlit].AltType==0);
    //assert(chlits.flat.n > 0 && QbFromLit[NewChlit] >= QbFromLit[GetLastChlit()]);
    uint declev = chlits.flat.n;
    if (GloDbgLev >= 2005 && !IsRedo)  printf("Chose Lit %i (at DL #%i)\n", decv(NewChlit), declev);
    chlits.append(NewChlit);
    //FreeVars.erase(absv(NewChlit));
    LitToDL[NewChlit] = declev;
    LitToChrono[NewChlit] = CurChrono;
    LitToAbsTime[NewChlit] = NumAppends;
    ChronoToLit[CurChrono] = NewChlit;
    UndoListByDL[declev].n = 0;
    UndoListByDL[declev].append(NewChlit);
    assert(ChlitByDL(declev) == NewChlit);
    LastHitDL[declev] = *pNumClauses;
}

int StopPoint;

void AsgnT::AppendForcedLit(int lit, GsT* pRef) {
    CurChrono++;
    NumAppends++;
    /*if (StopPoint == NumAppends) {
        stop();
    }*/
    uint NewDL = GetCurDL();
    int UndoChlit = ChlitByDL(NewDL);
    assert(UndoChlit == GetLastChlit());
    if (!IsPhant(lit)) assert(!HasLit(negv(lit)));
    //if (IsPhant(lit) && NoUnivGhost && PlyFromLit(lit)==0) assert(0);
    assert(!HasLit(lit));
    //assert(pRef->HasLit(negv(lit)) || pRef->HasLit(lit) || pRef->GateLit==lit);
    DepByLit[lit].pDepGs = pRef;
    DepByLit[lit].AltType = 0;
    AddUndoLit(UndoChlit, lit);
    //FreeVars.erase(absv(lit));
    LitToDL[lit] = LitToDL[UndoChlit];
    LitToAbsTime[lit] = NumAppends;
    //if (NumAppends == 155259) {
    //    stop();
    //}
    if (UseMonotone && IsGateLit(lit)) {
        if (GloDbgLev >= 2012) printf("NARF %i\n", lit);
        FixChildrenWatches(lit, 1);
        FixChildrenWatches(negv(lit), 1);
        while (!PendingDeadPars.empty()) {
            int CurLit = PendingDeadPars.top();
            PendingDeadPars.pop();
            FixChildrenWatches(CurLit, 1);
        }
    }
    if (GloDbgLev >= 2051) {
        int doit = (GloDbgLev >= 2050 || IsInputLit(lit));
        if (doit) {
            printf("Appending literal %i: forced at DL %i (chlit %i) ", 
                    decv(lit), LitToDL[UndoChlit], decv(UndoChlit));
            if (pRef) {printf("by clause[%i].\n", pRef->id);}
            else {printf("by something other than unit prop.\n");}
        }
    }
    if (IsPhant(lit) && HasLit(negv(lit))) {
        int losr = PlyFromLit(lit);
        int BatVar = LitToBat[Unphantomize(lit)];
        int BatLit = BatVar ^ losr;
        if (!HasLit(BatLit)) {
            AppendForcedLit(BatLit, NULL);
            DepByLit[BatLit].AltType = CONTRADICT_ANTECED;
            DepByLit[BatLit].AltLit = lit;
        }
    }
}



/*****************************************************************************
* Monotone literal rule.  }}}{{{1
*****************************************************************************/

#define NO_ALIVE_CACHE 0
#define USE_ALIVE_CACHE 1

void AsgnT::WatchParSanityChk() {
    for (int lit=2; lit <= pGloProb->LastInpLit; lit++) {
        PickParent(lit, 1);
    }
}

void AsgnT::PickParent(int lit, int ChkOnly=0) {
    if (ChkOnly && (HasLit(lit) || HasLit(negv(lit)))) return;
    if (ChkOnly==1 && LastChkPar[lit]==NumAppends) return;
    if (ChkOnly) LastChkPar[lit] = NumAppends;
    garri& poss = pGloProb->ParentLit[lit];
    int OldPar = LitToLiveParent[lit];
    if (IsAlive(OldPar)) {
        return;
    }
    for (int i=0; i < poss.n; i++) {
        int CurPar = poss[i];
        if (ChkOnly) {PickParent(CurPar, ChkOnly);}
        if (IsAlive(CurPar)) {
            if (ChkOnly) {
                if (LitToLiveParent[lit] != CurPar) {printf("...\n");}
                assert(LitToLiveParent[lit] == CurPar);
                return;
            }
            LitToLiveParent[lit] = CurPar;
            ParentToWatchers[OldPar].erase(lit);
            ParentToWatchers[CurPar].insert(lit);
            return;
        }
    }
}

bool AsgnT::IsAlive(int lit) {
    if (lit == pGloProb->OutGlit) return true;
    if (HasLit(lit) || HasLit(negv(lit))) return false;
    LvCacheEntry& cache = LvCache[lit];
    if (true || cache.timestamp < NumMonoProps) {
        if (GloSanChk >= 4) PickParent(LitToLiveParent[lit], 1);
        cache.IsAlive = IsAlive(LitToLiveParent[lit]);
        cache.timestamp = NumMonoProps;
    } else {
        assert(cache.IsAlive == IsAlive(LitToLiveParent[lit]));
    }
    return cache.IsAlive;
}

int NumChFix = 0;

void AsgnT::FixChildrenWatches(int ParLit, bool AndProp) {
    if (IsInputLit(ParLit)) {return;}
    NumChFix++;
    int PrintFix = (GloDbgLev >= 2012);
    if (PrintFix) printf("FixChild [%5i]: %i: ", NumChFix, ParLit);
    const int n = ParentToWatchers[ParLit].size();
    garri FrobLits = garri(n, n);
    int i=0;
    foreach (iter, ParentToWatchers[ParLit]) {FrobLits[i++] = (*iter);}
    assert(n == i);
    for (i=0; i < n; i++) {
        int CurLit = FrobLits[i];
        if (PrintFix) printf(" %i", CurLit);
        if (HasLit(CurLit) || HasLit(negv(CurLit))) continue;
        PickParent(CurLit);
        if (PrintFix) printf("(%i)", LitToLiveParent[CurLit]);
        if (!IsAlive(CurLit)) {
            if (IsGateLit(CurLit)) {
                PendingDeadPars.push(CurLit);
                //FixChildrenWatches(CurLit, AndProp);
            } else {
                assert(IsInputLit(CurLit));
                if (AndProp) { //  && IsAlive(negv(CurLit))
                    if (GloSanChk >= 3) PickParent(CurLit, 1);
                    //PickParent(negv(CurLit), 1);
                    int flit = (QtypeFromLit(CurLit)=='E' ? negv(CurLit) : CurLit);
                    AddPendingMonot(flit);//.second ? printf("YEAH!\n") : 0;
                    //PendingMonotoneLits.insert(flit);//.second ? printf("YEAH!\n") : 0;
                }
            }
        }
    }
    if (PrintFix) printf("\n");
    FrobLits.dealloc();
}


/*****************************************************************************
* Stuff  }}}{{{1
*****************************************************************************/

int AsgnT::ChooseLitByTree(int qb, int gate, int* pNumAttempts) {
    const int MAX_ATTEMPTS = 20;
    int qtype = GloQtype(qb);
    int ply = (qtype=='E' ? 1 : (qtype=='A' ? 0 : die()));
    if (qb != QbFromLit[GetLastChlit()]) {
        LastPickGate = 0;
    }
    if (HasLit(LastPickGate) || HasLit(negv(LastPickGate))) {
        LastPickGate = 0;
    }
    if (gate == 0) {
        if (LastPickGate != 0) {gate = LastPickGate;}
        else {gate = TargGate ^ (1 - ply);}
    }
    int IsOr = !IsAndGate(gate);
    LastPickGate = 0;
    GsT* pCurLs = AbsGlitToGs(gate);
    if (pCurLs->pChoices == 0) {
        pCurLs->pChoices = new garri(0, pCurLs->ReqLits.n);
    }
    garri& choices = *pCurLs->pChoices;
    choices.n = 0;
    int suic = rand() % 2;
    foreach (itTrig, pCurLs->TrigAE[ply]) {
        int CurLit = *itTrig ^ IsOr;
        if (QbFromLit[CurLit] > qb) break;
        if (HasLit(CurLit) || HasLit(negv(CurLit))) continue;
        if (suic) CurLit = negv(CurLit);
        choices.append(CurLit);
    }
    if (IsAndGate(gate) && choices.n >= 2) {
        LastPickGate = gate;
        int PickLit = choices[rand() % choices.n];
        return PickLit;
    }
    foreach (itTrig, pCurLs->TrigG) {
        int CurLit = *itTrig ^ IsOr;
        if (HasLit(CurLit) || HasLit(negv(CurLit))) continue;
        if (suic) CurLit = negv(CurLit);
        choices.append(CurLit);
    }
    while (1) {
        if (choices.n == 0) return 0;
        int r = rand() % (2*choices.n);
        int ixChoice = (r >> 1);
        bool GiveUp = (r & 1) && (absv(gate) != absv(TargGate));
        int PickLit = choices.RemoveAndFill(ixChoice);
        if (IsGateLit(PickLit)) {
            PickLit = ChooseLitByTree(qb, PickLit, pNumAttempts);
        } else {
            return PickLit;
        }
        if (PickLit != 0) {
            return PickLit;
        }
        else {
            if (GiveUp) return 0;
            if ((*pNumAttempts)++ >= MAX_ATTEMPTS) return 0;
        }
    }
}

    
int AsgnT::ChooseRandLit(int preferred) {
    static garri choices;
    choices.n = 0;
    int MinQb = (1 << 30);
    int MinLit = 0;
    QuantPfxT CurPfx = QuantPfx[absv(TargGate)];
    if (CurPfx.blks.n == 0) {
        return 0;
    }
    NumDecs++;
    int StartLit = CurPfx.blks[0].lits[0];
    int PrevQb = QbFromLit[GetLastChlit()];
    if (QbFromLit[StartLit] <= PrevQb && PrevQb <= CurPfx.blks.last().ix) {
        StartLit = pGloProb->QuantBlk[PrevQb].lits[0];
    }
    for (int lit=StartLit; lit <= pGloProb->LastInpLit; lit+=2) {
        if (LitToDL[lit] == NO_DECLEV && LitToDL[negv(lit)] == NO_DECLEV) {
            MinLit = lit;
            MinQb = QbFromLit[lit];
            break;
        }
    }
    if (MinQb > CurPfx.blks.last().ix) {
        if (!(HasLit(TargGate) || HasLit(negv(TargGate)))) {
            //PickNewTarg(); // for debugging
        }
        return 0;
    }
    if (MinQb == (1<<30)) {die("Internal error: No free literals remaining?!\n");}

    int TreeLit = 0;

    if (LastPickGate != 0 || (rand() % 256) >= 128) {
        //if (LastPickGate) stop();
        int tmp = 0;
        TreeLit = ChooseLitByTree(MinQb, 0, &tmp);
        if (TreeLit != 0) {
            AppendChlit(TreeLit);
            return TreeLit;
        }
    }

    for (int lit=MinLit; lit <= pGloProb->LastInpLit; lit+=2) {
        if (LitToDL[lit] == NO_DECLEV && LitToDL[negv(lit)] == NO_DECLEV) {
            if (QbFromLit[lit] != MinQb) break;
            if (lit == preferred) {
                AppendChlit(preferred);
                return preferred;
            }
            choices.append(lit);
            choices.append(negv(lit));
        }
    }
    if (TreeLit != 0) {
        //assert(choices.indexof(TreeLit) != -1);
        //return TreeLit;
    }


    //printf("Choices: "); DumpLitsEnc(choices);
    int x = rand();
    if (choices.n == 0) {printf("\nERROR: NO CHOICES!  (NumBt = %i)\n", NumBigBt);}
    int chosen = choices[x % choices.n];
    //printf("Choosing %i.\n", decv(chosen));
    AppendChlit(chosen);
    return chosen;
}




void AsgnT::dump(int GatesToo) {
    printf("Lits:\n");
    garri& blk = chlits.flat;
    int nch = 0;
    for (int i=0; i < blk.n; i++) {
        int chlit = blk[i];
        uint dl = LitToDL[chlit];
        nch = printf("    %4i (QB %i%c, DL %2i): ", decv(chlit), QbFromLit[chlit], tolower(QtypeFromLit(chlit)), dl);

        garri& CurUndo = UndoListByDL[LitToDL[chlit]];
        if (CurUndo.n == 0) {
            printf("(none)\n");
        } else {
            const char* sep = "";
            printf("[");
            foreach (it, CurUndo) {
                if (!GatesToo && IsGateLit(*it)) continue;
                nch += printf("%s%i", sep, decv(*it));
                if (nch > LINE_LEN - 7) {
                    if (nch < LINE_LEN) printf("\n");
                    nch = 0;
                }
                sep = " ";
            }
            printf("]\n");
        }
    }
}

void AsgnT::DumpChlits() {
    garri& blk = chlits.flat;
    printf("[ ");
    for (int i=0; i < blk.n; i++) {
        int chlit = blk[i];
        if (chlit != 0) printf("%i ", decv(chlit));
    }
    printf("]\n");
}

void AsgnT::DumpLits(int narf) {
    garri& blk = chlits.flat;
    printf("[ ");
    for (int i=0; i < blk.n; i++) {
        int chlit = blk[i];
        printf("[ ");
        garri& CurUndo = UndoListByDL[LitToDL[chlit]];
        foreach (it, CurUndo) {
            printf("%i ", *it);
        }
        printf("] ");
    }
    printf("]\n");
}


void AsgnT::restart() {
    // TODO eventually: optimize this; don't need to pop each level individually.
    while (chlits.flat.n > 1) {
        Backtrack();
    }
}



/*****************************************************************************
* Sanity Check for WatchedLits.    }}}{{{1
*****************************************************************************/
void AsgnT::WatchSanityChk1(GsT* pCurLs) {
        int FreeGates[4] = {0,0,0,0};
        int NewAE[4] = {0,0,0,0};
        int* OldAE = pCurLs->WatchRawAE;
        //PickWatchedLitsG(pCurLs, FreeGates);
        PickWatchedLitsAE(pCurLs, 0, NewAE, FreeGates);
        PickWatchedLitsAE(pCurLs, 1, NewAE+2, FreeGates);
        for (int j=0; j < 4; j++) {
            if (NewAE[j] == 0) NewAE[j] = OldAE[j];
        }
        for (int ply=0; ply <= 1; ply++) {
            if (!HasLitOrPhant(NewAE[ply*2+0],ply) && !HasLitOrPhant(NewAE[ply*2+1], ply)) continue;
            if ( HasLitOrPhant(NewAE[ply*2+0],ply) &&  HasLitOrPhant(NewAE[ply*2+1], ply)) continue;
            assert(NewAE[ply*2 + 0] == OldAE[ply*2 + 0]);
            assert(NewAE[ply*2 + 1] == OldAE[ply*2 + 1]);
        }
        if (pCurLs->OrigOrLrn==LRN_TAG) {
            int OldRes = pCurLs->WatchedRes;
            int NewRes = PickNewResWatch(pCurLs);
            if (NewRes != 0 && OldRes != NewRes) {
                int losr = 1 - pCurLs->FreeFmla;
                int LastTrig = NewAE[losr*2 + 0];
                if (IsInputLit(LastTrig)) {
                    assert(QbFromLit[OldRes] < QbFromLit[LastTrig]);
                }
                if (!HasLit(NewRes)) assert(!HasLit(OldRes));
            }
            if (FindFullCtlrs && !HasLit(negv(pCurLs->FirstGateWatch()))) {
                assert(pCurLs->ctlr == pCurLs->GetFullCtlr(0,0));
                if (pCurLs->ctlr != 0 /*&& i < pGloProb->NumDoneClauses*/) {
                    assert(FullCtlrToGSes[pCurLs->ctlr].count(pCurLs) == 1);
                    map<LitT,LitT> NewSemis = map<LitT,LitT>();
                    if (FindSemiCtlrs) {
                        GetSemiCtlrs(pCurLs->ctlr, 0, NewSemis); 
                        foreach (it, NewSemis) {
                            assert(pCurLs->ctlrs.count(it->first) == 1);
                        }
                    }
                    //assert(pCurLs->ctlrs == NewSemis);
                }
            }
        }
}
void AsgnT::WatchSanityChk3() {
    for (int i=1; i < pGloProb->clauses.n; i++) {
        GsT* pCurLs = pGloProb->clauses[i];
        if (pCurLs->OrigOrLrn==ORIG_TAG && pCurLs->ReqLits.n == 1) continue;
        WatchSanityChk1(pCurLs);
    }
}

void AsgnT::LatentExecsChk() {
    for (int i=1; i < pGloProb->NumDoneClauses; i++) {
        GsT* pCurLs = pGloProb->clauses[i];
        int ret = ExecLitSet(pCurLs, EM_DRYRUN);
        assert(ret <= 0);
    }
}



/*****************************************************************************
* Functions for Boolean Constraint Propagation proper.    }}}{{{1
*****************************************************************************/

int AsgnT::HasDeadLits(ap<int>* pTrigs) {
    foreach (it, *pTrigs) {
        if (HasLit(negv(*it))) return true;
    }
    return false;
}

int AsgnT::IsResCompat(ap<int>& TrigLits, ap<int>& ImpLits) {
    auto(itResCur, ImpLits.begin());
    auto(itResEnd, ImpLits.end());
    foreach (itWinrLit, TrigLits) {
        int negflit = negv(*itWinrLit);
        itResCur = lower_bound(itResCur, itResEnd, negflit);
        if (itResCur == itResEnd) break;
        if (*itResCur == negflit) return false;
    }
    return true;
}

int AsgnT::InCurQuantPfx(int lit, int TestTarg=0) {
    if (TestTarg==0) TestTarg = TargGate;
    QuantPfxT CurPfx = QuantPfx[absv(TestTarg)];
    if (CurPfx.blks.n == 0) return false;
    int MinQb = CurPfx.blks[0].ix;
    int MaxQb = CurPfx.blks.last().ix;
    return (MinQb <= QbFromLit[lit] && QbFromLit[lit] <= MaxQb);
}

int AsgnT::UnderQuantPfx(int lit, int TestTarg) {
    QuantPfxT CurPfx = QuantPfx[absv(TestTarg)];
    return (CurPfx.blks[0].ix <= QbFromLit[lit]);
}


int AsgnT::ExecLitSet(GsT* pCurLs, int mode, int ply=ALL_PLY, bool SkipChks=false) {
    if (mode > EM_SCHED && pCurLs->id >= pGloProb->NumDoneClauses) return -10;
    if (HasLitOrNeg(pCurLs->GateLit)) return -20;
    if (pCurLs->TrigRipe(0)==0 && pCurLs->TrigRipe(1)==0) return -40;
    if (!SkipChks) {
        if (pCurLs->HasBlockedImps()) return -50;
        if (HasDeadLits(&pCurLs->ReqLits)) return -60;
    }
    if (ply==ALL_PLY) {
        assert(mode <= EM_SCHED);
        int ret = 0;
        for (ply=-1; ply <=1; ply++) {
            if (pCurLs->FreeFmla == ply) continue;
            int tmp = ExecLitSet(pCurLs, mode, ply, true);
            if (tmp > 0) ret = 1;
        }
        return ret;
    }
    NumExecAttempts++;
    int IsDry = (mode == EM_DRYRUN);
    int GateForce = 0;
    if (!AltMeth1 && pCurLs->HasAllWatchedTrigs() && GetFirstFreeLit(pCurLs->ReqLits)==0) {
        GateForce = pCurLs->GateLit;
        if (HasLit(GateForce)) {return -__LINE__;}
        if (ply != -1) {return -__LINE__;}
        if (mode <= EM_SCHED) {
            if (!IsDry) pGloAsgn->AddPendingLs(pCurLs, DEFN_INS_GST_PRI);
            return GateForce;
        }
        return GateForce; 
    } else if (UsePhant) {
        if (ply != -1 && HasLitOrPhant(pCurLs->WatchAE(ply, 0), ply) && 
                GetFirstFreeLit(pCurLs->TrigAE[1-ply])==0) {
            GateForce = Phantomize(pCurLs->GateLit, ply);
            /*if (pCurLs->GateLit == TargGate && signv(pCurLs->GateLit) != (1^ply)) {
                GateForce = pCurLs->GateLit;
            }*/
            if (HasLit(GateForce)) {return 0;}
            if (HasLit(Unphantomize(GateForce))) {abort();}
            if (NoUnivGhost && ply==0) return 0;
            if (mode <= EM_SCHED) {
                if (!IsDry) pGloAsgn->AddPendingLs(pCurLs, DEFN_INS_GST_PRI);
                return GateForce;
            }
            return GateForce; 
        }
    }
    if (mode > EM_SCHED && !PendingExecs[DEFN_GST_PRI].is_empty()) {
        pGloAsgn->AddPendingLs(pCurLs, FORC_GST_PRI);
        return (GateForce != 0);
    }
    if (!PendingExecs[MONO_GST_PRI].is_empty()) abort();


    //if (UsePhant || pCurLs->OrigOrLrn == LRN_TAG) 
    return ExecForced(pCurLs, mode, ply);


    if (UsePhant) return -1000;
    else abort();

}


int AsgnT::ExecForced(GsT* pCurLs, int mode, int ply) {

        int IsDry = (mode == EM_DRYRUN);
        
        int winr, losr, BatLit;
        if (ply == -1) {
            if (pCurLs->OrigOrLrn != LRN_TAG) return -__LINE__;
            winr = pCurLs->FreeFmla;
            losr = (1 - pCurLs->FreeFmla);
            assert(winr == signv(pCurLs->GateLit) ^ 1);
            BatLit = absv(pCurLs->GateLit);
        } else {
            if (!UsePhant) return -__LINE__;
            if (pCurLs->OrigOrLrn == LRN_TAG) return -__LINE__;
            bool IsPhant = HasLit(Phantomize(negv(pCurLs->GateLit), ply));
            if (!IsPhant) return -__LINE__;
            losr = ply;
            winr = 1 - losr;
            BatLit = absv(LitToBat[pCurLs->GateLit]);
        }

        if (BatLit > TargGate) return -__LINE__;

        if ((pCurLs->TrigRipe(losr)==1)) {
            
            int ForcedLit = negv(pCurLs->WatchAE(losr, 0));
            if (ply != -1 && IsInputLit(ForcedLit) && QtypeNumFromLit(ForcedLit) != ply) {return -__LINE__;}
            if (IsInputLit(ForcedLit)) {
                if (!UnderQuantPfx(ForcedLit, BatLit)) {return -__LINE__;}
            } else {
                assert(!IsPhant(ForcedLit));
                if (LitToBat[ForcedLit] != BatLit) {return -__LINE__;}
            }
            if (GetUpstreamFreeLit(ForcedLit, pCurLs->TrigAE[winr]) != 0) {return -__LINE__;}
            if (GetUpstreamFreeLit(ForcedLit, pCurLs->ImpLits) != 0) {return -__LINE__;}
            //if (!UnderQuantPfx(GetFirstFreeLit(pCurLs->TrigAE[winr]), BatLit)) {abort();}
            //if (!UnderQuantPfx(GetFirstFreeLit(pCurLs->ImpLits), BatLit)) {abort();}
            if (IsGateLit(ForcedLit)) {
                if (!UsePhant) return -__LINE__;
                if (HasLit(ForcedLit)) return -__LINE__;
                ForcedLit += PhantOffset[losr];
                if (HasLit(ForcedLit)) return -__LINE__;
                if (HasLit(negv(ForcedLit))) abort(); //return -445;  // FIXME
            }
            if (NoUnivGhost && IsPhant(ForcedLit) && PlyFromLit(ForcedLit)==0) return -__LINE__;
            if (mode <= EM_SCHED) {
                if (!IsDry) pGloAsgn->AddPendingLs(pCurLs, FORC_GST_PRI); 
                return ForcedLit;
            }
            NumProp_OneStep++;
            GloNumLrnRejects++;
            return ForcedLit;

        } else if (pCurLs->TrigRipe(losr)==2) {

            return 0;
            
        } else {
            return -490;
        }
}


void AsgnT::SchedForProp(int CurLit, garri* pMultLits=NULL) {
    static garr<GsT*> CurLitWatch;  // static for optimization
    assert(CurLitWatch.n == 0);
    CurLitWatch.n = 0;
    static garri SingLit;
    SingLit.n = 0;
    SingLit.append(CurLit);
    if (pMultLits == NULL) {pMultLits = &SingLit;}
    else {assert(CurLit==0);}
    foreach (itLit, *pMultLits) {
        CurLit = *itLit;
        int OrigCurLit = CurLit;
        CurLit = Unphantomize(CurLit);
        foreach (it, pGloAsgn->LitWatch[CurLit]) {
            CurLitWatch.append((*it).second);
        }
        if (OrigCurLit != CurLit) {
            GsT* pDefGs = AbsGlitToGs(CurLit);
            CurLitWatch.append(pDefGs);
            for (int i=1; i <= pDefGs->ReqLits.n; i++) {
                GsT* pCurLs = pGloProb->clauses[pDefGs->id + i];
                assert(absv(pCurLs->GateLit) == absv(CurLit));
                CurLitWatch.append(pCurLs);
            }
        }
    }
    //init_garr_from_set(&CurLitWatch, pGloAsgn->LitWatch[CurLit]);
    //sort(CurLitWatch.begin(), CurLitWatch.end(), FnGreaterGsById);
    //sort(CurLitWatch.begin(), CurLitWatch.end(), FnLessGsById);
    int OldN = CurLitWatch.n;
    //CurLitWatch.remove_adj_dups();
    if (GloDbgLev >= 1800 && CurLitWatch.n != OldN) {
        printf("[%i] ", OldN - CurLitWatch.n);
    }
    for (int ixLs=0; ixLs < CurLitWatch.n; ixLs++) {
        GsT* pCurLs = CurLitWatch[ixLs];
        FixWatchedLits(pCurLs);
    }
    for (int ixLs=0; ixLs < CurLitWatch.n; ixLs++) {
        GsT* pCurLs = CurLitWatch[ixLs];
        if (!UsePhant) CtlrProp(pCurLs);
        ExecLitSet(pCurLs, EM_SCHED);
    }
    CurLitWatch.n = 0;
}

int AsgnT::Propagate() {
    static garr<GsIntPair> ForceList;
    static garri LitsForced;
    assert(ForceList.n == 0);

    int meth = 0;

    while (true) {
        while (true) {
            AntecedT ante = PopFirstPendingLs(false);
            if (ante == EmptyAnteced) break;
            if (ante.IsUnitProp()) {
                GsT* pCurLs = ante.pDepGs;
                if (ante.FinalCtlr != 0) {
                    abort();
                } else {
                    for (int ply=-1; ply <= 1; ply++) {
                        int flit = ExecLitSet(pCurLs, EM_WIN, ply);
                        if (flit > 0) {
                            NumProps++;
                            ForceList.append(GsIntPair(pCurLs,flit));
                            if (meth==0) {
                                pGloAsgn->AppendForcedLit(flit, pCurLs);
                                SchedForProp(flit);
                            }
                        }
                    }
                }
            } else if (ante.AltType == MONO_ANTECED) {
                int CurLit = ante.AltLit;
                if (HasLit(CurLit)) continue;
                if (LitToDL[negv(CurLit)] == GetCurDL()) {die("MONO!!!"); continue;}  // NARF !!!!!!!!!
                if (IsAlive(CurLit) ^ IsAlive(negv(CurLit))) {
                    AppendForcedLit(CurLit, NULL);
                    DepByLit[CurLit].AltType = MONO_ANTECED;
                    DepByLit[CurLit].AltLit = CurLit;
                    pGloAsgn->SchedForProp(CurLit);
                }
            } else {die();}
            if (HasTargVar()) {
                break;
            }
        }
        if (ForceList.n == 0) break;
        if (meth==1) {
            LitsForced.n = 0;
            foreach (it, ForceList) {
                GsT* pCurLs = it->first;
                int flit = it->second;
                if (HasLit(flit)) continue;
                if (HasLitOrNeg(Unphantomize(flit))) continue;
                if (HasLitOrNeg(pCurLs->GateLit)) continue;
                if (pCurLs->HasBlockedImps()) continue;
                if (HasDeadLits(&pCurLs->ReqLits)) continue;
                /*
                if (ExecLitSet(pCurLs, EM_DRYRUN, -1) != flit &&
                    ExecLitSet(pCurLs, EM_DRYRUN, 0) != flit &&
                    ExecLitSet(pCurLs, EM_DRYRUN, 1) != flit)
                {
                    continue;
                }
                */
                pGloAsgn->AppendForcedLit(flit, pCurLs);
                //SchedForProp(flit);
                LitsForced.append(flit);
                if (HasTargVar()) {break;}
            }
            SchedForProp(0, &LitsForced);
            /*foreach (it, ForceList) {
                if (HasTargVar()) {break;}
                int flit = it->second;
                SchedForProp(flit);
            }*/
        }
        if (HasTargVar()) {
            pGloAsgn->FreeFmla = (HasLit(TargGate) ? 1 : 0);
            break;
        }
        ForceList.n = 0;
    }
    //if (GloDbgLev > 1800) printf("%i ", ForceList.n);
    ForceList.n = 0;
    return 0;
}

int AsgnT::MonoProp() {
    bool flag = true;
    while (flag) {
        NumMonoProps++;
        flag = false;
        //Propagate();
        //int USE_MONOTONE_LIT = 1;
        /*if (USE_MONOTONE_LIT) {
            while (!PendingDeadPars.empty()) {
                int CurLit = PendingDeadPars.top();
                PendingDeadPars.pop();
                FixChildrenWatches(CurLit);
            }
        }*/
        /* TODO: Add the monotone literals as forced true or forced false literals. */
        /*foreach (itLit, PendingMonotoneLits) {
            AppendForcedLit(*itLit, NULL);
            DepByLit[*itLit].AltType = MONO_ANTECED;
            pGloAsgn->SchedForProp(*itLit);
        }*/
        PendingMonotoneLits.clear();
    }
    return 0;
}



/*****************************************************************************
* Stuff for learning a new game state.    }}}{{{1
*****************************************************************************/


bool DumpResols = true;
//typedef pair<ChronoT,LitT> ChronoLitPair;

struct LitsByChronoIterT {
    set<ChronoT>::iterator it;
    LitT operator*() {return pGloAsgn->ChronoToLit[*it];}
    LitsByChronoIterT& operator--() {--it; return *this;}
    LitsByChronoIterT& operator++() {++it; return *this;}
    bool operator==(LitsByChronoIterT other) {return (it == other.it);}
    bool operator!=(LitsByChronoIterT other) {return (it != other.it);}
    LitsByChronoIterT(set<ChronoT>::iterator _it) : it(_it) { }; 

};

struct LitsByChronoT {
    set<ChronoT> s;
    
    void insert(LitT lit) {s.insert(pGloAsgn->LitToChrono[lit]);}
    void erase(LitsByChronoIterT& it) {s.erase(it.it);}
    size_t size() {return s.size();}
    bool empty() {return s.empty();}
    
    LitsByChronoIterT begin() {return LitsByChronoIterT(s.begin());}
    LitsByChronoIterT end()   {return LitsByChronoIterT(s.end());}
    
    //void erase(int lit) {s.erase(ChronoLitPair(pGloAsgn->LitToChrono[lit], lit));}
    //size_t count(LitT lit) {return s.count(ChronoLitPair(pGloAsgn->LitToChrono[lit], lit));}
};

//typedef set<LitT,struct LitLessByChrono> LitSetByChronoT;
typedef LitsByChronoT LitSetByChronoT;

class TrigHeapT {
    public:
    garr<ChronoT> h;
    void insert(LitT lit) {
        int chrono = pGloAsgn->LitToChrono[lit];
        assert(chrono != NO_DECLEV);
        h.append(chrono);
        push_heap(h.begin(), h.end());
    }

    int pop() {
        assert(h.n > 0);
        ChronoT cret = h[0];
        while (h.n > 0 && h[0] == cret) {
            pop_heap(h.begin(), h.end());
            h.n--;
        }
        return pGloAsgn->ChronoToLit[cret];
    }

    int peek() {
        return pGloAsgn->ChronoToLit[h[0]];
    }

    void dealloc() {
        h.dealloc();
    }
};


class GsBuilder {
    public:
    set<LitT> ImpLits;
    TrigHeapT TrigHeap;
    garri ElimList;
    set<LitT, LitGtByChrono> ElimZero;
    set<LitT> AlreadyElim;
    int winr;
    int losr;
    int WinrLetr;
    int NumResolves;

    void init() {
        TrigHeap.h.init();
        ElimList = garri();
        ElimZero = set<LitT,LitGtByChrono>();
        AlreadyElim = set<LitT>();
    }

    void dealloc() {
        TrigHeap.dealloc();
        ElimList.dealloc();
    }

    /*void init(ap<LitT> BaseLits) {
        int n = BaseLits.n;
        TrigHeap.h.init(0, n + (n >> 2) + 64);
        foreach (itLit, BaseLits) {
            assert(pGloAsgn->HasLit(*itLit));
            TrigInsert(*itLit);
        }
    }*/
    
    void TrigInsert(int lit) {
        //if (IsGateLit(lit) && pGloProb->ParentLit[Unphantomize(negv(lit))][0] == pGloProb->OutGlit) {
        //    volnop();
        //}
        assert(pGloAsgn->HasLit(lit));
        if (IsPhant(lit) && pGloAsgn->LitToDL[lit]==0) {
            if (AlreadyElim.count(lit) != 1) {
                AlreadyElim.insert(lit);
                pGloAsgn->Resolve(*this, lit);
            }
            return;
        }
        if (QtypeFromLit(lit) == WinrLetr) {
            if (QbFromLit[lit] >= QuantPfx[pGloAsgn->TargGate].blks[0].ix) {
                ImpInsert(lit);
                return;
            }
        } else if (QtypeFromLit(lit) == ('A'+'E' - WinrLetr)) {
            //assert(pGloAsgn->IsChlit(lit) || pGloAsgn->DepByLit[lit].pDepGs != &EmptyLs);
            ///assert(pGloAsgn->HasLit(pGloAsgn->TargGate ^ 1^winr));
            // if (pGloAsgn->HasLit(BatLit ^ winr)) {
            //     volnop();
            // }
        }
        //if (GloDbgLev >= 1990) printf("PUSH %i\n", lit);
        /*if (pGloAsgn->LitToDL[lit] == 0) {
            //GsT* pDepGs = pGloAsgn->DepByLit[lit].pDepGs;
            //assert(pDepGs->ReqLits[0] == negv(lit) && pDepGs->ReqLits.n==1);
            volnop();
            ElimZero.insert(lit);
            //if (GloDbgLev >= 1950) printf("ElimZero.insert(%i)\n", lit);
        } else*/ {
            TrigHeap.insert(lit);
        }
        //assert(ImpLits.count(negv(lit)) == 0);
    }

    void ImpInsert(int lit) {
        assert(QbFromLit[lit] >= QuantPfx[pGloAsgn->TargGate].blks[0].ix);
        assert(QtypeFromLit(lit) == WinrLetr);
        ImpLits.insert(lit);
        //assert(!pGloAsgn->HasLit(negv(lit)));
    }

    LitT PopTrig() {
        if (ElimList.n > 0) abort();
        int ret = TrigHeap.pop();
        //if (GloDbgLev >= 1990) printf("POP  %i\n", ret);
        return ret;
    }

    LitT PeekTrig() {
        if (ElimList.n > 0) abort();
        int ret = TrigHeap.peek();
        return ret;
    }

    bool TrigEmpty() {
        return TrigHeap.h.n == 0;
    }

    void CopyTrigsTo(ap<int>& TrigLits) {
        int NumLits = TrigLits.n;
        assert(NumLits == TrigHeap.h.n);
        for (int i=0; i < NumLits; i++) {
            TrigLits[i] = Unphantomize(pGloAsgn->ChronoToLit[TrigHeap.h[i]]);
            //push_heap(TrigLits.begin(), &TrigLits[i+1]);
        }
        //sort_heap(TrigLits.begin(), TrigLits.end());
    }
};

bool AsgnT::HasUIP(GsBuilder& gb) {
    LitT top = gb.TrigHeap.pop();
    LitT penult = gb.TrigHeap.peek();
    gb.TrigHeap.insert(top);
    if (gb.TrigHeap.h.n == 1) return true;
    return (LitToDL[top] != LitToDL[penult]);
}

int GsT::GetLastGlit() {
    // FIXME: This is slow.
    int BestLit=0;
    ChronoT BestChrono=0;
    for (int i=0; i < TrigG.n; i++) {
        ChronoT LitChrono = pGloAsgn->LitToChrono[TrigG[i]];
        if (BestChrono < LitChrono) {
            BestChrono = LitChrono;
            BestLit = TrigG[i];
        }
    }
    return BestLit;
}

int AsgnT::GetChronoLastLit(ap<int> Lits) {
    int BestLit=0;
    for (int i=0; i < Lits.n; i++) {
        int CurLit = Lits[i];
        if (LitToChrono[BestLit] < LitToChrono[CurLit]) {
            BestLit = CurLit;
        }
    }
    return BestLit;
}

int AsgnT::GetChronoLastTwoLits(ap<int> Lits, int* p2nd) {
    int BestLit=0, PenLit=0;
    for (int i=0; i < Lits.n; i++) {
        int CurLit = Lits[i];
        if (LitToChrono[BestLit] < LitToChrono[CurLit]) {
            PenLit = BestLit;
            BestLit = CurLit;
        } else if (LitToChrono[PenLit] < LitToChrono[CurLit]) {
            PenLit = CurLit;
        }
    }
    *p2nd = PenLit;
    return BestLit;
}

void AsgnT::FinalCtlrResolve(GsBuilder& gb, int glit, int BaseLit, garri& CumTrigs, bool AllowImps) {
    GsT* pCurLs = AbsGlitToGs(glit);
    assert(CumTrigs.last() == glit);
    CumTrigs.PopLast();
    if (!IsAndGate(glit)) {
        if (PrfLog) PrintDischargeMsg("    ", glit, pCurLs, BaseLit);
        if (GloSanChk >= 2) {
            ap<LitT> trigs;
            if (IsInputLit(BaseLit)) {trigs = pCurLs->TrigAE[QtypeNumFromLit(BaseLit)];}
            else {trigs = pCurLs->TrigG;}
            assert(binary_search(trigs.begin(), trigs.end(), negv(BaseLit)) == true);
        }
        CumTrigs.append(BaseLit);
        return;
    }
    if (PrfLog) PrintDischargeMsg("    ", glit, pCurLs);
    int losr = 0;
    if (AllowImps) losr = QtypeNumFromLit(BaseLit);
    assert(pCurLs->WatchAE(losr,0) == BaseLit);
    int WinrLetr = (losr==0 ? 'E' : 'A');
    int SkipLit = BaseLit;
    int GotSkip = false;
    for (int i=0; i < pCurLs->ReqLits.n; i++) {
        int lit = pCurLs->ReqLits[i];
        if (lit == SkipLit) {GotSkip=1; continue;}
        CumTrigs.append(lit);
        if (AllowImps && QtypeFromLit(lit)==WinrLetr && QbFromLit[BaseLit] < QbFromLit[lit]) {
            gb.ImpInsert(lit);
        } else {
            assert(HasLit(lit));
            gb.TrigInsert(lit);
        }
    }
    assert(GotSkip);
    CumTrigs.append(SkipLit);
}

int AsgnT::FullCtlrResolve(GsBuilder& gb, int glit, int BaseLit, garri& CumTrigs) {
    GsT* pCurLs = AbsGlitToGs(glit);
    assert(CumTrigs.last() == glit);
    int IsOr = !IsAndGate(glit);
    if (pCurLs->IsTrigLit(BaseLit ^ IsOr)) {
        return glit;
    }
    int SubCtlr = pCurLs->GetLastGlit();
    CumTrigs.PopLast();
    if (!IsAndGate(glit)) {
        if (PrfLog) PrintDischargeMsg("    ", glit, pCurLs, SubCtlr);
        SubCtlr = negv(SubCtlr);
        CumTrigs.append(SubCtlr);
        return FullCtlrResolve(gb, SubCtlr, BaseLit, CumTrigs);
    }
    if (PrfLog) PrintDischargeMsg("    ", glit, pCurLs);
    int SkipLit = SubCtlr;
    int GotSkip = false;
    for (int i=0; i < pCurLs->ReqLits.n; i++) {
        int lit = pCurLs->ReqLits[i] ^ IsOr;
        if (lit == SkipLit) {GotSkip=1; continue;}
        CumTrigs.append(lit);
        assert(HasLit(lit));
        gb.TrigInsert(lit);
    }
    assert(GotSkip);
    CumTrigs.append(SkipLit);
    return FullCtlrResolve(gb, SubCtlr, BaseLit, CumTrigs);
}

void AsgnT::SemiCtlrResolve(GsBuilder& gb, GsT* pLrn, int FinalCtlr, int BaseLit, garri& CumTrigs) {
    int TopmostCtlr = pLrn->GetLastGlit();
    int CurCtlr = FinalCtlr;
    static garri CtlrLine;
    assert(CtlrLine.n == 0);
    while (true) {
        assert(CurCtlr != 0);
        CtlrLine.append(CurCtlr);
        if (CurCtlr == TopmostCtlr) break;
        auto(it, SemiCtlrToGSes[CurCtlr].find(pLrn));
        assert(it != SemiCtlrToGSes[CurCtlr].end());
        CurCtlr = it->second;
    }
    //for (int i=0; i < CtlrLine.n; i++)
    for (int i = CtlrLine.n-1; i >= 1; i--) {
        int UpperCtlr = CtlrLine[i];
        int LowerCtlr = CtlrLine[i-1];
        int narf = FullCtlrResolve(gb, UpperCtlr, LowerCtlr, CumTrigs);
        FinalCtlrResolve(gb, narf, LowerCtlr, CumTrigs, false);
    }
    FinalCtlrResolve(gb, FinalCtlr, BaseLit, CumTrigs, true);
    CtlrLine.n = 0;
}

void AsgnT::PrintSemiCtlrGs(int ctlr, int LrnId, int ForcedLit) {
    GsT* pLrn = pGloProb->clauses[LrnId];
    GsBuilder gb;
    garri CumTrigs;
    int FinalCtlr = DepByLit[ForcedLit].FinalCtlr;
    int CurCtlr = pLrn->GetLastGlit();
    CumTrigs.append(CurCtlr);
    SemiCtlrResolve(gb, pLrn, FinalCtlr, negv(ForcedLit), CumTrigs);
    printf("Natural order: "); DumpLitsEnc(CumTrigs, "\n");
    CumTrigs.sort(CmpLitByDL);
    printf("Sorted by DL:  "); DumpLitsEnc(CumTrigs, "\n");
    gb.dealloc();
    CumTrigs.dealloc();
}

void AsgnT::MonoResolve(GsBuilder& gb, int DeadLit, int OrigChrono) {
    assert(DeadLit != pGloProb->OutGlit);
    garri& ParLits = pGloProb->ParentLit[DeadLit];
    if (LitToLiveParent[DeadLit] == 0) return;  // DeadLit was never alive.
    if (DeadLit == negv(pGloProb->OutGlit)) return;
    foreach (itPar, ParLits) {
        int NegPar = negv(*itPar);
        if (LitToChrono[NegPar] < OrigChrono) {
            gb.TrigInsert(NegPar);
        } else if (LitToChrono[*itPar] < OrigChrono) {
            gb.TrigInsert(*itPar);
        } else {
            MonoResolve(gb, *itPar, OrigChrono);
        }
    }
}

void AsgnT::PrintDischargeMsg(char const* indent, int CurLit, GsT* pCurLs, int WithLit=0) {
    if (pCurLs->id <= LastInitGs) {
        if (absv(CurLit) == absv(pCurLs->GateLit)) {
            if (pCurLs->ReqLits.n == 1) {
                    fprintf(PrfLog, "%s(gate_disjunct,  %5i,  %5i),\n", 
                            indent, decv(CurLit), decv(pCurLs->ReqLits[0]));
            } else {
                assert(pCurLs->OrigOrLrn == ORIG_TAG);
                if (WithLit != 0) {
                    fprintf(PrfLog, "%s(gate_disjunct,  %5i,  %5i),\n", 
                            indent, decv(CurLit), decv(WithLit));
                } else {
                    fprintf(PrfLog, "%s(gate_conjuncts, %5i),\n", 
                            indent, decv(CurLit));
                }
            }
        } else {
                    assert(PlyFromLit(CurLit) == 1 - FreeFmla);
                    fprintf(PrfLog, "%s(via_phantom_gs, %5i,  defn[%i]),\n", 
                            indent, decv(Unphantomize(CurLit)), decv(pCurLs->GateLit));
        }
    }
    else {
                    fprintf(PrfLog, "%s(via_learned_gs, %5i,  LrnGS[%i]),\n", 
                            indent, decv(CurLit), pCurLs->id);
    }
}

void AsgnT::Resolve(GsBuilder& gb, int CurLit) {
    if (IsBattleLit(CurLit)) {assert(signv(CurLit) == 1^FreeFmla);}
    //if (IsInputLit(CurLit)) assert(LitToBat[CurLit] == TargGate);
    if (QtypeFromLit(CurLit) == GetWinner()) {
        assert(QbFromDL(LitToDL[CurLit]) <= QbFromLit[CurLit]);
        gb.ImpInsert(CurLit);
        return;
    }
    assert(!IsChlit(CurLit));
    GsT* pCurLs = DepByLit[CurLit].pDepGs;
    int FinalCtlr = DepByLit[CurLit].FinalCtlr;
    if (IsInputLit(CurLit)) {
        if (UsePhant && pCurLs->OrigOrLrn==LRN_TAG) assert(pGloAsgn->FreeFmla == pCurLs->FreeFmla);
        assert(absv(pCurLs->GateLit) <= absv(TargGate));
    }
    gb.NumResolves++;
    if (gb.NumResolves > pGloProb->LastPlit) {
        die("Detected an infinite loop in constructing a game-state.\n");
    }
    if (pCurLs == NULL) {
        if (DepByLit[CurLit].AltType == MONO_ANTECED) {assert(0);}
        else assert(0);
    }
    //assert(pCurLs != &EmptyLs);
    int CurCtlr = 0;
    if (FinalCtlr != 0) {
        assert(!UsePhant);
        assert(IsInputLit(CurLit));
        assert(pCurLs->OrigOrLrn==LRN_TAG);
        CurCtlr = pCurLs->GetLastGlit();
        if (PrfLog) {
            fprintf(PrfLog, "  (via_learned_gs, %5i,  MakeGS(\n", decv(CurLit));
            fprintf(PrfLog, "    (WinnerLit,      %5i),\n", decv(pCurLs->GateLit));
            if (pCurLs->GateLit == pCurLs->ReqLits[0]) {
                assert(pCurLs->ReqLits.n==1);
            } else {
                PrintDischargeMsg("    ", pCurLs->GateLit, pCurLs, CurCtlr);
            }
        }
    }
    else if (PrfLog) PrintDischargeMsg("  ", CurLit, pCurLs);
    int HasSpecial = 0;
    int ElimUnphant = Unphantomize(CurLit);
    if (ElimUnphant == pCurLs->GateLit) {HasSpecial = 1;}
    for (int i=0; i < pCurLs->ReqLits.n; i++) {
        int lit = pCurLs->ReqLits[i];
        if (lit == negv(ElimUnphant)) {HasSpecial=lit; continue;}
        if (lit == CurCtlr) {HasSpecial=lit; continue;}
        if (IsInputLit(CurLit) || IsPhant(CurLit)) {
            assert(PlyFromLit(CurLit) == gb.losr);
        }
        if (IsGateLit(lit)) {
            int plit = Phantomize(lit, gb.losr);
            if (LitToChrono[plit] < LitToChrono[lit]) {
                lit = plit;
            }
        }
        if (HasLit(lit)) {
            if (LitToChrono[lit] < LitToChrono[CurLit]) {
                gb.TrigInsert(lit);
            } else {
                assert(IsInputLit(lit));
                assert(QbFromLit[lit] > QbFromLit[CurLit]);
                assert(PlyFromLit(lit) == gb.winr);
                gb.ImpInsert(lit);
            }
        } else {
            assert(IsInputLit(lit));
            assert(QbFromLit[lit] > QbFromLit[CurLit]);
            assert(PlyFromLit(lit) == gb.winr);
            gb.ImpInsert(lit);
        }

    }
    assert(HasSpecial);
    if (pCurLs->OrigOrLrn==ORIG_TAG && ElimUnphant != pCurLs->GateLit) {
        int PhantGate = negv(Phantomize(pCurLs->GateLit, PlyFromLit(CurLit)));
        assert(HasLit(PhantGate));
        // if (LitToDL[PhantGate]==0) {
        //     Resolve(gb, PhantGate);
        // } else {
             gb.TrigInsert(PhantGate);
        // }
    }
    int SwordIsGlowing = (pCurLs->OrigOrLrn==LRN_TAG && pGloAsgn->FreeFmla != pCurLs->FreeFmla);
    if (UsePhant) assert(!SwordIsGlowing);
    if (FinalCtlr != 0) {
        assert(!UsePhant);
        assert(HasSpecial == CurCtlr);
        if (DumpResols && GloDbgLev >= 2007) {
            if (NeedNewl) {printf("\n"); NeedNewl=false;}
            printf("Final Ctlr %i via %i: forced %i\n", FinalCtlr, CurCtlr, CurLit);
            /*printf("Before: ");
            DumpIntSet(&gb.ConflictAsgn, ", ");
            DumpIntSet(&gb.ImpLits, "\n");*/
        }
        static garri CumTrigs;  // Cumulative triggering lits.
        assert(CumTrigs.n == 0);
        CumTrigs.append(CurCtlr);
        int BaseLit = negv(CurLit);
        if (SwordIsGlowing) {
            BaseLit = CurLit;
        }
        if (FindSemiCtlrs) {
            SemiCtlrResolve(gb, pCurLs, FinalCtlr, BaseLit, CumTrigs);
        } else {
            int narf = FullCtlrResolve(gb, CurCtlr, BaseLit, CumTrigs);
            assert(narf == FinalCtlr);
            FinalCtlrResolve(gb, FinalCtlr, BaseLit, CumTrigs, true);
        }
        if (PrfLog) fprintf(PrfLog, "  )),\n");
        if (SwordIsGlowing) {
            volnop();
        }
        CumTrigs.n = 0;
    }
    for (int i=0; i < pCurLs->ImpLits.n; i++) {
        int CurImp = pCurLs->ImpLits[i];
        if (!SwordIsGlowing) {
            gb.ImpInsert(CurImp);
        }
    }
}

int NumRedunCacheHits;

#define SOME_ELIMINATED true

/*****************************************************************************
* MakeLearnedGs: Learns a new game state from the current assignment.    }}}{{{1
*****************************************************************************/
GsT* AsgnT::MakeLearnedGs(garr<GsT*>& MoreLrn, GsT* ChosenGs=NULL) {
    GsBuilder gb = GsBuilder();
    gb.NumResolves = 0;
    gb.winr = FreeFmla;
    gb.losr = 1 - FreeFmla;
    gb.WinrLetr = GetWinner();
    assert(pGloAsgn->HasTargVar());

    /*if (1) {
        int GotOne = false;
        if (GloDbgLev >= 2012 && NumDoneGSes > 1) {
            for (int i=0; i < DoneLitSets.n; i++) {
                GsT* pCurLs = DoneLitSets[i];
                if (pCurLs->id > LastInitGs) GotOne = true;
                if (pCurLs == NULL) continue;
                printf("\n DoneGS[%i]: pGloProb->clauses.arr[%i], ctlr=%i\n", 
                    i, pCurLs->id, pCurLs->ctlr);
                pCurLs->DumpBrief();
            }
            if (GotOne) {
                volnop();
            }
        }
    }*/

    if (PrfLog) {
        fprintf(PrfLog, "LrnGS[%i] = MakeGS(\n", *pNumClauses);
    }
    AntecedT* pInitAnteced = &DepByLit[HasLit(TargGate) ? TargGate : negv(TargGate)];
    GsT* pInitGs = pInitAnteced->pDepGs;
    {
        assert(HasTargVar());
        if (ChosenGs) {assert(0); pInitGs = ChosenGs;}
        int WinrLit = (TargGate ^ gb.losr);
        if (PrfLog) {fprintf(PrfLog, "  (WinnerLit,      %5i),\n", decv(WinrLit));}
        if (pInitGs != NULL) {
            assert(pInitGs->HasAllWatchedTrigs());
            LitsInConflict.insert(GetChronoLastLit(pInitGs->ReqLits));
            assert(pInitGs->GateLit == WinrLit);
            Resolve(gb, WinrLit);
            /*if (PrfLog) {PrintDischargeMsg("  ", pInitGs->GateLit, pInitGs);}
            gb.init(pInitGs->ReqLits);
            for (int i=0; i < pInitGs->ImpLits.n; i++) {
                gb.ImpInsert(pInitGs->ImpLits[i]);
            }*/
        } else if (pInitAnteced->AltType == CONTRADICT_ANTECED) {
            int ContraLit2 = pInitAnteced->AltLit;
            //int BaseContra2 = Unphantomize(ContraLit2);
            assert(PlyFromLit(ContraLit2) == gb.losr);
            int ContraLit1 = negv(ContraLit2);
            assert(HasLit(ContraLit1));
            LitsInConflict.insert(ContraLit2);
            LitsInConflict.insert(ContraLit1);
            if (PrfLog) {
                fprintf(PrfLog, "%s(contradiction,  %5i,  %5i),\n", 
                        "  ", decv(ContraLit2), decv(negv(ContraLit2)));
            }
            if (Unphantomize(ContraLit1) != (TargGate ^ gb.winr)) {
                gb.TrigInsert(ContraLit2);
                gb.TrigInsert(ContraLit1);
            } else {
                Resolve(gb, ContraLit2);
            }
        }
    }

    if (DumpResols && GloDbgLev >= 2010) {
        if (GloDbgLev >= 2015) {printf("Lits: "); DumpLits(0);}
        //DumpLitsEnc(gb.ConflictAsgn, "\n");
    }

    bool AlreadyLearnedUIP = false;
    //bool AlreadyLearnedUIP3 = false;
    bool AlreadyLearnedSuic = false;
    int CurLit=0;
    int elig = true; //false;
    int SuicElig = false;
    int winner = GetWinner();
    /*if (*pNumClauses == 7227) { // 008-020-03-04-001.ncf.cqbf --seed 1922870488
        volnop();
    }*/
    //if (!gb.TrigEmpty()) assert(!IsChlit(gb.PeekTrig()));
    while (true) {
        while (gb.ElimList.n > 0) {
            Resolve(gb, gb.ElimList.PopLast());
        }
        if (gb.TrigEmpty()) break;
        CurLit = gb.PeekTrig();
        int QtypeCurDL = GloQtype(QbFromDL(LitToDL[CurLit]));
        int ok;
        if (TargGate==pGloProb->OutGlit || QbFromDL(LitToDL[CurLit]) >= QuantPfx[TargGate].blks[0].ix) {
        //if (IsGateLit(CurLit) || QbFromLit[CurLit] >= QuantPfx[pGloAsgn->TargGate].blks[0].ix) 
            ok = (QtypeFromLit(CurLit) != winner && QtypeCurDL != winner) && HasUIP(gb);
        } else {
            if (IsInputLit(CurLit)) {
                break;
            }
            ok = false;
            //if (QbFromLit[CurLit] < QuantPfx[pGloAsgn->TargGate].blks[0].ix) break;
        }
        if (ok && elig) {
            /*while (!gb.ElimZero.empty()) {
                int CurElim = *gb.ElimZero.begin();
                gb.ElimZero.erase(gb.ElimZero.begin());
                //int foo = gb.ImpLits.size();
                Resolve(gb, CurElim);
                //assert(foo == gb.ImpLits.size());
                //assert(CurLit == gb.PeekTrig());
            }*/
            foreach (itImp, gb.ImpLits) {
                if (QbFromLit[CurLit] < QbFromLit[*itImp]) break;
                if (LitToDL[CurLit] <= LitToDL[*itImp]) goto TryAgainUIP;
            }
            if (IsChlit(CurLit)) break;
            if (!AlreadyLearnedUIP && LitsInConflict.count(CurLit)==0) {
                if ((0 && AltMeth1 && UsePhant) || IsInputLit(CurLit)) {
                    MoreLrn[LRN_TYPE_2] = CreateGsFromGb(gb);
                    MoreLrn[LRN_TYPE_2]->LrnType = LRN_TYPE_2;
                    AlreadyLearnedUIP = true;
                    AlreadyLearnedSuic = true;
                    SuicElig = false;
                    if (PrfLog) {
                        int WinrLit = (TargGate ^ gb.losr);
                        fprintf(PrfLog, ");\n");
                        fprintf(PrfLog, "LrnGS[%i] = MakeGS(\n", *pNumClauses+1);
                        fprintf(PrfLog, "  (WinnerLit,      %5i),\n", decv(WinrLit));
                        fprintf(PrfLog, "  (via_learned_gs, %5i,  LrnGS[%i]),\n", 
                                decv(WinrLit), *pNumClauses);
                    }
                    continue;
                }
            }
            TryAgainUIP:;
        }
        elig = true;
        /*if (QtypeCurDL != winner && SuicElig && UseLrnType4 && LitToChrono[CurLit] < DeathChrono) {
            MoreLrn[LRN_TYPE_4] = CreateGsFromGb(gb);
            MoreLrn[LRN_TYPE_4]->LrnType = LRN_TYPE_4;
            AlreadyLearnedSuic = true;
            SuicElig = false;
            continue;
        }*/
        SuicElig = (!AlreadyLearnedSuic);
        gb.PopTrig();
        Resolve(gb, CurLit);
    }
    if (PrfLog) {
        fprintf(PrfLog, ");\n");
    }

    /*if (GloDbgLev >= 2009) {
        printf(" Narf  learned GS:      ");
        DumpLitsEnc(gb.ConflictAsgn, "");
        printf(" %i\n", CurLit);
    }*/

    //ElimWinrGates(gb);
    if (0 && AltMeth1 && MoreLrn[LRN_TYPE_2] != NULL) {
        MoreLrn[LRN_TYPE_1] = MoreLrn[LRN_TYPE_2];
        MoreLrn[LRN_TYPE_2] = NULL;
        return MoreLrn[LRN_TYPE_1];
    }

    MoreLrn[LRN_TYPE_1] = CreateGsFromGb(gb);
    MoreLrn[LRN_TYPE_1]->LrnType = LRN_TYPE_1;
    if (MoreLrn[LRN_TYPE_4] != NULL) {
        for (int i=1; i <= 3; i++) {
            if (MoreLrn[i] == NULL) continue;
            if (DeepEq(MoreLrn[i]->ReqLits, MoreLrn[LRN_TYPE_4]->ReqLits)) {
                MoreLrn[LRN_TYPE_4]->dealloc();
                MoreLrn[LRN_TYPE_4] = NULL;
            }
        }
    }
    gb.dealloc();

    return MoreLrn[LRN_TYPE_1];
}

/*****************************************************************************
* More stuff related to learning.    }}}{{{1
*****************************************************************************/

bool AsgnT::IsRedundant(int CurLit, GsBuilder& gb, map<LitT,bool>& cache, ap<LitT>& TrigLits) {
    int ret2 = IsRedundant2(CurLit, gb, cache, TrigLits);
    //int ret1 = IsRedundant1(CurLit, gb, cache, TrigLits);
    //assert(ret1==ret2);
    cache[CurLit] = ret2;
    return ret2;
}
bool AsgnT::IsRedundant2(int CurLit, GsBuilder& gb, map<LitT,bool>& cache, ap<LitT>& TrigLits) {
    // Precond: IsInputLit(CurLit)==false
    auto(CacheIter, cache.find(CurLit));
    if (CacheIter != cache.end()) {NumRedunCacheHits++; return CacheIter->second;}
    assert(IsGateLit(CurLit));
    int IsOr = !IsAndGate(CurLit);
    GsT* pCurLs = GlitToGs[CurLit ^ IsOr];
    GsT* pDepLs = GetDepByLit(CurLit);
    int ret = !IsOr;
    for (int i=0; i < pCurLs->ReqLits.n; i++) {
        int lit = pCurLs->ReqLits[i] ^ IsOr;
        if (lit == negv(CurLit)) {abort();}
        int HasIt = false;
        if (!UsePhant && IsOr) {
            assert(pDepLs->ReqLits.n == 1);
            if (lit != pDepLs->ReqLits[0]) continue;
        }
        if (binary_search(TrigLits.begin(), TrigLits.end(), lit) == true) {HasIt=1;}
        else if (IsInputLit(lit) && gb.ImpLits.count(lit) == 1 && gb.ImpLits.count(negv(lit)) == 0) {HasIt=1;}
        else if (!IsInputLit(lit) && IsRedundant(lit, gb, cache, TrigLits)) {HasIt=1;}
        else {HasIt = 0;}
        if (IsOr) {
            if (!HasIt) continue;
            else {ret=true; break;}
        } else {
            if (HasIt) continue;
            else {ret=false; break;}
        }
    }
    return ret;
}

bool AsgnT::IsRedundant1(int CurLit, GsBuilder& gb, map<LitT,bool>& cache, ap<LitT>& TrigLits) {
    // Precond: IsInputLit(CurLit)==false
    auto(CacheIter, cache.find(CurLit));
    if (CacheIter != cache.end()) {NumRedunCacheHits++; return CacheIter->second;}
    GsT* pCurLs = GetDepByLit(CurLit);
    int ret = true;
    for (int i=0; i < pCurLs->ReqLits.n; i++) {
        int lit = pCurLs->ReqLits[i];
        if (lit == negv(CurLit)) {continue;}
        if (binary_search(TrigLits.begin(), TrigLits.end(), lit) == true) {continue;}
        if (gb.ImpLits.count(lit) == 1 && gb.ImpLits.count(negv(lit)) == 0) {continue;}
        if (!IsInputLit(lit) && IsRedundant(lit, gb, cache, TrigLits)) {continue;}
        ret = false;
        break;
    }
    return ret;
}

void AsgnT::MinimizeConflict(GsBuilder& gb, ap<LitT>& TrigLits) {
    /* Sorensson & Biere, "Minimizing Learned Clauses", SAT 2009. */
    map<LitT, bool> cache;
    for (int i=1; i < TrigLits.n; i++) {
        int CurLit = TrigLits[i];
        if (!IsInputLit(CurLit) && IsRedundant(CurLit, gb, cache, TrigLits)) {
            TrigLits[i] = TrigLits[i-1];
        } 
    }
}

map<vector<int>, GsT*> DupLrnTest;

GsT* AsgnT::CreateGsFromGb(GsBuilder& gb) {
    GsT* pCurLs = talloc(GsT, 1);
    pCurLs->init();
    pCurLs->OrigOrLrn = LRN_TAG;

    int NumLits = gb.TrigHeap.h.n;
    pCurLs->ReqLits.alloc(NumLits);
    gb.CopyTrigsTo(pCurLs->ReqLits);
    ap<LitT>& TrigLits = pCurLs->ReqLits;
    sort(TrigLits.begin(), TrigLits.end());
    TrigLits.remove_adj_dups();
    if (TrigLits.n > 0) {
        MinimizeConflict(gb, TrigLits);
        TrigLits.remove_adj_dups();
    }

    int NumImps = gb.ImpLits.size();
    pCurLs->ImpLits.alloc(NumImps);
    {auto(itLit, gb.ImpLits.begin());
    for (int i=0; i < NumImps; i++) {
        pCurLs->ImpLits[i] = *(itLit++);
    }}

    pCurLs->FreeFmla = this->FreeFmla;
    pCurLs->GateLit = TargGate ^ 1 ^ this->FreeFmla;
    assert(HasLit(pCurLs->GateLit));
    assert(FreeFmla==true || FreeFmla==false);

    if (NumLits == 0) {
        return pCurLs;
    }
    assert(IsResCompat(pCurLs->TrigAE[FreeFmla], pCurLs->ImpLits));

    /* Pick watched literals. */
    /*int iwA=0;
    int iwE=0;
    vector<int> PopSave = vector<int>();
    while (true) {
        if (gb.TrigHeap.h.n == 0) break;
        int CurLit = gb.PopTrig();
        int CurUnphant = Unphantomize(CurLit);
        if (binary_search(TrigLits.begin(), TrigLits.end(), CurUnphant) == 0) continue;
        PopSave.push_back(CurLit);
        //assert(binary_search(TrigLits.begin(), TrigLits.end(), CurUnphant));
        int qtype = QtypeFromLit(CurUnphant);
        if (IsGateLit(CurUnphant)) {
            if (!IsPhant(CurLit)) {
                int plit = Phantomize(CurLit, gb.losr);
                if (HasLit(plit)) {
                    assert(LitToChrono[plit] < LitToChrono[CurLit]);
                    continue;
                }
            }
            //int NotPhant = (CurLit == CurUnphant);
            //if (!NotPhant) assert(PlyFromLit(CurLit) == FreeFmla ^ 1);
            int okA = 1; //(NotPhant || PlyFromLit(CurLit)==0) && pCurLs->WatchAE(0,0) != CurUnphant;
            int okE = 1; //(NotPhant || PlyFromLit(CurLit)==1) && pCurLs->WatchAE(1,0) != CurUnphant;
            if (iwA < 2 && okA) {pCurLs->WatchAE(0, iwA++) = CurUnphant;}
            if (iwE < 2 && okE) {pCurLs->WatchAE(1, iwE++) = CurUnphant;}
        }
        else if (qtype == 'A') {if (iwA<2) pCurLs->WatchAE(0, iwA++) = CurUnphant;}
        else if (qtype == 'E') {if (iwE<2) pCurLs->WatchAE(1, iwE++) = CurUnphant;}
        else assert(false);
        if (iwA==2 && iwE==2) break;
    }
    foreach (itLit, PopSave) {
        gb.TrigInsert(*itLit);
    }*/
    if (*pNumClauses >= 5283) {
        stop();
    }

    pCurLs->ReqLits.sort(CmpLitByNarf);
    pCurLs->InitTrigLits(pCurLs->ReqLits);

    
    for (int ply=0; ply <= 1; ply++) {
        pCurLs->WatchAE(ply,0) = pCurLs->WatchAE(ply,1) = 0;
        int* WatchW = pCurLs->WatchRawAE + (ply << 1);
        foreach (itW, pCurLs->TrigAE[ply]) {
            int wlit = *itW;
            if (LitToChrono[wlit] > LitToChrono[WatchW[0]]) {
                WatchW[1] = WatchW[0];
                WatchW[0] = wlit;
            } else
            if (LitToChrono[wlit] > LitToChrono[WatchW[1]]) {
                WatchW[1] = wlit;
            }
        }
        if (UsePhant && ply == FreeFmla) continue;
        foreach (itW, pCurLs->TrigG) {
            int wlit = *itW;
            if (GlitToMinChrono(wlit, ply) > GlitToMinChrono(WatchW[0], ply)) {
                WatchW[1] = WatchW[0];
                WatchW[0] = wlit;
            } else
            if (GlitToMinChrono(wlit, ply) > GlitToMinChrono(WatchW[1], ply)) {
                WatchW[1] = wlit;
            }
        }
    }
    
    if (pCurLs->WatchAE(0,0) != 0) assert(pCurLs->WatchAE(0,0) != pCurLs->WatchAE(0,1));
    if (pCurLs->WatchAE(1,0) != 0) assert(pCurLs->WatchAE(1,0) != pCurLs->WatchAE(1,1));
    
    volnop();

    int BestImp = 0;
    int LastTrig = pCurLs->WatchAE(1 - pCurLs->FreeFmla, 0);
    int Penult  =  pCurLs->WatchAE(1 - pCurLs->FreeFmla, 1);
    if (!gb.ImpLits.empty()) {BestImp = *gb.ImpLits.begin();}
    foreach (itLit, gb.ImpLits) {
        if (QbFromLit[*itLit] > QbFromLit[LastTrig]) break;  // IsInputLit(LastTrig) && 
        if (LitToDL[*itLit] > LitToDL[Penult]) {
            if (LitToDL[*itLit] > LitToDL[BestImp]) {
                BestImp = *itLit;
            }
        }
    }
    pCurLs->WatchedRes = BestImp;
    if (pCurLs->ImpLits.n > 0) assert(pCurLs->WatchedRes != 0);
    if (GloSanChk >= 3) {
        int WinrLetr = GetWinner();
        foreach (it, pCurLs->ImpLits) {
            assert(QtypeFromLit(*it) == WinrLetr);
        }
    }

    if (GloSanChk >= 2) {
        vector<int> RawInts(pCurLs->ReqLits.begin(), pCurLs->ReqLits.end());
        foreach (itImp, pCurLs->ImpLits) {
            RawInts.push_back(*itImp);
        }
        RawInts.push_back(absv(pCurLs->GateLit));
        auto(itOld, DupLrnTest.find(RawInts));
        if (itOld != DupLrnTest.end()) {
            GsT* pOld = itOld->second;
            pOld->PrintIndex();
            printf("THIS SHOULDN'T BE HAPPENING\n");
            assert(0);
        }
        DupLrnTest[RawInts] = pCurLs;
    }
    /*{
    int* pCur1 = pCurLs->ReqLits.begin();  int* pEnd1 = pCurLs->ReqLits.begin();
    int* pCur2 = pCurLs->ImpLits.begin();  int* pEnd2 = pCurLs->ImpLits.begin();
    while (pCur1 != pEnd1 && pCur2 != pEnd2) {
        if      (*pCur1 < *pCur2) {++pCur1;}
        else if (*pCur2 < *pCur1) {++pCur2;}
        else {
            if (GloDbgLev >= 1950) {
                printf("NARF! %i\n", *pCur1);
            }
            break;
        }
    }
    }*/

    return pCurLs;
}




void AsgnT::AssignLrnId(GsT* pCurLs) {
    pCurLs->id = (pGloProb->clauses.n);
    pGloProb->clauses.append(pCurLs);
}

void AsgnT::AddNewLitSet(GsT* pCurLs) {
    //int NumLits = pCurLs->ReqLits.n;
    //assert(NumLits > 0);
    pCurLs->RegisterLitsHave();
    int* NewWatchesB = pCurLs->WatchRawAE;
    for (int i=0; i < 4; i++) {
        if (NewWatchesB[i] == 0) continue;
        bool ok = (i < 2) || (
                (NewWatchesB[0] != NewWatchesB[i]) &&
                (NewWatchesB[1] != NewWatchesB[i]));
        if (!ok) continue;
        pCurLs->itWatchRawAE[i] = pGloAsgn->LitWatchInsert(pCurLs->WatchRawAE[i], pCurLs);
    }
    pCurLs->itWatchRes = pGloAsgn->LitWatchInsert(pCurLs->WatchedRes, pCurLs);
    //pGloAsgn->LitWatch[pCurLs->WatchedLit[0]].insert(pCurLs);
    //pGloAsgn->LitWatch[pCurLs->WatchedLit[1]].insert(pCurLs);
    //volatile int CurRand = rand();

}

void PrintNewLitSet(GsT* pCurLs, const char* extra, int DeltaBT) {
    if (GloDbgLev >= 1900) {
        printf("%s Lrn GS #%i (L%i). ", extra, pCurLs->id, pCurLs->LrnType);
    }
    if (GloDbgLev >= 1800) {
        int winner = (pCurLs->FreeFmla==true ? 'E' : 'A');
        printf("(");
        if (DeltaBT) printf("BT%3i, ", DeltaBT);
        if (!DeltaBT) printf("       ");
        printf("%c wins %4i", winner, absv(pCurLs->GateLit));
        printf(")  ");
    }
    if (GloDbgLev >= 2000) {
        DumpLitsEnc(pCurLs->ReqLits, " , imps ");
        DumpLitsEnc(pCurLs->ImpLits, " ");
    }
    if (GloDbgLev >= 1950 && GloDbgLev < 2000) {
        printf("A ");
        DumpLitsEnc(pCurLs->TrigAE[0], ", E ");
        DumpLitsEnc(pCurLs->TrigAE[1], " ");
    }
    if (GloDbgLev >= 1800) {
        const char* sep;
        if (GloDbgLev >= 2000) {sep = "\n\n"; NeedNewl=false;}
        else if (GloDbgLev >= 1900) {sep = "\n"; NeedNewl=false;}
        else if (NumBigBt % 2 == 0) {sep = "\n"; NeedNewl=false;}
        else {sep = ""; NeedNewl = true;}
        printf("%s", sep);
    }

}


/*GsT* MakeDummyOneLitClause(int flit) {
    GsT* ret = talloc(GsT, 1);
    ret->init();
    ret->OrigOrLrn = LRN_TAG;
    ret->ReqLits.alloc(1);
    ret->ReqLits[0] = negv(flit);
    ret->FreeFmla = (QtypeFromLit(flit)=='A' ? 1 : QtypeFromLit(flit)=='E' ? 0 : die());
    return ret;
}*/



uint AsgnT::FindBtLev(GsT* pLrn) {
    /* Find the backtrack decision level. */
    int winr = FreeFmla;
    int losr = 1 - FreeFmla;
    uint BtDL;
    int MaxQb;

    int FinalLit = pLrn->WatchAE(losr, 0);
    if (LitToDL[   pLrn->WatchAE(winr, 0)] > LitToDL[FinalLit]) {
        FinalLit = pLrn->WatchAE(winr, 0);
    }
    if (IsGateLit(FinalLit) || QbFromLit[FinalLit] < QuantPfx[TargGate].blks[0].ix) {
        BtDL = LitToDL[FinalLit];
        return BtDL;
    }

    int ctlr = pLrn->GetFullCtlr(0,0);
    assert(ctlr == 0);

    BtDL = LitToDL[pLrn->WatchAE(losr, 1)];
    MaxQb = QbFromLit[pLrn->WatchAE(losr, 0)];
    assert(IsInputLit(pLrn->WatchAE(losr, 0)));
    foreach (it, pLrn->ImpLits) {
        uint ImpDL = LitToDL[*it];
        if (BtDL < ImpDL && QbFromLit[*it] < MaxQb) {
            BtDL = ImpDL;
        }
    }
    foreach (it, pLrn->TrigAE[winr]) {
        uint ImpDL = LitToDL[*it];
        if (BtDL < ImpDL && QbFromLit[*it] < MaxQb) {
            BtDL = ImpDL;
        }
    }
    return BtDL;
}
    

/*****************************************************************************
* LearnAndBacktrack.    }}}{{{1
*****************************************************************************/
GsT* AsgnT::LearnAndBacktrack(int GotTwo) {
    garr<GsT*> MoreLrn; MoreLrn.init(LRN_TYPE_4+1);
    for (int i=0; i < LRN_TYPE_4+1; i++) {MoreLrn[i] = NULL;}
    GsT* pLrn = pGloAsgn->MakeLearnedGs(MoreLrn);
    //if (pLrn == NULL) {assert(0);}
    /*
    uint BtDL = FindBtLev(pLrn);
    for (int i=LRN_TYPE_2; i < LRN_TYPE_4+1; i++) {
        GsT* pLrnX = MoreLrn[i];
        if (pLrnX != NULL) {
            uint BtLX = FindBtLev(MoreLrn[i]);
            if (BtLX < BtDL) {
                BtDL = BtLX;
                pLrn = MoreLrn[i];
            }
        }
    }
    int DeltaBT = GetCurDL() - BtDL;
    */
    for (int i=LRN_TYPE_2; i >= LRN_TYPE_1; i--) {
        GsT* pLrnX = MoreLrn[i];
        if (pLrnX) AssignLrnId(pLrnX);
    }
    if (GloDbgLev >= 2002) {
        DumpLits(0);
    }
    /*
    int losr = 1 - FreeFmla; int winr = FreeFmla;
    int FinalLit = pLrn->WatchAE(losr, 0);
    if (LitToDL[   pLrn->WatchAE(winr, 0)] > LitToDL[FinalLit]) {
        FinalLit = pLrn->WatchAE(winr, 0);
    }
    bool IsFinal = (QbFromDL(LitToDL[FinalLit]) < QuantPfx[TargGate].blks[0].ix);
    if (IsGateLit(FinalLit)) assert(IsFinal);
    */
    foreach (itGs, MoreLrn) {
        GsT* pLrnX = *itGs;
        if (pLrnX == NULL) continue;
        AddNewLitSet(pLrnX);
        if (absv(pLrnX->GateLit) == absv(pGloProb->OutGlit) && pLrnX->ReqLits.n == 0) {
            PrintNewLitSet(pLrnX, "$", 0);
            return pLrnX;
        }
    }
    int InitTarg = TargGate;
    int PrevLit;
    int InitDL = GetCurDL();

    if (MoreLrn[1]->id == 5343) {
        stop();
    }
    
    while (true) {
        if (GetLastChlit() == ZERO_LIT) break;
        if (GloDbgLev >= 2009) printf("Popping %i (DL %i)\n", GetLastChlit(), GetCurDL());
        int OldTarg = TargGate;
        int OldUndoTarg = UndoTarg[GetCurDL()];
        PrevLit = pGloAsgn->Backtrack();
        //uint PoitDL = max((int)BtDL - 1, 0);
        //assert((((pLrn->HasAllTrigs() || ExecLitSet(pLrn, EM_DRYRUN) > 0)) && GetCurDL() > 0) == (GetCurDL() != PoitDL));
        bool IsAsserting = 0;
        bool IsBlocked = 0;
        for (int i=LRN_TYPE_1; i <= LRN_TYPE_2; i++) {
            GsT* pLrnX = MoreLrn[i];
            if (pLrnX == NULL) continue;
            if (pLrnX->HasBlockedImps()) {
                IsBlocked = 1;
                break;
            }
            IsAsserting |= (pLrnX->HasAllWatchedTrigs() || ExecLitSet(pLrnX, EM_DRYRUN) > 0);
            if (IsAsserting && i==LRN_TYPE_2) pLrn = MoreLrn[LRN_TYPE_2];
            if (IsAsserting) continue;
        }
        if (IsBlocked) continue;
        if ((IsAsserting && GetCurDL() > 0)) {
            continue;
        } else {
            pGloProb->NumDoneClauses = pGloProb->clauses.n;
            if (GetCurDL() != 0) {
                TargGate = OldTarg;
                UndoTarg[GetCurDL()+1] = OldUndoTarg;
                AppendChlit(PrevLit);
                SchedForProp(PrevLit);
            } else {
                volnop();
                foreach (itGs, MoreLrn) {
                    if (*itGs == NULL) continue;
                    WatchSanityChk1(*itGs);
                    ExecLitSet(*itGs, EM_SCHED);
                    //int ok = ExecRec(*itGs, EM_SCHED);
                    //if (*itGs == pLrn && *itGs != MoreLrn[LRN_TYPE_4]) assert(ok > 0);
                }
            }
            pGloAsgn->Propagate();
            break;
        }
    }
    int DeltaBT = InitDL - GetCurDL();
    assert(DeltaBT > 0);
    int IsFinal = (InitTarg != TargGate);
    PrintNewLitSet(pLrn, (IsFinal ? "$" : (GotTwo ? "*" : "+")), DeltaBT);
    foreach (itGs, MoreLrn) {
        if (*itGs == NULL || *itGs == pLrn) continue;
        PrintNewLitSet(*itGs, "\\", 0);
    }
    if (GloDbgLev >= 2002) {
        DumpLits(0);
    }
    NumBigBt++;

    if (GloSanChk >= 3) WatchSanityChk3();
    if (GloSanChk >= 4 && !pGloAsgn->HasTargVar()) LatentExecsChk();
    if (GloDbgLev >= 2012) printf("Done Lrn #%i\n", pGloProb->clauses.n-1);

    MoreLrn.dealloc();
    return NULL;
}


int AsgnT::PickNewTarg() {
    foreach (itBlk, QuantPfx[absv(TargGate)].blks) {
        foreach (itLit, (*itBlk).lits) {
            assert(HasLit(*itLit) || HasLit(negv(*itLit)));
        }
    }
    static garri choices;
    assert(choices.n == 0);
    int IsNeg = !IsAndGate(TargGate);
    foreach (it, GlitToGs[TargGate ^ IsNeg]->TrigG) {
        int CurLit = *it ^ IsNeg;
        if (HasLit(CurLit) || HasLit(negv(CurLit))) continue;
        choices.append(CurLit);
    }
    int ret = choices[rand() % choices.n];
    choices.n = 0;
    return ret;
}

bool AsgnT::OccursOnlyOnce(int lit) {
    garri pars = pGloProb->ParentLit[lit];
    if (pars.n == 0) return true;
    int count = 0;
    foreach (it, pars) {
        int CurPar = *it;
        if (LitToLiveParent[CurPar] != 0) {count++;}
    }
    if (count > 1) return false;
    return OccursOnlyOnce(pars[0]);
}

void AsgnT::PreSolve() {
    /*for (int i=2; i <= pGloProb->LastInpLit; i += 2) {
        FreeVars.insert(i);
    }*/
    TargGate = pGloProb->OutGlit;
    LastDefnGs = pGloProb->clauses.n - 1;
    foreach (it, BattleVars) {
        pGloProb->AddTopGs(*it);
        pGloProb->AddTopGs(negv(*it));
        BattleVarList.append(*it);
    }

    LastInitGs = pGloProb->clauses.n - 1;
    pGloProb->NumDoneClauses = pGloProb->clauses.n;
    pGloAsgn->Propagate();
    foreach (it, pGloProb->clauses) {
        if (*it == NULL) continue;
        CtlrProp(*it);
        ExecLitSet(*it, EM_SCHED);
    }
    pGloAsgn->Propagate();
    WatchSanityChk3();
    if (!pGloAsgn->HasTargVar()) LatentExecsChk();

    if (0) {
        for (int lit=0; lit <= pGloProb->LastInpLit; lit++) {
            //if (LitToLiveParent[lit] == 0) 
            if (IsAlive(lit) && !IsAlive(negv(lit))) {
                die_f("Error: Input formula has a monotone literal: %i.\n", lit);
                /*if (HasLit(lit) || HasLit(negv(lit))) continue;
                int qtype = QtypeFromLit(lit);
                int flit = (qtype=='A' ? lit : (qtype=='E' ? negv(lit) : die()));
                GsT* DummyLs = MakeDummyOneLitClause(flit);
                AppendForcedLit(flit, DummyLs);
                pGloAsgn->SchedForProp(flit);*/
            }
        }
    }
    // if (0) {
    //     for (int lit=0; lit <= pGloProb->LastInpLit; lit++) {
    //         garri pars = pGloProb->ParentLit[lit];
    //         if (pars.n == 0) continue;
    //         if (!OccursOnlyOnce(lit)) continue;
    //         GsT* pCurLs = pGloProb->LitHave[lit][0];
    //         if (pCurLs->ReqLits.n != 2) continue;
    //         int owner = QtypeNumFromLit(lit);
    //         int OtherLit = pCurLs->ReqLits[0] + pCurLs->ReqLits[1] - lit;
    //         if (!IsInputLit(OtherLit)) continue;
    //         if (QbFromLit[OtherLit] <= QbFromLit[lit]) continue;
    //         GsT* pNewLs = pGloProb->AddTwoLitLrnGs(OtherLit, (lit ^ owner), owner);
    //         assert(pNewLs);
    //         stop();
    //     }
    // }
    pGloAsgn->Propagate();
    WatchSanityChk3();

    //IsRestarting = 0;
    ixChoi = 0;
    NumSinceCut = 0;
    NumUntilRestart = 2;
    LastPrintTime = GetUserTimeMicro();
    NumRestarts = 0;
    //TargStack.init();
}


/*****************************************************************************
* Top-level solver function.    }}}{{{1
*****************************************************************************/
FmlaT AsgnT::solve() {
    while (1) {
        int NewLit;
        int choice = 0;
        if (UndoTarg[GetCurDL()+1] == 0) {
            UndoTarg[GetCurDL()+1] = TargGate;
        }
        NextChoice:
        if (ixChoi < InitChoices.n) {
            choice = InitChoices[ixChoi++];
            if (pGloAsgn->HasLit(choice)) {goto NextChoice;}
            if (pGloAsgn->HasLit(negv(choice))) {InitChoices.n = 0;}
        }
        if (GloSanChk >= 5 && !pGloAsgn->HasTargVar()) LatentExecsChk();
        NewLit = pGloAsgn->ChooseRandLit(choice);
        if (NewLit == 0) {
            if (HasLit(TargGate) || HasLit(negv(TargGate))) {
                assert(0);
            }
            //TargStack.append(TargGate);
            TargGate = PickNewTarg();
            continue;
        } else {
            TriedLit[NewLit] = true;
            if (NewLit==(1<<30)) {
                assert(pGloAsgn->HasTargVar());
            } else {
                pGloAsgn->SchedForProp(NewLit);
                pGloAsgn->Propagate();
                //if (GloSanChk >= 4 && !pGloAsgn->HasTargVar()) LatentExecsChk();
            }
        }
        GloStep++;
        int GotOne = false;
        int GotTwo = false;
        GsT* pFinal;
        //if (GloSanChk >= 3) WatchSanityChk3();
        while (pGloAsgn->HasTargVar()) {
            if (GotOne) GotTwo = true;
            GotOne = true;

            pFinal = LearnAndBacktrack(GotTwo);
            if (pFinal != NULL) {
                if (absv(pFinal->GateLit) == absv(pGloProb->OutGlit) && pFinal->ReqLits.n == 0) {
                    return pFinal->FreeFmla;
                }
            }

            NumUntilRestart--;
            NumSinceCut++;
            if (GloTimeOut != 0  &&  (GetUserTimeMicro() - StartTime) > GloTimeOut) {
                fprintf(stderr, "Timed out!  (%d backtracks)\n", NumBigBt);
                exit(0);
            }
        }

        if (GotOne) {
            /* Random restarts. */
            //GsT* pLrn = pGloProb->clauses.last();
            if (NumUntilRestart <= 0 && !NoRestart) {
                NumUntilRestart = RestartCycle;
                NumRestarts++;
                if (NumRestarts <= 20) {
                    NumUntilRestart = 8;
                }
                if (GloDbgLev >= 2000) printf("\n");
                if (GloDbgLev >= 100) {
                    long long CurTime = GetUserTimeMicro();
                    if (GloDbgLev >= 2000 || (CurTime - LastPrintTime) >= 200*1000) {
                        LastPrintTime = CurTime;
                        if (NeedNewl) {printf("\n"); NeedNewl=false;}
                        printf("### RESTART %i ###  (NumBt = %i)\n",
                            NumRestarts, NumBigBt);
                    }
                }
                if (1 || NumSinceCut >= 16) {
                    //pGloAsgn->CutVsids(31);
                    NumSinceCut = 0;
                }
                pGloAsgn->restart();
                /*while (TargStack.n > 0) {
                    TargGate = TargStack.PopLast();
                }*/
                //runs.clear();
            }

            /*const uint RunSize = 3;
            if (runs.size() < RunSize && pGloAsgn->chlits.flat.n > 2) {
                pair<int, vector<int> > tmp;
                tmp.first = pLrn->ReqLits.n;
                for (int i=0; i < pGloAsgn->chlits.flat.n; i++) {
                    tmp.second.push_back(pGloAsgn->chlits.flat[i]);
                }
                runs.insert(tmp);
                pGloAsgn->restart();
                if (runs.size() == RunSize) {
                    InitChoices.n = 0;
                    foreach (iter, runs.begin()->second) {
                        InitChoices.append(*iter);
                    }
                    ixChoi = 1;
                    pGloAsgn->restart();
                    if (GloDbgLev >= 2000) printf("Selected choosen run.\n");
                }
            }*/

            /*InitChoices = pGloAsgn->chlits.flat.dup();
            ixChoi = 1;
            pGloAsgn->restart();*/
        }
    }
    assert(false);
}


/*****************************************************************************
* Helper funcs for parsing the input file.    }}}{{{1
*****************************************************************************/


/*void AddWatchOrig(GsT *pLitSet, int k, int lit) {
    assert(k==0 || k==1);
    assert(pLitSet->ReqLits.indexof(lit) != -1);
    pLitSet->WatchedLit[k] = lit;
    pGloAsgn->LitWatch[lit].insert(pLitSet);
}*/



GsT* ProbInst::AddTwoLitClause(int x, int GateLit) {
    GsT* pCurLs = (GsT*) calloc(1, sizeof(GsT) + 2*sizeof(int));
    pCurLs->id = clauses.n;
    clauses.append(pCurLs);
    pCurLs->init();
    pCurLs->OrigOrLrn = ORIG_TAG;
    pCurLs->FreeFmla = NO_WINR;
    pCurLs->ReqLits.arr = (int*)(pCurLs+1);
    pCurLs->ReqLits.n = 1;//2;
    pCurLs->ReqLits.arr[0] = x;
    //pCurLs->ReqLits.arr[1] = GateLit;
    pCurLs->GateLit = negv(GateLit);
    if (pGloAsgn->BattleVars.count(absv(GateLit)) == 1) {
        pCurLs->FreeFmla = 1^signv(pCurLs->GateLit);
        pCurLs->OrigOrLrn = LRN_TAG;
    }

    ap<int>* pTrigGrp;
    int ply = 0;
    if (IsGateLit(x)) {pTrigGrp = &pCurLs->TrigG;}
    else {
        ply = QtypeNumFromLit(x);
        pTrigGrp = &pCurLs->TrigAE[ply];
    }
    *pTrigGrp = pCurLs->ReqLits;
    pCurLs->WatchAE(ply,0) = x;
    pCurLs->WatchAE(1-ply,0) = x;
    pCurLs->itWatchRawAE[ply*2] = pGloAsgn->LitWatchInsert(pCurLs->WatchRawAE[ply*2], pCurLs);

    //pGloAsgn->FixWatchedLits(pCurLs);
    pCurLs->RegisterLitsHave();
    return pCurLs;
}

GsT* ProbInst::AddTwoLitLrnGs(int x, int y, int winr) {
    GsT* pCurLs = (GsT*) calloc(1, sizeof(GsT) + 2*sizeof(int));
    pCurLs->id = clauses.n;
    clauses.append(pCurLs);
    pCurLs->init();
    pCurLs->OrigOrLrn = LRN_TAG;
    pCurLs->FreeFmla = winr;
    pCurLs->ReqLits.arr = (int*)(pCurLs+1);
    pCurLs->ReqLits.n = 2;
    pCurLs->ReqLits.arr[0] = x;
    pCurLs->ReqLits.arr[1] = y;
    pCurLs->GateLit = pGloProb->OutGlit ^ 1 ^ winr;
    pCurLs->InitTrigLits(pCurLs->ReqLits);
    pGloAsgn->FixWatchedLits(pCurLs);
    //pCurLs->RegisterLitsHave();
    return pCurLs;
}

GsT* ProbInst::AddTopGs(int GateLit) {
    //if (UsePhant) LitToDL[Phantomize(GateLit, signv(GateLit))] = 0;
    GsT* pCurGs = (GsT*) calloc(1, sizeof(GsT) + 2*sizeof(int));
    pCurGs->OrigOrLrn = LRN_TAG;
    pCurGs->GateLit = GateLit;
    pCurGs->FreeFmla = 1^signv(pCurGs->GateLit);

    pCurGs->ReqLits.arr = (int*)(pCurGs+1);
    pCurGs->ReqLits.n = 1;
    pCurGs->ReqLits.arr[0] = GateLit;
    pCurGs->TrigG = pCurGs->ReqLits;

    pCurGs->id = pGloProb->clauses.n;
    pGloProb->clauses.append(pCurGs);
    pCurGs->init();

    pCurGs->WatchAE(0,0) = GateLit;
    pCurGs->WatchAE(1,0) = GateLit;
    pGloAsgn->FixWatchedLits(pCurGs);
    pCurGs->RegisterLitsHave();
    pGloAsgn->CtlrProp(pCurGs);
    pGloAsgn->ExecLitSet(pCurGs, EM_SCHED);
    return pCurGs;
}


pair<int,int> ProbInst::CalcExprMaxMinQb(int CurLit) {
    CurLit &= ~1;
    int ret = QbFromLit[CurLit];
    if (ret != 0) {
        int ret2 = MinQbFromGlit[CurLit];
        assert(ret2 != 0);
        return pair<int,int>(ret, ret2);
    }

    ap<int> args = AbsGlitToGs(CurLit)->ReqLits;
    int max=0;
    int min=(1<<30);
    for (int ii=0; ii < args.n; ii++) {
        int SubLit = args[ii];
        pair<int,int> MaxMin = CalcExprMaxMinQb(SubLit);
        int SubMax = MaxMin.first;
        int SubMin = MaxMin.second;
        if (SubMax > max) {max = SubMax;}
        if (SubMin < min) {min = SubMin;}
    }
    QbFromLit[CurLit] = max;
    QbFromLit[negv(CurLit)] = max;
    MinQbFromGlit[CurLit] = min;
    MinQbFromGlit[negv(CurLit)] = min;
    return pair<int,int>(max,min);
}

void CalcLitToBat(int glit, int bat) {
    if (IsInputLit(glit)) {
        LitToBat[glit]       = pGloProb->QuantBlk[QbFromLit[glit]].bat;
        LitToBat[negv(glit)] = pGloProb->QuantBlk[QbFromLit[glit]].bat;
        return;
    }
    if (QuantPfx[glit].blks.n > 0) {
        bat = glit;
    }
    if (LitToBat[glit] != 0) {
        if (LitToBat[glit] != bat) {
            die_f("Error: Gate %i appears under both %i and %i.\n", glit, LitToBat[glit], bat);
        }
        return;
    }
    LitToBat[glit] = bat;
    LitToBat[negv(glit)] = bat;
    foreach (it, AbsGlitToGs(glit)->ReqLits) {
        CalcLitToBat(*it, bat);
    }
}

/*void ProbInst::CalcGlitToBat(int CurLit, int CurBat) {
    if (IsInputLit(CurLit)) {
        CurBat = pGloProb->QuantBlk[MinQbFromGlit[CurLit]].bat;
        GlitToBat[CurLit] = CurBat;
        GlitToBat[negv(CurLit)] = CurBat;
        return;
    }
    if (pGloAsgn->IsBattleLit(CurLit)) {CurBat = CurLit;}
    if (GlitToBat[CurLit] != 0 && GlitToBat[CurLit] != CurBat) {
        die_f("Error in input file: Gate %i occurs under both %i and %i.\n",
            GlitToBat[CurLit], CurBat);
    }
    GlitToBat[CurLit] = CurBat;
    GlitToBat[negv(CurLit)] = CurBat;
    foreach (it, AbsGlitToGs(CurLit)->TrigG) {
        CalcGlitToBat(*it, absv(CurBat));
    }
}*/

void DieParse(const char* pErrMsg, Lexer* lex, ...) {
    fprintf(stderr, "Error on line %i: ", lex->LineNum);
    va_list ap;
    va_start(ap, lex);
    vfprintf(stderr, pErrMsg, ap);
    va_end(ap);
    printf("\n");
    abort();
}



/*****************************************************************************
* ParseFile.    }}}{{{1
*****************************************************************************/
ProbInst ParseFile(char* filename) {
    pGloProb = talloc(ProbInst,1);
    pNumClauses = &pGloProb->clauses.n;
    ProbInst& prob = *pGloProb;
    Lexer lex(filename);

    if (StrEqc(lex.CurTok.p, "CktQBF")) {lex.AdvTok(); lex.ReadLineSep();}
    prob.LastInpLit = lex.ReadIntParamEnc("LastInputVar") | 1;
    if (StrEqc(lex.CurTok.p, "FirstGateVar")) {lex.ReadIntParamEnc("FirstGateVar");}
    prob.LastGlit = lex.ReadIntParamEnc("LastGateVar") | 1;
    prob.OutGlit = lex.ReadIntParamEnc("OutputGateLit");
    if (StrEqc(lex.CurTok.p, "PreprocTimeMilli")) {
        lex.AdvTok();
        prob.PreprocTimeMilli = lex.ReadIntRaw();
        lex.ReadLineSep();
    }

    PhantOffset[0] = ((prob.LastGlit / 100) + 1) * 100;
    LastPhant[0] = PhantOffset[0] + prob.LastGlit;
    PhantOffset[1] = ((LastPhant[0] / 100) + 1) * 100;
    LastPhant[1] = PhantOffset[1] + prob.LastGlit;
    prob.LastPlit = LastPhant[1];

    prob.clauses.init(0, prob.LastGlit*2 + 4000);
    prob.clauses.AppendBlank();
    prob.LitHave  = talloc(garr<GsT*>, prob.LastPlit+1);
    GateDef = talloc(GateDefT, prob.LastPlit+1);
    GateType = talloc(char, prob.LastPlit+1);
    GlitToGs = talloc(GsT*, prob.LastPlit+1);
    prob.ParentLit = talloc(garri, prob.LastPlit+1);
    prob.ImmedForce = talloc(garri, prob.LastPlit+1);
    QuantPfx = talloc(QuantPfxT, prob.LastPlit+1);
    QbFromLit = talloc(short, prob.LastPlit+1);
    LitToBat = talloc(int, prob.LastPlit+1);
    MinQbFromGlit = talloc(short, prob.LastPlit+1);
    //GlitToBat = talloc(int, prob.LastPlit+1);
    prob.QuantBlk.init(0);
    prob.QuantBlk.AppendBlank();
    
    pGloAsgn = new AsgnT;
    pGloAsgn->init();
    pGloAsgn->LitWatch = new GwSetT[prob.LastPlit+1];

    /* Read quantifier blocks */
    prob.NumQuantVars = 0;
    int PrevMaxLit = 0;
    int CurMaxLit = 0;
    int PrevQvar = 0;
    while (true) {
        if (lex.CurTok.p == NULL) {DieParse("Unexpected end-of-file.", &lex, 0);}
        if (lex.CurTokType == INT_TOK_TYPE) {break;}
        int QuantGate=0;
        sscanf(lex.CurLine.p, "<q gate=%u", &QuantGate);
        if (QuantGate == 0) DieParse("Expected a line beginning with \"<q gate=\" followed by a positive integer.", &lex, 0);
        pGloAsgn->BattleVars.insert(QuantGate);
        lex.SkipLine();
        while (true) {
            if (strncmp(lex.CurLine.p, "</q>", 4) == 0) {
                lex.SkipLine();
                break;
            }
            char qtype=0;
            if      (StrEqc(lex.CurTok.p, "A")) {qtype = 'A';}
            else if (StrEqc(lex.CurTok.p, "E")) {qtype = 'E';}
            else {DieParse("Expected 'a' or 'e', but found '%.1s'.", &lex, lex.CurTok.p);}
            lex.AdvTok();
            int ixQb = prob.QuantBlk.n;
            prob.QuantBlk.AppendBlank();
            prob.QuantBlk[ixQb].qtype = qtype;
            prob.QuantBlk[ixQb].ix = ixQb;
            prob.QuantBlk[ixQb].bat = QuantGate;
            while (lex.CurTokType == INT_TOK_TYPE) {
                int qvar = lex.ReadIntEnc();
                if (qvar != absv(qvar)) {DieParse("Variable number is not even.", &lex, 0);}
                if (qvar != PrevQvar+2) {DieParse("CurQvar = %i != %i = PrevQvar+2\n", &lex, qvar, PrevQvar+2);}
                PrevQvar += 2;
                if (!prob.IsValidInpVar(qvar)) {DieParse("Variable number is out of range.", &lex, 0);}
                if (qvar < PrevMaxLit) {
                    DieParse("Vars quantified at an inner QB must be numbered higher than those quantified at an outer QB.", &lex, 0);
                }
                CurMaxLit = max(CurMaxLit, qvar);
                prob.QuantBlk[ixQb].lits.append(qvar);
                QbFromLit[qvar] = ixQb;
                QbFromLit[negv(qvar)] = ixQb;
                MinQbFromGlit[qvar] = ixQb;
                MinQbFromGlit[negv(qvar)] = ixQb;
            }
            PrevMaxLit = CurMaxLit;
            prob.NumQuantVars += prob.QuantBlk[ixQb].lits.n;
            QuantPfx[QuantGate].blks.append(prob.QuantBlk[ixQb]);
            if (ixQb > MAX_QB) {
                DieParse("Too many quantification blocks.", &lex, 0);
            }
            lex.ReadLineSep();
        }
    }



    /* Read propositional formula */
    while (lex.CurTok.p != NULL) {
        int ixExpr = lex.ReadIntEnc();   
        if (ixExpr != absv(ixExpr)) {DieParse("Expr number is not even.", &lex, 0);}
        assert(prob.IsValidExprIx(ixExpr));
        lex.ChkTok("=");
        const char* sOp = lex.ReadOpIdfr();
        garri args = lex.ReadArgs();
        if (StrEqc(sOp, "AND")) {
        } else if (StrEqc(sOp, "OR")) {
            // [x = a || b || c] is equivalent to [~x = ~a && ~b && ~c].
            ixExpr = negv(ixExpr);
            sOp = "AND";
            foreach (it, args) {*it = negv(*it);}
        } else {
            die_f("Error on line %i: Unexpected '%s'.\n", lex.LineNum, sOp);
        }
        {
            set<int> TmpSet(args.begin(), args.end());
            for (int i=0; i < args.n; i++) {
                if (TmpSet.count(negv(args[i])) != 0) {
                    DieParse("Defn contains contradictory literals: %i and its negation.", &lex, args[i]);
                }
            }
            if (TmpSet.size() < (uint) args.n) {
                DieParse("Duplicated argument.", &lex);
            }
        }
        if (args.n == 1) {
            DieParse("Operator needs at least two distinct arguments.", &lex, 0);
        }
        GsT* pCurLs = talloc(GsT, 1);
        pCurLs->id = prob.clauses.n;
        prob.clauses.append(pCurLs);
        pCurLs->init();
        GlitToGs[ixExpr] = pCurLs;
        pCurLs->OrigOrLrn = ORIG_TAG;
        pCurLs->FreeFmla = NO_WINR;
        if (StrEqc(sOp, "AND")) {
            // ((x1 & x2 & x3) <==> v) implies
            // ((x1 & x2 & x3 & ~v) ==> False) & (~x1 => ~v) & (~x2 => ~v) & (~x3 => ~v)
            GateDef[ixExpr].IsAnd = true;
            garri tmp = args.dup();
            tmp.sort(CmpLitByNarf);
            pCurLs->GateLit = ixExpr;
            if (pGloAsgn->BattleVars.count(absv(pCurLs->GateLit)) == 1) {
                pCurLs->FreeFmla = 1^signv(pCurLs->GateLit);
            }
            pCurLs->ReqLits = dup_as_ap(tmp);
            GateDef[ixExpr].args = dup_as_ap(tmp);
            GateDef[negv(ixExpr)].args = dup_as_ap(tmp);
            foreach (it, GateDef[negv(ixExpr)].args) {*it = negv(*it);}
            tmp.dealloc();
            for (int i=0; i < args.n; i++) {
                assert(prob.IsValidExprIx(args[i]));
                if (args[i] >= ixExpr) {DieParse("Argument must be less than parent.", &lex, 0);}
                prob.AddTwoLitClause(negv(args[i]), ixExpr);
                prob.ImmedForce[negv(args[i])].append(negv(ixExpr));
                prob.ParentLit[args[i]].append(ixExpr);
                prob.ParentLit[negv(args[i])].append(negv(ixExpr));
            }
            pCurLs->InitTrigLits(pCurLs->ReqLits);
            //pCurLs->InitOrigWatchLits();
        } else {assert(false);}
        assert(pCurLs->ReqLits.n == args.n);
        assert(pCurLs->ReqLits.n == pCurLs->TrigAE[0].n + pCurLs->TrigAE[1].n + pCurLs->TrigG.n);
        //AddWatchOrig(pCurLs, 1, pCurLs->ReqLits[0]);
        //AddWatchOrig(pCurLs, 0, pCurLs->ReqLits[1 + (rand() % (args.n-1))]);
        pCurLs->RegisterLitsHave();
        //if (ixExpr != prob.OutGlit) pCurLs->MakePhantom();
        lex.ReadLineSep();
        args.dealloc();
    }
    

    return prob;
}


void PrintImpUses() {
    printf("NumLrn{Rejects,Wins} = {%i, %i}.  Sf{Uses,Aborts,Revives} = {%i, %i, %i}.\n",
        GloNumLrnRejects, GloNumLrnWins, NumSfUses, NumSfAborts, NumSfRevives);
}


/*****************************************************************************
* main.    }}}{{{1
*****************************************************************************/

int main(int argc, char** argv) {
    
    GloRandSeed = RandSeedFromTime();
    int ShowTime = 1;

    /*set<int> a; a.insert(7); a.insert(5);
    set<int> b; b.insert(5); b.insert(7);
    set<set<int> > foo;
    foo.insert(a);
    printf("ASDF %i\n", foo.size());*/

    /* Parse command-line options. */
    int NumRepeat = 1;
    int ixArg = 2;
    bool AddPreprocTime = false;
    bool CondensedOut = false;
    char* PlogName = NULL;
    for (; ixArg < argc; ixArg++) {
        char* sArg = argv[ixArg];
        if (false) {}
        else if (StrEqc(sArg, "-pt")) {AddPreprocTime = true;}
        else if (StrEqc(sArg, "-q1")) {CondensedOut = true;}
        else if (StrEqc(sArg, "-OldCube")) {OldCubeWay = true;}
        else if (StrEqc(sArg, "-NoRestart")) {NoRestart = true;}
        else if (StrEqc(sArg, "-NoTime")) { ShowTime = 0; }
        else if (StrEqc(sArg, "-NoUnivGhost")) { NoUnivGhost = 1; }
        else if (ixArg+1 >= argc) {
            die_f("Either the option '%s' doesn't exists or it requires a parameter.\n", sArg);
        }
        else if (StrEqc(sArg, "-monotone")) { UseMonotone = atoi(argv[++ixArg]); }
        else if (StrEqc(sArg, "-LrnType4")) { UseLrnType4 = atoi(argv[++ixArg]); }
        else if (StrEqc(sArg, "-FindFullCtlrs")) { FindFullCtlrs = atoi(argv[++ixArg]); }
        else if (StrEqc(sArg, "-FindSemiCtlrs")) { FindSemiCtlrs = atoi(argv[++ixArg]); }
        else if (StrEqc(sArg, "-UseSemiCtlrs")) { UseSemiCtlrs = atoi(argv[++ixArg]); }
        else if (StrEqc(sArg, "-plog")) { PlogName = argv[++ixArg]; }
        else if (StrEqc(sArg, "-s") || StrEqc(sArg, "-seed")) {
            assert(++ixArg < argc);
            GloRandSeed = atoi(argv[ixArg]);
        }
        else if (StrEqc(sArg, "-d") || StrEqc(sArg, "-DbgLev")) {
            assert(++ixArg < argc);
            GloDbgLev = atoi(argv[ixArg]);
        }
        else if (StrEqc(sArg, "-SanChk")) {
            assert(++ixArg < argc);
            GloSanChk = atoi(argv[ixArg]);
        }
        else if (StrEqc(sArg, "-RestartCycle")) {
            assert(++ixArg < argc);
            RestartCycle = atoi(argv[ixArg]);
        }
        else if (StrEqc(sArg, "-TimeOut")) {
            assert(++ixArg < argc);
            GloTimeOut = (long long)(atof(argv[ixArg])*1000000);
        }
        else if (StrEqc(sArg, "-Repeat")) {
            assert(++ixArg < argc);
            NumRepeat = atoi(argv[ixArg]);
        }
        else {
            printf("Unrecognized option: '%s'\n", sArg);
            exit(1);
        }
    }

    if (PlogName) {
        PrfLog = fopen(PlogName, "w");
    }

    if (argc < 2 || (StrEqc(argv[1], "-h") || StrEqc(argv[1], "--help"))) {
        fprintf(stderr, "Usage: %s infile [OPTIONS]\n", argv[0]);
        fprintf(stderr, 
            "Options:\n"
            " '-s N'        Use N as the seed for the random number generator.\n"
            " '-d N'        Print debugging info; higher N for more info.\n"
            " '-plog FILE'  Write a proof log to FILE\n"
            " '-SanChk N'   Higher N means more sanity checks.  Default is __.\n"
            " '-TimeOut N'  Abort after N seconds.\n"
            " '-pt'         Include preprocessing time in time limit.\n"
            " '-NoTime'     Don't show elapsed time.\n"
            " '-Restart N'  Restart every N learned game states.\n"
            " '-Repeat N'   Repeat N times (for profiling purposes).\n"
            " See source code for additional options.\n"
        );
        exit(1);
    }

    if (GloDbgLev > 100) printf(
        "\n\n##############################################################################\n\n");

    /* If we extend the solver to handle QBFs with free variables, then FreeFmla
     * would be a formula in terms of the free variables.  At present, FreeFmla
     * is either true or false. */
    FmlaT FreeFmla = -1;
    while (NumRepeat-- > 0) {
        /* Seed the random number generator. */
        srand(GloRandSeed);
        
        /* Read the input QBF file. */
        try {
            ParseFile(argv[1]);
        } catch (char* pErrMsg) {
            printf("%s\n", pErrMsg);
            exit(1);
        }

        StartTime = GetUserTimeMicro();
        if (AddPreprocTime) {
            StartTime -= (((long long) pGloProb->PreprocTimeMilli) * 1000);
        }
        pGloProb->CalcExprMaxMinQb(pGloProb->OutGlit);
        CalcLitToBat(pGloProb->OutGlit, 0);
        garr<GsT*>& clauses = pGloProb->clauses;
        for (int i=1; i < clauses.n; i++) {
            pGloAsgn->FixWatchedLits(clauses[i]);
        }
        pGloAsgn->InitParents();

        QbFromLit[ZERO_LIT] = ZERO_DECLEV;
        DupLrnTest.clear();

        if (GloDbgLev >= 2020) {printf("OutGlit: %i\n", pGloProb->OutGlit);}

        printf("Seed: %2i. ", GloRandSeed);
        fflush(stdout);
        if (GloDbgLev >= 1000) {printf("\n");}

        /* Call the solver procedure. */
        pGloAsgn->PreSolve();
        if (pGloAsgn->HasTargVar()) {
            FreeFmla = pGloAsgn->FreeFmla;
        } else {
            pGloAsgn->FreeFmla = NOT_DONE;
            FreeFmla = pGloAsgn->solve();
        }
        FixNeedNewl();
        if (CondensedOut) {
            printf("%s Bt:%5i, Dec:%5i.  T: %0.3f s.\n", 
                (FreeFmla ? "true. " : "false."), NumBigBt, NumDecs,
                ShowTime * (GetUserTimeMicro() - StartTime) / 1000000.0);
        } else {
            printf("%s Bt:%5i, Dec:%5i.  ImpB:%3i, tmp1:%6i, tmp2:%7i, T: %0.3f s.\n", 
                (FreeFmla ? "true. " : "false."), NumBigBt, NumDecs, NumImpBlocked,
                NumProps,
                NumExecAttempts,
                ShowTime * (GetUserTimeMicro() - StartTime) / 1000000.0);
        }

        if (GloDbgLev >= 100) {
            printf("Num props: {1-step: %i,  n-step base: %i,  n-step lrn: %i}\n",
                NumProp_OneStep, NumProp_ManyStep_Base, NumProp_ManyStep_Lrn);
            printf("NumLrn{Rejects,Wins} = {%i, %i}.\n",
                GloNumLrnRejects, GloNumLrnWins);
        }
        if (0 && GloDbgLev >= 100) {
            PrintImpUses();
            garri& TopLits = pGloAsgn->UndoListByDL[ZERO_DECLEV];
            printf("Top unit lits: [");
            for (int i=0; i < TopLits.n; i++) {
                int CurLit = TopLits[i];
                if (!IsInputLit(CurLit)) continue;
                if (i != 0) printf(", ");
                printf("%i(%c)", decv(CurLit), tolower(QtypeFromLit(CurLit)));
            }
            printf("]\n");
            /*if (pGloAsgn->HasDoneLitSet(0)) {
                printf("Done LitSet: ");
                DumpLitsEnc(pGloAsgn->pBestDoneLs->ReqLits);
            }
            printf("CurAsgn: ");
            pGloAsgn->chlits.dump();*/
        }
    }
    

    if (FreeFmla != true && FreeFmla != false) {
        return 30;
    }
    return (FreeFmla ? 10 : 20);

}
